/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/
#ifndef _H_EXTENDIBLE_SERVER_TREE
#define _H_EXTENDIBLE_SERVER_TREE

#define PROXY_APPLICATION_PORT 7880

    // some hash functions
int hash_one_at_a_time(const char *key, int len);
int hash_additive(const char *key, int len, int prime);
int hash_rotating(const char *key, int len, int prime);

    // other utilities
int most_significant_bits(  unsigned word, int bits );
int least_significant_bits( unsigned word, int bits ); 

    // for sending dynamic lard indices over networks 
typedef struct DirectoryMessage {
    char *msg;          // this is the message that gets sent
    int  msgsize;       // this is the size of that message
    int  hdrsize;       // this is the size of the header of the message 
    int  offset;        // this is the offset at the end of the message
};

typedef char * ServerToStringFunc( void *server );
typedef void * StringToServerFunc( char *string, void *args );
typedef bool (serversAreSameFunction)( void *, void * );


class
DynamicLard {
    
public:

        // pass a toString func if you want to be able to send
        // index over a network
        // receiver will need a corresponding StringToServerFunc . . .
    DynamicLard( ServerToStringFunc *toString = NULL );

    ~DynamicLard( );
    
    void addServer( void *server );

    void setServerEqualityFunction( serversAreSameFunction *fp );

    void setServerToString( ServerToStringFunc *toString );

    bool addHelper( void *server, void *helper );

	bool replaceServer( void *server, void *replacement );
	
    bool removeServer( void *server, bool clean = false );

    void *getServer( const char *file );
   
    void dump( FILE *out, ServerToStringFunc *foreachserver );

    int send( int sd );

    int receive( int sd, StringToServerFunc *foreachserver, void *args );

    void setName( const char *name );

    const char *getName( );
    
    int numServers( );
    
    int getSize();

    void clear( bool clean = false );

private:

    unsigned depth;
    unsigned size;
    unsigned num_servers;

    void **servers;
    DirectoryMessage dmessage;
    ServerToStringFunc *toString;
    serversAreSameFunction *equality_check;
    char *name;
    
    void debug( char *fmt, ... );
    void addServerAtIndex( void *server, unsigned index );
    bool refreshDirectoryMessage( bool grow = false );
    int findIndex( void *server );
    int findAncestor( int index );
    int findDescendant( int index, int generation );
    void grow();
};

#include "nest_common.h"

#endif
