/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#ifdef HAS_LINUX_EXT2_FS_H

#include <stdio.h>

#include "fs_ext2_quota_instance.h"

/*
    Doesn't have to take care of persistency as the QuotaManager
    guarantees that!
*/

NestReplyStatus
FsExt2QuotaInstance::add_entity(const nestUserOrGroup *name,
                                const char *dev,
                                nestLotSize size,
                                int inode_count)
{
    return NEST_SUCCESS;
}

NestReplyStatus
FsExt2QuotaInstance::delete_entity(const nestUserOrGroup *name,
                                   const char *dev)
{
    return NEST_SUCCESS;
}

NestReplyStatus
FsExt2QuotaInstance::update_entity(const nestUserOrGroup *name,
                                   const char *dev,
                                   nestLotSize new_size,
                                   int new_inode_count)
{

# ifdef DQF_MASK
    struct mem_dqblk    dqb;
# else
	struct dqblk		dqb;
# endif
    dqb.dqb_bhardlimit = dqb.dqb_bsoftlimit = NUM_BLKS(new_size);
    dqb.dqb_ihardlimit = dqb.dqb_isoftlimit = new_inode_count;
    dqb.dqb_btime = dqb.dqb_itime = 0;

    int cmd = QCMD(Q_SETQLIM, name->isUser() ? USRQUOTA : GRPQUOTA);
    if (quotactl(cmd, dev, name->getId(), (caddr_t) &dqb) < 0) {
        return NEST_QUOTA_ENFORCEMENT_ERROR;
    }

    cmd = QCMD(Q_SYNC, name->isUser() ? USRQUOTA : GRPQUOTA);
    if (quotactl(cmd, 0, name->getId(), NULL) < 0) {
        return NEST_QUOTA_ENFORCEMENT_ERROR;
    }

    return NEST_SUCCESS;
}

NestReplyStatus
FsExt2QuotaInstance::view_entity_usage(const nestUserOrGroup *name,
                                       const NestDevice *dev,
                                       nestLotSize *size,
                                       int *inode_count)
{

    if (!name || !dev || !size || !inode_count) {
        return NEST_QUOTA_ENFORCEMENT_ERROR;
    }

# ifdef DQF_MASK
    struct mem_dqblk    dqb;
# else
	struct dqblk		dqb;
# endif

    int cmd = QCMD(Q_GETQUOTA, name->isUser() ? USRQUOTA : GRPQUOTA);
    if (quotactl(cmd, dev->getDevName(), name->getId(), (caddr_t) &dqb) < 0) {
        return NEST_QUOTA_ENFORCEMENT_ERROR;
    }

    /*
    *size = SIZE_IN_MB(dqb.dqb_curblocks);
    */
# ifdef DQF_MASK
    *size = dqb.dqb_curspace;		// In bytes -- good for us.
# else
    *size = dqb.dqb_curblocks * dev->getBlockSize( );
# endif
    *inode_count = dqb.dqb_curinodes;

    return NEST_SUCCESS;
}

NestReplyStatus
FsExt2QuotaInstance::initialize(const NestDevice *disk )
{
# ifdef DQF_MASK
    struct mem_dqblk    dqb;
# else
	struct dqblk		dqb;
# endif
    int             cmd;
    bool            first_time = true;

	nestUser	root( "root" );
	root.setUid( 0 );		// Force it, just in case!

    while (1) {

        cmd = QCMD(Q_GETQUOTA, USRQUOTA);
		int status;
        status = quotactl( cmd,
						   disk->getDevName(),
						   root.getId(),
						   (caddr_t) &dqb );
		if ( status >= 0 ) {
            return NEST_SUCCESS;
        }

        nestLotSize size;
        int         inode_count;

        if (this->view_entity_usage( &root, disk, &size, &inode_count)
			== NEST_SUCCESS) {
            return NEST_SUCCESS;
        }

        if (!first_time) {
            return NEST_NO_QUOTA_ENFORCEMENT;
        }

        cmd = QCMD(Q_QUOTAON, USRQUOTA);
        if (quotactl(cmd, disk->getDevName(), root.getId(), NULL) < 0) {
            if (errno != EBUSY) {
                return NEST_NO_QUOTA_ENFORCEMENT;
            }
        }

        first_time = false;
    }

    return NEST_NO_QUOTA_ENFORCEMENT;
}

#endif



