/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#ifndef NEST_CLIENT_H
#define NEST_CLIENT_H

#include "nest_common.h"
#include "nest_filetransfer.h"
#include "nest_request.h"
#include "nest_protocol.h"
#include "nest_user.h"

#include "nest_proxy_fetch.h"

#include "nest_pthread_util.h"

#include "dynamic-lard.h"

#include <queue>

// why do I need forward declarations?????
class nestProtocol;
class nestFiletransfer;
class nestRequest;
class nestAd;

typedef NestReplyStatus ( taskFunction )( nestClient * ); 
typedef NestReplyStatus ( taskCleanupFunction ) (nestClient *,
												 NestReplyStatus); 

class
nestClient {

public:

    nestClient( nestProtocol* connection, sockaddr_in *address );

    nestClient( const char *initial_cwd,
				nestProtocol* connection,
				sockaddr_in *address );
	nestClient( const nestClient *);

    ~nestClient( );

    bool sendDirectoryList( NestFileInfo *list );

    bool sendAd( nestAd *ad );

    bool sendDiskspace( nestLotSize availablebytes, nestLotSize totalbytes );

    bool sendErrorMessage(  const char* filename,
							const char* func_name,
							int error );

    bool sendAck( );

    bool sendFinalAck( );

    bool sendFinalAck(  int transferType );

    bool sendInitialAck( );

    bool sendTransferInfo( );

    bool sendTimestamp( );

    bool sendPwd( const char *pwd );

    bool sendVersion( const char *version );

    bool sendBandwidthStatistics( );

    bool sendFilesize(  nestFileSize filesize );

    NestReplyStatus sendFileinfo( const NestFileInfo *info );

    NestReplyStatus setFileattr( NestFileInfo *info,
								 const char *fname,
								 const void *fattr );
    
    bool sendPhysName( const char *name );

    NestReplyStatus sendProxyIndex( DynamicLard *index );
    
    bool sendAcl( const char *user,
                  int is_user,
                  nest_access_rights_t access );

    bool sendAcl( const NestACL *Acl );

    bool sendNestLot( NestLotInternal *r);

    bool sendLotHandle( NestLotHandle *handle);

    NestReplyStatus sendFilehandle( NestFileInfo *handle );
    
    nestRequest *receiveRequest( void );

    void dump( );

    int storeRequest( nestRequest *req );

    void endRequest( NestReplyStatus status );

    NestReplyStatus assignStorageConnection( nestProtocol *storage );

    NestReplyStatus assignOtherStorageConnection( nestProtocol *storage);

    NestReplyStatus startTransfer( );
    NestReplyStatus startTransfer( nestProtocol *storage_con );
    NestReplyStatus startTransfer( int type, long length );

    NestReplyStatus endThirdPartyTransfer( );

    NestReplyStatus openThirdParty( nestProtocolType proto );
    
    void endTransfer( NestReplyStatus status );

    bool persistent();
    
    const char *toString();

    const char *getCwd(void) const;

    const char *getHostIp();
    void setHostIp( const char *ip );

    short getClientPort();

    sockaddr_in *getSocketAddress();

    NestReplyStatus setCwd( const char *dir );

    void setProxy( bool whether );
    void setProxy( const char *proxy_address );
    bool isProxy();
    const char *proxyAddress();
    
    nestRequest* getRequest( );
    nestRequest* getLastRequest();

    nestRequest* getProxyRequest();
    void assignProxyRequest( nestRequest *req );
    nestRequest* getLastProxyRequest();

    void assignProxyClient( nestClient *client );
    nestClient *getProxyClient( );

    void assignProxyFetch( nestProxyFetch * );
    nestProxyFetch *getProxyFetch();

    void assignTaskFunction( taskFunction *fp, taskCleanupFunction *fp );
    taskFunction * getTaskFunction();
    taskCleanupFunction * getTaskCleanupFunction();

    void setTaskStatus( NestReplyStatus );
    NestReplyStatus getTaskStatus( );

    nestFiletransfer* getTransfer( );
    nestFiletransfer* getLastTransfer( );

    char *getBuffer(void);

    nestProtocol* getConnection( );

    nestProtocol* getStorageConnection( );

    nestProtocol* getOtherStorageConnection( );

    void assignConnection( nestProtocol *connection );

    float getLastBandwidth( );

    int getId( void ) const ;

	// Shortcuts
	const char *getUserIdentity( void ) const;
	const char *getUserName( void ) const;

	// Access to host info
	const char *getHostName( void ) const;
	void setHostName( const char * );
    const char *getHostIdentity( void ) const;
    void setHostIdentity( const char * );

	void setUser( nestUser *user );
	void copyUser( const nestUser *user );
	nestUser *getUser( void ) { return user; };

    bool isActive();

    bool hasSocket();

	NestReplyStatus addDirtyWrite( nestRequest *req );
    nestRequest *getDirtyWrite( );
	
    void setFlushing( bool flushing );
    bool isFlushing( );

    void setActive();
    void setPassive();

    int getProtocol() { return getConnection()->getType(); } 

    void assignRequest( nestRequest *req );

    void   setLastAllocation( nestLotSize allocation );
    nestLotSize getLastAllocation( );

    void   setAllocation( nestLotSize allocation );
    nestLotSize getAllocation( );

#if  0
    void        setUid( user_id_t uid );
    user_id_t   getUid();
#endif

    int operator==( const nestClient &other );

    void assignThirdPartyConnection( nestProtocol *third );
    nestProtocol *getThirdPartyConnection( );

private:

    void setState( int state );
    void my_endRequest( NestReplyStatus status );

    nestProtocol *connection;
    nestProtocol *storage;
    nestProtocol *otherStorage;
    nestProtocol *third_party;
    nestRequest *req, *last_req, *proxy_req, *last_proxy_req;
    nestFiletransfer *transfer, *last_transfer;

    queue< nestRequest *>  pending_reqs;
	queue< nestRequest *> *dirty_writes;

    bool flushing;  // flag that means it is flushing dirty writes

        // TODO: some configurations may not need thread safety
        // with globus threaded library they all do however
    pthread_mutex_t mutex;

    char buffer[FILEBUFFER];
	nestUser	*user;
    char *cwd;

    char *proxy_address;
    struct in_addr address;
    unsigned port;
    char *ip_address;
	char *host_name;
	char *host_identity;

    nestClient *proxy_client;
    taskFunction *task_fp;
    taskCleanupFunction *cleanup_fp;
    NestReplyStatus taskStatus;
    long taskLength;
    nestProxyFetch *proxy_fetch;

    int state;
    bool waiting_to_close;
    bool is_proxy;
    int id;
    int receivedRequests;
    struct timespec connectTime;
    nestLotSize last_allocation;
    nestLotSize allocation;

};

#endif
