/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#ifndef NEST_CLIENT_POOL_H
#define NEST_CLIENT_POOL_H

#include "nest_common.h"
#include "nest_request.h"
#include "nest_client.h"
#include "nest_pthread_util.h"
#include "nest_protocol.h"
#include "nest_protocol_manager.h"
#include <list>

#define CLIENT_ITERATOR list< nestClient * >::iterator

typedef int nestConnection;


class
nestClientPool {

  public:

    nestClientPool( bool thread_safe = false );

    ~nestClientPool( );
    
    int addClient( nestProtocol* protocol, sockaddr_in *address );

    int addClient( nestClient *client );
    
    int removeClient( nestClient *client );

    bool contains( nestClient *client );
    
    nestClient *getFront( bool block = false );
    
    CLIENT_ITERATOR begin( );

    CLIENT_ITERATOR erase( CLIENT_ITERATOR itr );
    
    CLIENT_ITERATOR end( );
    
    int getMaxfd();

    bool isThreadSafe();

    float getAveLen() { return (float)total_sizes / admissions; }

    int getMaxLen() { return max_len; }

    long getAdmissions() { return admissions; }

    bool empty();
    
    ssize_t size( );

    ssize_t clear( );

  private:
    bool signal();
    bool wait();
    bool lockMutex( const char *func, int line );
    bool unlockMutex( const char *func, int line );

    list <nestClient*> clients;
    long total_sizes;
    int max_len;
    long admissions;
    int maxfd;
    bool thread_safe;
    pthread_mutex_t *mutex;
    pthread_cond_t  *cond_var;
};

#endif

