/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "nest_error.h"
#include "nest_common.h"
#include "nest_shared.h"
#include "nest_client_util.h"

void NestFormatAccessString( nest_access_rights_t access,
                             char *access_string )
{

    // Format the access string..
    *access_string++ = (access & NEST_READ_ACCESS) ? 'r' : '-';
    *access_string++ = (access & NEST_WRITE_ACCESS) ? 'w' : '-';
    *access_string++ = (access & NEST_LIST_ACCESS) ? 'l' : '-';
    *access_string++ = (access & NEST_DELETE_ACCESS) ? 'd' : '-';
    *access_string++ = (access & NEST_INSERT_ACCESS) ? 'i' : '-';
    *access_string++ = (access & NEST_MODIFYAC_ACCESS) ? 'a' : '-';
    *access_string = '\0';
}

NestReplyStatus NestParseAccessString( nest_access_rights_t *access,
                                       const char *string )
{

    char    c;

    // Check passed in pointers..
    if (  ( NULL == string )  || ( NULL == access )  ) {
        return NEST_BAD_PARAMETERS;
    }

    *access = 0;

    if (!strcasecmp(string,"none")) {
        return NEST_SUCCESS;
    }

    // Walk through the string, one char at a time
    while (  ( c = toupper (*string) ) != '\0' )
    {
        string++;
        if ( 'R' == c ) *access |= NEST_READ_ACCESS;
        else if ( 'W' == c ) *access |= NEST_WRITE_ACCESS;
        else if ( 'L' == c ) *access |= NEST_LIST_ACCESS;
        else if ( 'D' == c ) *access |= NEST_DELETE_ACCESS;
        else if ( 'I' == c ) *access |= NEST_INSERT_ACCESS;
        else if ( 'A' == c ) *access |= NEST_MODIFYAC_ACCESS;
        else return NEST_BAD_PARAMETERS;
    }

    // Done
    return NEST_SUCCESS;
}


NestReplyStatus
NestParseLotSize( const char *arg, nestLotSize *size )
{
	return NestParseSizeString( arg, size, 'B' );
}

NestReplyStatus
NestParseLotTime( const char *arg, int *seconds )
{
	return NestParseTimeString( arg, seconds, 'S' );
}

const char *
NestFormatLotSize( nestLotSize size, char *buf, int bufsize, bool raw )
{
	return NestFormatSizeString( size, buf, bufsize, raw );
}

const char *
NestFormatFileSize( nestFileSize size, char *buf, int bufsize, bool raw )
{
	return NestFormatSizeString( size, buf, bufsize, raw );
}

const char *
NestFormatLotTime( int seconds, char *buf, int bufsize, bool raw )
{
	return NestFormatTimeString( seconds, buf, bufsize, raw );
}
