/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#ifndef NEST_COLLECTION_MANAGER_H
#define NEST_COLLECTION_MANAGER_H

#include "nest_common.h"
#include <string>
#include "time.h"

class 
nestCollection
{

  public:
    nestCollection( const string &name, const string &path, ClassAd *ad );
    nestCollection( ClassAd *ad );
    nestCollection( void );
    ~nestCollection();

	NestReplyStatus Init( const string &name,
						  const string &path,
						  ClassAd *ad = NULL );
	void Clear( void );
	bool nestCollection::buildCollectionFromAd( ClassAd *ad );

	bool isValid( void ) const { return valid; };
    const string * getName( ) const { return &name; };
    const string * getPath( ) const { return &path; };

	time_t getMtime( ) const { return mtime; };
	const string * getVersion( ) const { return &version; };
	const string * getComment( ) const { return &comment; };

	NestReplyStatus setMtime( time_t mtime );
	NestReplyStatus setVersion( const string &version );
	NestReplyStatus setVersion( const char *version );
	NestReplyStatus setComment( const string &comment );
	NestReplyStatus setComment( const char *comment );

	// Get / set my ClassAd (for CollectionManager only)
	ClassAd * getAd( void ) { return this->ad; };

private:
	bool	valid;
	string 	name;
    string	path;
	time_t 	mtime;
	string	version;
	string	comment;
	ClassAd	*ad;
};

class
nestCollectionVisitor
{
  public:
	nestCollectionVisitor( void ) { };
	virtual ~nestCollectionVisitor( void ) { };
	virtual NestReplyStatus visit( const nestCollection & ) = 0;
};

class
nestCollectionManager 
{

  public:

	nestCollectionManager( void );
	~nestCollectionManager( void );

    int init( );
	NestReplyStatus initialize(ClassAdCollection *resources);
	NestReplyStatus recover(ClassAdCollection *resources);

    NestReplyStatus createAssociation( const string &name,
									   const string &path );

    NestReplyStatus removeAssociation( const string &name );

    const string * getPath( const string &name );

    nestCollection* getCollection( const string &name );
    bool checkName( const string &name );

	NestReplyStatus updateMtime( nestCollection *col, time_t mtime );
	NestReplyStatus updateVersion( nestCollection *col,
								   const string &version );
	NestReplyStatus updateComment( nestCollection *col,
								   const string &comment );
	NestReplyStatus update( nestCollection *col );

	NestReplyStatus VisitAllCollections( nestCollectionVisitor &visitor );

  private:
	ClassAdCollection		*collection_list;
	ClassAdParser			parser;

	LocalCollectionQuery    query;
	ViewName                collection_view;

	// Private methods
	NestReplyStatus setupView( ClassAdCollection *resources );
	
};

extern nestCollectionManager	*collection_manager;

#endif
