/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "nest_common.h"
#include "nest_request.h"
#include "nest_pthread_util.h"


bool    NestEnableLots      = true;
const char    *NestAdminUser      = "";
user_id_t   AdminUid;

const char    *NestSystemUser     = "system";
const char    *NestAnonymousUser  = "";

const char    *SystemPwdFile      = "/etc/passwd";
const char    *SystemGrpFile      = "/etc/group";
const char    *GridMapFile        = "/etc/grid-security/grid-mapfile";

const char    *MetaDir            = "./__NEST__";

const char    *ActualGroup        = "";
user_id_t   ActualGid;


int admin_pid = 0;
void
NestcheckPid( const char *file, const char *function )
{
    nest_debug( NEST_DEBUG_ADMIN, "%s:%s - %d =?= %d\n",
            file, function, NestGetpid(), admin_pid );
    if ( NestGetpid() != admin_pid ) {
        nest_die( "Admin violoation in %s:%s.  %d != %d\n",
                file, function, NestGetpid(), admin_pid );
    }
}


    // User management utilities

int
Nestset_grid_mapfile( const char *fname )
{
    return NestOpen( fname, O_RDONLY );
}

grid_mapfile_entry  *
Nestget_grid_mapfile_entry(int grid_mapfile_fd)
{
    static  grid_mapfile_entry  entry;
    if (grid_mapfile_fd < 0) {
        return NULL;
    }

    char    ch; //line[sizeof(grid_mapfile_entry)];
    bool    error;
    int     i;

    error = true;
    while (read( grid_mapfile_fd, &ch , 1) == 1) {
        if ( (ch != ' ') && (ch != '\t') && (ch != '\n')) {
            error = false;
            break;
        }
    }
    if (error || ch != '"') {
        return NULL;
    }

    i = 0;
    //entry.identity[i++] = ch;
    error = true;
    while (read( grid_mapfile_fd, &ch , 1) == 1) {
        if (ch == '"') {
            error = false;
            break;
        }
        entry.identity[i++] = ch;
    }
    if (error) {
        return NULL;
    }
    entry.identity[i] = 0;

    error = true;
    while (read( grid_mapfile_fd, &ch , 1) == 1) {
        if ((ch != ' ') &&
            (ch != '\t') &&
            (ch != '\n')) {
            error = false;
            break;
        }
    }
    if (error) {
        return NULL;
    }

    i = 0;
    entry.name[i++] = ch;
    error = true;
    while (read( grid_mapfile_fd, &ch , 1) == 1) {
        if ((ch == ' ') ||
            (ch == '\t') ||
            (ch == '\n')) {
            error = false;
            break;
        }
        entry.name[i++] = ch;
    }
    if (error) {
        return NULL;
    }
    entry.name[i] = 0;

    return &entry;
}

int
Nestend_grid_mapfile(int grid_mapfile_fd)
{
    if (grid_mapfile_fd >= 0) {
        NestClose(grid_mapfile_fd);
    }
    return 0;
}

#if 0
nestPwent::getpwuid(  user_id_t uid, const char *pwfile )
{
    FILE    *p_fp = fopen(pwfile, "r");
    int     ret_val;

    if (p_fp == NULL) {
        return -1;
    }

    ret_val = Nestfgetpwuid(u, p, p_fp);

    fclose(p_fp);
    return(ret_val);
}

struct passwd *
Nestnew_pw_ent()
{

    struct passwd    *ret_ptr = (struct passwd *) 
        NestMalloc(sizeof(struct passwd));

    ret_ptr->pw_name   = (char *) NestMalloc (NEST_NAMELEN);
    ret_ptr->pw_passwd = (char *) NestMalloc (NEST_NAMELEN);
    ret_ptr->pw_gecos  = (char *) NestMalloc (NEST_NAMELEN * 10);
    ret_ptr->pw_dir    = (char *) NestMalloc (NEST_NAMELEN);
    ret_ptr->pw_shell  = (char *) NestMalloc (NEST_NAMELEN);

    return ret_ptr;
}


void
Nestfree_pw_ent(struct passwd *ptr)
{

    NestFree(ptr->pw_name);
    NestFree(ptr->pw_passwd);
    NestFree(ptr->pw_gecos);
    NestFree(ptr->pw_dir);
    NestFree(ptr->pw_shell);

    NestFree(ptr);
}

void
Nestpw_assign(struct passwd *dst, struct passwd *src)
{

    //memcpy((void *) dst, (void *) src, sizeof(struct passwd));

    dst->pw_uid = src->pw_uid;
    dst->pw_gid = src->pw_gid;
    strcpy(dst->pw_name, src->pw_name);
    strcpy(dst->pw_passwd, src->pw_passwd);
    strcpy(dst->pw_gecos, src->pw_gecos);
    strcpy(dst->pw_dir, src->pw_dir);
    strcpy(dst->pw_shell, src->pw_shell);

}


int
Nestfgetpwuid(user_id_t u, struct passwd *p, FILE *fp)
{

    fpos_t    save_pos;
    struct passwd *tmp_p;

    if (!fp || !p){
        return -1;
    }

    if (fgetpos(fp, &save_pos) < 0) {
        return -1;
    }

    while ( (tmp_p = fgetpwent(fp)) != NULL){
        if ((user_id_t)u == (user_id_t)(tmp_p->pw_uid)) {
            break;
        }
    }

    fsetpos(fp, &save_pos);

    if (!tmp_p) {
        return -1;
    }

    Nestpw_assign(p, tmp_p);
    return 0;
}


int
Nestfgetpwnam(const char *user, struct passwd *p, FILE *fp)
{

    fpos_t    save_pos;
    struct passwd *tmp_p;

    if (!user || !p || !fp){
        return -1;
    }

    if (fgetpos(fp, &save_pos) < 0) {
        return -1;
    }

    LOCK_MUTEX( &nest_common_mutex );
    while( (tmp_p = fgetpwent(fp)) != NULL){
        if(!strcmp(user, tmp_p->pw_name)) {
            break;
        }
    }
    UNLOCK_MUTEX( &nest_common_mutex );

    fsetpos(fp, &save_pos);

    if (!tmp_p) {
        return -1;
    }

    Nestpw_assign(p, tmp_p);
    return 0;
}


int
Nestmy_getpwuid(user_id_t u, struct passwd *p, const char *fname)
{

    FILE    *p_fp = fopen(fname, "r");
    int     ret_val;

    if (p_fp == NULL) {
        return -1;
    }

    ret_val = Nestfgetpwuid(u, p, p_fp);

    fclose(p_fp);
    return(ret_val);
}

int
Nestmy_getpwnam(const char *u, struct passwd *p, const char *fname)
{

    FILE    *p_fp = fopen(fname, "r");
    int     ret_val;

    if (p_fp == NULL) {
        return -1;
    }

    ret_val = Nestfgetpwnam(u, p, p_fp);

    fclose(p_fp);
    return(ret_val);
}
#endif

#if 0
int
NestGet_uid(user_id_t *u, const char *uname, const char *fname)
{
    struct passwd *pw;

    pw = Nestnew_pw_ent();

    nest_debug( NEST_DEBUG_CLIENT, "Looking up uid for uname %s from %s.\n",
            uname, fname );

    if (Nestmy_getpwnam(uname, pw, fname) < 0) {
		nest_debug( NEST_DEBUG_CLIENT, "No passwd entry found for user '%s'\n",
					uname );
        Nestfree_pw_ent(pw);
        return -1;
    }

	nest_debug( NEST_DEBUG_CLIENT, "UID for '%s' found %d\n",
				uname, pw->pw_uid );
    *u = pw->pw_uid;
    Nestfree_pw_ent(pw);
    return 0;
}
#endif

#if 0
int
NestGet_uname(char *uname, user_id_t uid, const char *fname )
{

    struct passwd *pw;

    nest_debug( NEST_DEBUG_CLIENT, "Looking up uname for uid %d from %s.\n",
				uid, fname );

	nestPwent	pwent;
	pwent.findent( uid, fname );

	const char *name = pwent.getent_name( );
	if ( name ) {
		if( this->userName ) {
			free( (void*) this->userName );
			}
			this->userName = strdup( name );
		}


    if (Nestmy_getpwuid(u, pw, fname) < 0) {
        Nestfree_pw_ent(pw);
        return -1;
    }

    strcpy(uname, pw->pw_name);
    Nestfree_pw_ent(pw);
    return 0;
}
#endif

    //  Group management utilities

int
Nestgetgrpent(group_ent *grp_ent, FILE *fp)
{

    char    c,
            tmp_buf[NEST_NAMELEN];

    int        fld = 1,
            i=0;

    c = getc(fp);

    while(!feof(fp)){
        switch(c) {
		case ',':
		case ':':
		case '\n':    tmp_buf[i++] = '\0';
			
			switch(fld) {
				
			case 1:
				strncpy(grp_ent->grp_name, tmp_buf, i);
				break;

			case 2:
				strncpy(grp_ent->grp_passwd, tmp_buf, i);
				break;

			case 3:
				sscanf(tmp_buf,"%d",&i);
				grp_ent->grp_gid = i;
				break;

			default:
				strncpy(grp_ent->grp_entries[fld - 4], 
						tmp_buf, i);
			}

			fld++;

			if (fld > MAX_USR_ENTRIES + 3) {
				nest_debug( NEST_DEBUG_GROUP_MNGMNT, 
							"Number of users in group %s > %d\n",
							grp_ent->grp_name,
							MAX_USR_ENTRIES);
                            return 0;
			}

			i = 0;

			if ( c == '\n') {
				strcpy(grp_ent->grp_entries[fld - 4], "");
				strcpy(grp_ent->grp_entries[fld - 3], "");//Paranoia
				return 0;
			}
			break;

		default:
			tmp_buf[i++] = c;
			if (i > NEST_NAMELEN) {
				nest_debug( NEST_DEBUG_GROUP_MNGMNT, 
							"Name too big to accomodate!\n");
				i--;
			}
        }
        c = getc(fp);
    }
    return -1;
}


int
Nestputgrpent(group_ent *grp_ent, FILE *fp)
{

	char    tmp_buf[NEST_NAMELEN];
    char    flag = 'f';

    sprintf(tmp_buf,"%lu",(unsigned long)grp_ent->grp_gid);

    fprintf(fp,"%s:%s:%s:", grp_ent->grp_name, grp_ent->grp_passwd, tmp_buf);
    fflush(fp);

    for(int i=0; i < MAX_USR_ENTRIES; i++) {
        if (!strcmp(grp_ent->grp_entries[i],"")) {
            if (flag == 't') {
                fseek(fp, -1, SEEK_CUR);
            }
            fprintf(fp,"\n");
            return 0;
        }
        flag = 't';
        fprintf(fp,"%s,",grp_ent->grp_entries[i]);
    }
    return 0;
}


int
Nestgetgrpnam(const char *grp_name, group_ent *grp_ent, const char *fname)
{

    FILE    *tmp_fp = fopen(fname, "r");

    if (tmp_fp == NULL) {
        nest_perror("fopen");
        return -1;
    }

    while(Nestgetgrpent(grp_ent, tmp_fp) == 0) {
        if (!strcmp(grp_ent->grp_name, grp_name)) {
            fclose(tmp_fp);
            return 0;
        }
    }

    fclose(tmp_fp);
    return -1;
}

int
Nestgetgrpgid(user_id_t grp_gid, group_ent *grp_ent, const char *fname)
{

    FILE    *tmp_fp = fopen(fname, "r");

    if (tmp_fp == NULL) {
        nest_perror("fopen");
        return -1;
    }

    while(Nestgetgrpent(grp_ent, tmp_fp) == 0) {
        if ((unsigned)grp_ent->grp_gid == (unsigned)grp_gid) {
            fclose(tmp_fp);
            return 0;
        }
    }

    fclose(tmp_fp);
    return -1;
}



char *NEST_TRANSFER_NAMES[] = {
    "NestTime Request",
    "Client Send File",
    "Client Recv File",
    "File Delete Request",
    "NestBandwidth Statistics Request",
    "List Available Files Request",
    "Filesize Request",
    "Server Terminate Request",
    "Client Terminate Request",
    "Test Message",
    "Test Client to Server NestBandwidth",
    "Test Server to Client NestBandwidth",
    "Request Done"
};

int
NestAccept( int sockfd, struct sockaddr *cliaddr, socklen_t *addrlen )
{
    int accepted;

    while( true ) {
        accepted = accept( sockfd, cliaddr, addrlen );
        if ( accepted < 0 ) {
            if ( errno == EINTR ) {
                continue;
            } else {
                nest_perror( "accept" );
                exit( ERROR_EXIT );
            }
        } else {
            nest_debug( NEST_DEBUG_FD | NEST_DEBUG_FD, 
						"Accepted %d on port %d\n",
						accepted, sockfd );
            return accepted;
        }
    }
}

int
NestChangeUser( const char *name )
{
    struct passwd *p;

    LOCK_MUTEX( &nest_common_mutex );
    if( ( p = getpwnam( name ) ) == NULL ) return -1;
    UNLOCK_MUTEX( &nest_common_mutex );

    return setuid( p->pw_uid );
}

uid_t
NestGetuid()
{
    return getuid();
}

gid_t
NestGetgid()
{
    return getgid();
}

int
NestSeteuid( uid_t uid )
{
    return seteuid(uid);
}

int
NestSetegid( gid_t gid )
{
    return setegid(gid);
}

pid_t
NestSetpgrp()
{
    int ret = setpgrp();
    if ( ret < 0 ) {
        perror( "setpgrp" );
        assert( 0 );
    }
    return ret;
}

pid_t
NestGetpid( )
{
    return getpid();
}

// this function has been added for the test bandwidth protocols which
// just send and receive into memory buffers and don't go to disk
// insure was complaining however about uninit'ed mem reads 
void
NestBufferInit( char * buffer, unsigned long size )
{
    memset( buffer, 0, size );
}


int
NestSpipe( int fd[2] )
{
    return( socketpair( AF_UNIX, SOCK_STREAM, 0, fd ) );
}

char *
NestFdGets( char *buffer, ssize_t bufsize, int fd )
{

    for( ssize_t i = 0; i < bufsize; i++ ) {
        if ( NestReadn( fd, (void *)&buffer[i], 1 ) < 0 ) return NULL;
        if ( buffer[i] == '\n' ) {
            if ( i + 1 < bufsize ) buffer[i+1] = '\0';
            return buffer; 
        }
    }
    return NULL;
}

NestReplyStatus
NestMergeFdsets( fd_set *dst, fd_set *src, int maxfd )
{

    int m = 0;
    for( int i = 0; i < maxfd; i++ ) {
        if ( NestFd_isset( i, src ) ) {
            NestFd_set( i, dst );
            m++;
        }
    }
    nest_debug( NEST_DEBUG_FD,
				"Merged fd_sets (%d fds were merged)\n", m );
        
    return NEST_SUCCESS;
}

int
NestSelect( int numfds, fd_set * readfds, fd_set * writefds, 
        fd_set * exceptfds, struct timeval * timeout )
{
    CHECK_PID;
    int n = -1;
    while( n < 0 ) {
		n = select( numfds, readfds, writefds, exceptfds, timeout );
		if ( n < 0 ) {
			if ( errno == EINTR )
				continue;
			nest_warn( __FILE__, __FUNCTION__, "select: %s (maxfd %d)\n", 
					   strerror( errno ), numfds );
			int old_level = get_nest_debug_level();
			nest_debug( NEST_DEBUG_ALWAYS, "select failed: %s (maxfd %d)\n",
						strerror( errno ), numfds );
			NestFd_print( NEST_DEBUG_FD, readfds, numfds, true );
			set_nest_debug_level( old_level );
			assert( 0 ); 
			return -1;
		}
    }

    nest_debug( NEST_DEBUG_FD, "Ready reads: " );
    NestFd_print( NEST_DEBUG_FD, readfds, numfds );
    nest_debug( NEST_DEBUG_FD, "\nReady writes: " );
    NestFd_print( NEST_DEBUG_FD, writefds, numfds );

    return n;
}

bool
NestReadLink( const char *path, char *linkto, int length )
{
    int result;
    result = readlink(path,linkto,length);
    if(result>0) {
        linkto[result] = 0;
        return true;
    } else {
        return false;
    }
}

#ifdef HAS_SCM_RIGHTS

const char*
NestInet_ntop( int family, const void*addrptr, char*strptr, ssize_t len)
{
    LOCK_MUTEX( &nest_common_mutex );
    const char * p;
    if ( ( p = inet_ntop( family, addrptr, strptr, len ) ) == NULL ) {
        nest_perror( "inet_ntop" );
        exit (ERROR_EXIT );
    }
    const char *ret = NestStrdup( p );
    UNLOCK_MUTEX( &nest_common_mutex );
    return ret;    
}

// this functions receives a fd from the sd param and returns the fd 
int 
NestReceiveFd( int sd )
{
    
    struct iovec vector;    /* file name for the passed file - phony */
    struct msghdr msg;        /* full message */
    static struct cmsghdr *cmsg = NULL;
                            /* control message which includes the fd */
    int fd;

        // set up the iovec to read in the filename
    char filename[NEST_MESSAGESIZE];
    vector.iov_base = filename;
    vector.iov_len  = NEST_MESSAGESIZE;

        // the message we're expecting to receive
    msg.msg_name = NULL;
    msg.msg_namelen = 0;
    msg.msg_iov = &vector;
    msg.msg_iovlen = 1;

        // the control message
    if ( cmsg == NULL ) {
            // it will be NULL the first time the function is called
        cmsg = (struct cmsghdr *)
            NestMalloc( sizeof(struct cmsghdr) + sizeof(fd) );
        cmsg->cmsg_len = sizeof(struct cmsghdr) + sizeof(fd);
    }

        // connect the message and the control message
    msg.msg_control = cmsg;
    msg.msg_controllen = cmsg->cmsg_len;
    
    if ( recvmsg( sd, &msg, 0 ) <= 0 ) {
        nest_perror( "recvmsg" );
        assert( 0 );
        return ERROR_EXIT;
    }

        // grab the file descriptor
    memcpy( &fd, CMSG_DATA( cmsg ), sizeof( fd ) );
    nest_debug( NEST_DEBUG_FD | NEST_DEBUG_FD, "Received fd %d\n",fd);
    return fd;
}

// this function passes the param fd to the socket sd
// returns OK or ERROR_EXIT on failure
NestReplyStatus
NestSendFd( int sd, int fd )
{
    struct iovec   vector;     /* some data to pass with fd */
    struct msghdr  msg;         /* the complete message */
    static struct cmsghdr *cmsg = NULL;
                            /* the control message which includes the fd */

        // send a bogus filename to make the code work
        // I can't figure out why this is necessary
    char filename[NEST_MESSAGESIZE];
    memset( filename, 0, NEST_MESSAGESIZE );
    NestSnprintf( filename, sizeof(filename), "%s", "foo" );
    vector.iov_base = filename;
    vector.iov_len  = NestStrlen( filename ) + 1;

        // assemble first part of message
    msg.msg_name = NULL;
    msg.msg_namelen = 0;
    msg.msg_iov = &vector;
    msg.msg_iovlen = 1;

        // assemble the control message
    if ( cmsg == NULL ) {
            // cmsg is NULL the first time the function is called
        cmsg = (struct cmsghdr *)NestMalloc(
                sizeof(struct cmsghdr) + sizeof(fd) );
        cmsg->cmsg_len   = sizeof( struct cmsghdr) + sizeof( fd );
        cmsg->cmsg_level = SOL_SOCKET;
        cmsg->cmsg_type  = SCM_RIGHTS;
    }

        // copy the file descriptor onto the end of the control message
    memcpy( CMSG_DATA( cmsg ), &fd, sizeof( fd ) );

        // connect the message and the control message
    msg.msg_control = cmsg;
    msg.msg_controllen = cmsg->cmsg_len;

    if ( sendmsg( sd, &msg, 0 ) != (int) vector.iov_len ) {
        nest_perror( "sendmsg" );
        return ErrnoToNest( errno );
    }

    return NEST_SUCCESS;
}

#else /* HAS_SCM_RIGHTS */

int
NestReceiveFd( int sd )
{

    struct strrecvfd recvfd;

    if( ioctl(sd, I_RECVFD, &recvfd) < 0)
        return -1;
    else
        return recvfd.fd;

}

NestReplyStatus
NestSendFd( int sd, int fd )
{

    if( ioctl(sd, I_SENDFD, fd) < 0){
        nest_perror( "ioctl" );
        return ErrnoToNest( errno );
    }

    return NEST_SUCCESS;
}

#endif /* HAS_SCM_RIGHTS */

int
NestChmod( const char *pathname, mode_t mode )
{
    int reply = chmod(pathname, mode);

    if (reply < 0) {
        nest_perror("chmod");
    }

    return reply;
}

int
NestFchmod( int fd, mode_t mode )
{
    int reply = fchmod(fd, mode);

    if (reply < 0) {
        nest_perror("fchmod");
    }

    return reply;
}

int
NestChown ( const char *pathname, uid_t uid, gid_t gid )
{
    int reply = chown(pathname, uid, gid);

    if (reply < 0) {
        nest_perror("chown");
    }

    return reply;
}

int
NestLchown ( const char *pathname, uid_t uid, gid_t gid )
{
    int reply = lchown(pathname, uid, gid);

    if (reply < 0) {
        nest_perror("lchown");
    }

    return reply;
}

int
NestFchown( int fd, uid_t uid, gid_t gid )
{
    int reply = fchown(fd, uid, gid);

    if (reply < 0) {
        nest_perror("fchown");
    }

    return reply;
}


/**
takes a sin and return the hostname
*/
char *
NestDns_lookup( const struct sockaddr_in *from )
{
    
        // turn this off
    //static char *deprecated = "dns lookup turned off";
    //return deprecated; 

    LOCK_MUTEX( &nest_common_mutex );
    struct hostent * mach;
    char *ret = NULL;
    mach = gethostbyaddr( (char *)&from->sin_addr,
        sizeof( struct in_addr ), AF_INET );
    if (mach == NULL) {
        const char *error_string = NULL;
        #ifdef HAS_HSTRERROR
            error_string = hstrerror( h_errno );
        #else
            error_string = strerror( h_errno );
        #endif
        nest_warn( __FILE__, __FUNCTION__, "gethostbyaddr %s: %s\n",
                (char *)&from->sin_addr, error_string );    
        ret = NestStrdup( (char *)&from->sin_addr );
    } else {
        ret = NestStrdup( mach->h_name );
    }
    UNLOCK_MUTEX( &nest_common_mutex );
    return ret;
}

char *
NestGetIP( struct in_addr addr )
{
    LOCK_MUTEX( &nest_common_mutex );
    char *ret = NestStrdup( inet_ntoa( addr ) );
    UNLOCK_MUTEX( &nest_common_mutex );
    return ret;
}

NestReplyStatus 
NestGetcwd( char *buffer, size_t size )
{
    LOCK_MUTEX( &nest_common_mutex );
    char *reply = getcwd(buffer, size);
    UNLOCK_MUTEX( &nest_common_mutex );

    if (reply == NULL) {
        nest_perror("getcwd");
        return ErrnoToNest( errno );
    } else {
        return NEST_SUCCESS;
    }
}


void
NestChdir( const char *pathname )
{
    if ( chdir( pathname ) < 0 ) {
        nest_die( "chdir to %s failed: %s\n", pathname, strerror(errno));
    }
}

int
Nestsafe_Chdir( const char *pathname )
{
    return chdir( pathname );
}

int
NestMkdir( const char *pathname, mode_t mode )
{
    int reply = mkdir(pathname, mode);

    if (reply < 0 && errno != EEXIST ) {
        nest_warn( __FILE__, __FUNCTION__,"mkdir of %s failed: %s\n", 
                pathname, strerror( errno ) );
    }

    return reply;
}

int
NestRmdir( const char *pathname )
{
    int reply = rmdir(pathname);

    if (reply < 0 && errno != ENOENT) {
        nest_perror("rmdir");
    }

    return reply;
}

void
NestClearFileFlags( int fd, int flags ) {
    int val;
    if ( ( val = fcntl( fd, F_GETFL, 0 ) ) < 0 ) {
        nest_die( "fcntl: %s", strerror( errno ) );
    }
    val &= ~(flags);
    if ( fcntl( fd, F_SETFL, val ) < 0 ) {
        nest_die( "fcntl: %s", strerror( errno ) );
    }
}

void 
NestSetFileFlags( int fd, int flags )
{
    int val;
    if ( ( val = fcntl( fd, F_GETFL, 0 ) ) < 0 ) {
        nest_die( "fcntl: %s", strerror( errno ) );
    }
    val |= flags;
    if ( fcntl( fd, F_SETFL, val ) < 0 ) {
        nest_die( "fcntl: %s", strerror( errno ) );
    }
}

int
NestShutdown( int sd, int how )
{
    //CHECK_PID;
    int ret = shutdown( sd, how );
    if ( ret != 0 ) {
        if ( errno != ENOTCONN ) {
            nest_warn( __FILE__, __FUNCTION__,
					   "shutdown of %d failed: %s\n", sd, strerror( errno ) );
        }
    } else {
        nest_debug( NEST_DEBUG_FD | NEST_DEBUG_ADMIN, 
        //nest_printf( 
                "NestShutdown %d: %s\n", sd, 
                ( (ret<0) ? strerror( errno ) : "success" ) );
    }
    return ret;
}

int
NestCreat( const char *pathname, mode_t mode ) {

    return creat( pathname, mode );
}

off_t
NestLseek( int fd, off_t offset, int whence ) {
    off_t retval = lseek( fd, offset, whence );
    if ( retval == (off_t)-1 ) nest_perror( "lseek" );
    return retval;
}

int
NestRename( const char *oldpath, const char *newpath )
{
    int retval = rename(oldpath, newpath);

    if (retval < 0) {
        char out[3*NEST_MAXPATHLEN];
        NestSnprintf(out,
                     3*NEST_MAXPATHLEN,
                     "rename - %s %s",
                     oldpath,
                     newpath);
        out[3*NEST_MAXPATHLEN - 1] = 0;
        nest_perror(out);
    }

    return retval;
}

int
NestStat( const char *file, nest_stat *st )
{
    int retval = stat(file, st);

    if (retval < 0) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "stat of %s: %s\n",
                    file,
                    strerror( errno ) );
    }

    return retval;
}

int
NestFstat( int fd, nest_stat *st ) {
    int retval = fstat(fd, st);

    if (retval < 0) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "fstat %s\n",
                    strerror( errno ));
    }

    return retval;
}

int
NestLstat( const char *file, nest_stat *st )
{
    int retval = lstat(file, st);

    if (retval < 0) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "lstat of %s: %s\n",
                    file,
                    strerror( errno ) );
    }

    return retval;
}


int NestStatFS( const char *path, nest_fsstat *s ) 
{
#if defined(HAS_SYS_STATFS_H)
    return statfs(path,s);
#elif defined(HAS_SYS_STATVFS_H)
    return statvfs(path,s);
#else
#error Must have either statvfs or statfs.
#endif
}

int
NestSymlink( const char *oldpath, const char *newpath )
{

    return symlink( oldpath, newpath );
}

int
NestReadlink( const char *path, char *buf, size_t buf_len )
{

    int retval;

    if (!buf || !buf_len) {
        return -1;
    }

    retval = readlink( path, buf, buf_len );

    /*
        stupid readlink doesn't put a '\0' at the end of the string
        I hate readlink!!
    */

    if (retval < 0) {
        *buf = 0;
        return -1;
    }

    buf[retval] = 0;
    return 0;
}

int
NestTruncUnlink( char *file )
{

    /*
        first O_TRUNC the file and then unlink it

        the stupid quota system doesn't update the usage statistics,
        sometimes, weirdly, if you just do a unlink()

        this was actually found out accidentally :)
        and this O_TRUNC idea seems to solve problems!
    */

    int tmp_fd = NestOpen( file, O_WRONLY | O_TRUNC );

    if (tmp_fd >= 0) {
        NestClose(tmp_fd);
    }

    int retval = unlink(file);

    if (retval < 0) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "unlink %s: %s\n",
                    file,
                    strerror( errno ));
    }

    return retval;
}

int
NestUnlink( const char *file )
{

    int retval = unlink(file);

    if (retval < 0) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "unlink %s: %s\n",
                    file,
                    strerror( errno ));
    }

    return retval;
}

int
NestSync( )
{
    // Some sync()'s return a value, some void.
    // We'll just ignore them all, and assume success.
    // Ugly, yes, but it works.  -NRL
    sync();
    return 0;
}

int
NestGetFileflags( int transfer_type )
{
    if ( transfer_type == CLIENT_RECV_FILE ) {
        return READ_FLAGS;
    } else {
        assert( transfer_type == CLIENT_SEND_FILE );
        return WRITE_FLAGS;
    }
}

int 
NestSetenv( const char *name, const char *value, int overwrite )
{
#ifdef HAS_SETENV
    LOCK_MUTEX( &nest_common_mutex );
    int ret = setenv( name, value, overwrite );
    UNLOCK_MUTEX( &nest_common_mutex );
    return ret;
#else
    char command[NEST_MESSAGESIZE];
    NestSnprintf( command, NEST_MESSAGESIZE, "export %s=%s\n",
            name, value );
    if ( overwrite ) {
        nest_warn( __FILE__, __FUNCTION__, "Overwrite ignored in setenv\n" );
    }
    nest_warn( __FILE__, __FUNCTION__, "Not setting env %s -> %s", 
            name, value );
    return 1;
#endif
}


void 
NestSignal( int signum, void (*handler)(int))
{
    bool ok = true;
    LOCK_MUTEX( &nest_common_mutex );
    if ( signal( signum, handler ) == SIG_ERR ) {
        ok = false;
    }
    UNLOCK_MUTEX( &nest_common_mutex );
    if ( ! ok ) {
        nest_perror( "signal" );
        exit( ERROR_EXIT );
    }
}

void
NestNanosleep( unsigned long nanos )
{
#ifdef HAS_NANOSLEEP
    struct timespec w,r;
    int result;

    w.tv_sec = 0;
    w.tv_nsec = nanos;

    while(w.tv_nsec>BILLION) {
        w.tv_sec++;
        w.tv_nsec-=BILLION;
    }

    while(1) {
        result = nanosleep(&w,&r);
        if(result<0) {
            w = r;
        } else {
            break;
        }
    }
#else
    nest_warn( __FILE__, __FUNCTION__,
			   "This platform does not have nanosleep.\n" );
#endif
}

// fdset utilities
int 
NestFd_print( int debug_level, fd_set *fds, int maxfd )
{
    return NestFd_print( debug_level, fds, maxfd, false );
}

int
NestFd_print( int debug_level, fd_set *fds, int maxfd, bool use_fstat )
{

    struct stat sbuf;

    debug_level |= NEST_DEBUG_FD;

    for( int i = 0; i < maxfd; i++ ) {
        if ( NestFd_isset( i, fds ) ) {
            nest_debug( debug_level, "%d ", i );
            if ( use_fstat ) {
                int ret = fstat( i, &sbuf );
                if ( ret < 0 ) {
                    nest_warn( __FILE__, __FUNCTION__,
							   "fstat of %d failed: %s\n", i, strerror(errno));
                }
            }
        }
    }
    nest_debug( debug_level, "\n" );
    return 0;
}

void
NestFd_set( int fd, fd_set *fds )
{
    //CHECK_PID;
    nest_debug( NEST_DEBUG_FD, "Setting %d in fdset 0x%x\n", fd, fds );
    FD_SET( fd, fds );
}

void
NestFd_clear( int fd, fd_set *fds )
{
    CHECK_PID;
    nest_debug( NEST_DEBUG_FD, "Clearing %d in fdset 0x%x\n", fd, fds );
    FD_CLR( fd, fds );
}

int
NestFd_isset( int fd, fd_set *fds )
{
    return( FD_ISSET( fd, fds ) );
}

double
NestGetLotDuration(char *duration)
{
    char   *time;
    double days;
    double hours;
    double minutes;

    time    = NestStrdup(duration);
    days    = atoi( strtok(duration,":") );
    hours   = atoi( strtok(NULL,":") );
    minutes = atoi( strtok(NULL,":") );

    return days*24*60 + hours*60 + minutes;

}

nestLotSize  
NestStrLotList_to_LotSize(const char *lot_str)
{
    char    *str_lot_list = strdup(lot_str);
    nestLotSize  lot_allocation = 0;
	nestLotSize  temp_allocation;
    char    *token = strtok(str_lot_list, ":");

    
    while(token) {
        token = strtok(NULL, ":");
        if(token) {
            sscanf(token,"%lld", &temp_allocation);
            if(temp_allocation < 0) {
                lot_allocation = 0;
                break;
            }
            lot_allocation += temp_allocation;
        }
    }

    free( str_lot_list );
    return lot_allocation;
}

int
NestFileInfo_to_string(const NestFileInfo *file,
                       char *msg,
                       int len)
{

    if (!msg || !len) {
        return -1;
    }

    if (!file) {
        *msg = 0;
        return 0;
    }

    return NestSnprintf(msg,
						len,
						"%s %ld:%d:%d:%d:%d %s",
						file->name,
						file->size,
						file->type,
						file->atime,
						file->ctime,
						file->mtime,
						file->LotList);
}

NestReplyStatus
NestAdd_ACL( NestACL **result,
             const char *name,
             int is_user,
             nest_access_rights_t right )
{

    if (!result || !name) {
        return NEST_BAD_PARAMETERS;
    }
    NestACL *cur;

    for( cur = *result; cur; cur = cur->next) {
        if (!strcmp(name, cur->acl.name) &&
            (is_user == cur->acl.is_user)) {
            cur->acl.rights |= right;
            return NEST_SUCCESS;
        }
    }

    cur = (NestACL *) NestMalloc (sizeof(NestACL));
    if (!cur) {
        return NEST_OUT_OF_RESOURCES;
    }

    strcpy(cur->acl.name, name);
    cur->acl.is_user = is_user;
    cur->acl.rights = right;

    cur->next = (*result);
    *result = cur;

    return NEST_SUCCESS;
}

/*
    Helper function to add an entry (user) to a list (members)
*/

NestReplyStatus
NestAdd_entry_to_list(string &new_members,
                      string members,
                      string user)
{

    unsigned int i;
    char flag = 's';         //  s == start

    new_members = members;
    if (new_members.find(user)
        != string::npos) {
        return NEST_ALREADY_EXISTS;
    }

    for(i = 0; i < new_members.length(); i++) {
        if (new_members[i] == '{') {
            flag = 'o';  //  o == open braces
            continue;
        }
        if (new_members[i] != ' ' &&
            new_members[i] != '\t' &&
            new_members[i] != '\n' &&
            new_members[i] != '}') {
            if (flag == 'f') {
                continue;
            }
            if (flag == 'o') {
                flag = 'f';  //  f == found an alphabet after '{'
            }
            continue;
        }
        if (new_members[i] == '}') {
            break;
        }
    }

    string append_str;

    switch (flag) {
        case 's':   nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                        "Improperly formed list!\n" );
                    return NEST_UNKNOWN_ERROR;
        case 'o':   /*  First member */
                    append_str = "\"" + user + "\"}";
                    break;
        case 'f':   /* Not the first member */
                    append_str = ",\"" + user + "\"}";
                    break;
    };

    new_members.replace(i, 1, append_str);
    return NEST_SUCCESS;
}

/*
    Helper function to delete an entry (user) from a list (members)
*/

NestReplyStatus
NestDelete_entry_from_list(string &new_members,
                           string members,
                           string user)
{
    string tmp_user;
    unsigned int pos;

    new_members = members;
    if (new_members.find(user)
        == string::npos) {
        return NEST_USER_NOT_FOUND;
    }
    tmp_user = ",\"" + user + "\"";
    pos = new_members.find(tmp_user);
    if (pos == string::npos) {
        tmp_user = "\"" + user + "\",";
        pos = new_members.find(tmp_user);
        if (pos == string::npos) {
            tmp_user = "\"" + user + "\"";
            pos = new_members.find(tmp_user);
            if (pos == string::npos) {
                return NEST_UNKNOWN_ERROR;
            }
        }
    }

    new_members.erase(pos, tmp_user.length());
    return NEST_SUCCESS;
}

/*
Check to see if this path corresponds to a device.
If so, fill in that device name and return true.
This funciton was included within #ifdef HAS_FSTAB_H.
This is okay, but sometimes to make the system more robust
the volume label instead of device name is listed in
/etc/fstab. Also /etc/fstab is a static snap shot. So
we are better off using /etc/mtab.
*/
#if 0
int NestPathIsDevice( const char *path, char *device )
{
    struct fstab *f;

    setfsent();
    f = getfsfile(path);
    endfsent();

    if(f) {
        strncpy(device, f->fs_spec, NEST_MAXPATHLEN);
        return 1;
    } else {
        return 0;
    }
}
#endif

int 
NestPathIsDevice( const char *path, char *device )
{
    FILE *file;
    char mpath[NEST_MAXPATHLEN];
    char line[NEST_MAXPATHLEN*2];
    int foundit = 0;

    file = fopen("/etc/mtab","r");
    if(!file) return 0;

    while(fgets(line,sizeof(line),file)) {
        if(sscanf(line,"%s %s",device,mpath)==2) {
            if(!strcmp(path,mpath)) {
                foundit = 1;
                break;
            }
        }
    }

    fclose(file);

    return foundit;
}

/*
Given a path name, try to determine the device containing it.
*/

int 
NestPathToDevice( const char *path, char *device )
{
    char cwd[NEST_MAXPATHLEN];

    strcpy(cwd,path);

    while (cwd[0]) {
        int i;

        if(NestPathIsDevice(cwd,device)) break;

        for(i=NestStrlen(cwd) - 1; i >= 0; i--) {
            if (cwd[i] == '/') {
                cwd[i] = 0;
                break;
            }
        }
    }

    if(cwd[0]) {
        return 1;
    } else {
        return NestPathIsDevice("/",device);
    }
}

/*
Return the pathname where this device is mounted.
This funciton was included within #ifdef HAS_FSTAB_H.
This is okay, but sometimes to make the system more robust
the volume label instead of device name is listed in
/etc/fstab. Also /etc/fstab is a static snap shot. So
we are better off using /etc/mtab.
*/
#if 0
int NestDeviceToPath( const char *device, char *path )
{
    struct fstab *f;

    setfsent();
    f = getfsspec(device);
    endfsent();

    if(f) {
        strncpy(path, f->fs_file, NEST_MAXPATHLEN);
        return 1;
    } else {
        return 0;
    }
}
#endif

int 
NestDeviceToPath( const char *device, char *path ) {
    FILE *file;
    char tempdev[NEST_MAXPATHLEN];
    char line[NEST_MAXPATHLEN*2];
    int foundit = 0;

    file = fopen("/etc/mtab","r");
    if(!file) return 0;

    while(fgets(line,sizeof(line),file)) {
        if(sscanf(line,"%s %s",tempdev,path)==2) {
            if(!strcmp(device,tempdev)) {
                foundit = 1;
                break;
            }
        }
    }

    fclose(file);

    return foundit;
}

/*
Return the filesystem type of this device.
This funciton was included within #ifdef HAS_FSTAB_H.
This is okay, but sometimes to make the system more robust
the volume label instead of device name is listed in
/etc/fstab. Also /etc/fstab is a static snap shot. So
we are better off using /etc/mtab.
*/
#if 0
int NestFSType( const char *device, char *type )
{
    struct fstab *f;

    setfsent();
    f = getfsspec(device);
    endfsent();

    if(f) {
        strncpy(type, f->fs_vfstype, NEST_MAXPATHLEN);
        return 1;
    } else {
        return 0;
    }
}
#endif

int 
NestFSType( const char *device, char *type )
{
    FILE *file;
    char tempdev[NEST_MAXPATHLEN];
    char path[NEST_MAXPATHLEN];
    char line[NEST_MAXPATHLEN*2];
    int foundit = 0;

    file = fopen("/etc/mtab","r");
    if(!file) return 0;

    while(fgets(line,sizeof(line),file)) {
        if(sscanf(line,"%s %s %s",tempdev,path,type)==3) {
            if(!strcmp(device,tempdev)) {
                foundit = 1;
                break;
            }
        }
    }

    fclose(file);

    return foundit;
}

struct timeval *
NestNextTimeout( int seconds, struct timeval *last )
{

    static struct timeval now;
    static struct timeval timeout;
    long micros;

    NestGettimeofday( &now );

        // how long since the last one and how long til the next one
    micros = NestTimediff( last, &now, __FUNCTION__ );
    if ( micros < 0 ) {
        nest_warn( __FILE__, __FUNCTION__, "%s.%s: negative micros: %ld (will set to zero).\n",
                __FILE__,
                __FUNCTION__,
                micros );
        micros = 0;
    }

    micros = ( seconds * 1000 ) - micros;
    nest_debug( NEST_DEBUG_SERVER, "%ld millis until next maintenance\n", 
            ( micros / THOUSAND ) );

    if ( micros <= 0 ) {
            // overdue already
        timeout.tv_sec = 0;
        timeout.tv_usec = 0;
    } else {
            // compute remaining time by subtracting elapsed from max
        timeout.tv_sec = micros / ( THOUSAND * THOUSAND );
        timeout.tv_usec = micros % ( THOUSAND * THOUSAND );
    }

    return &timeout; 
}

const char *
NestUname( )
{
    static struct utsname *buf = NULL;
    if ( ! buf ) { //initialize
        buf = (struct utsname*)NestMalloc(sizeof(*buf));
        uname(buf);
    }
    return buf->nodename; 
}

const char *
NestGetModifiedIdentity( const char *identity )
{
    /* 
     * This function creates a physical file name from the
     * global identity which will be used for creating 
     * physical file names and as key in the classad 
     * collections. To make sure special chars like ' ','/',
     * '=','"',etc. in the identity are not mis-interpreted,
     * they are all converted to '_'. Change the following code
     * suitably to include more special chars.
     */

    if(!identity) {
        return NULL;
    }

    char *file_entry = (char*) NestMalloc( strlen(identity)+1 );
    int  i = 0;

    const char *cursor = identity;
    while(*cursor) {
        switch(*cursor) {
            case '"':
                cursor++;
                continue;                 
            case ' ':
            case '/':
            case '=':
                *(file_entry+i) = '_';
                break;
            default:
                *(file_entry+i) = *cursor;
                break;
        }
        cursor++;
        i++;
    }
    *(file_entry+i) = 0;

    return file_entry; 
}

// Match this identity to the a mapfile entry & return the matching user
const char *
NestGetUserFromIdentity( const char *identity )
{
    if(!strcmp(identity,NEST_EVERYBODY) ||
       !strcmp(identity,NEST_NOBODY) ||
       !strcmp(identity,NEST_NO_GROUP))
        return (char*)identity;
            
    char *user = NULL;
    int  state = Nestset_grid_mapfile( GridMapFile );
    
    if ( state < 0) {
        nest_warn( __FILE__, __FUNCTION__,  
                    "Couldn't read grid-mapfile at %s: Can't lookup %s!\n",
                    GridMapFile,
                    identity );
        return user;
    }

    grid_mapfile_entry  *entry;
    while( (entry = Nestget_grid_mapfile_entry(state)) != NULL) {
        if( !strcmp(identity, entry->identity) ) {
            user = entry->name;
            break;
        }
    }

    if ( !user ) {
        nest_warn( __FILE__, __FUNCTION__,
				   "Unable to find a gridmap entry for identity '%s'.\n",
				   identity );
    }
    nest_debug( NEST_DEBUG_CLIENT | NEST_DEBUG_AUTH,
				"%s returning %s from identity '%s'.\n",
				__FUNCTION__, user, identity );
    Nestend_grid_mapfile( state );
    return user;
}

// Match this user to the the first matching mapfile entry & return the
// matching identity
const char *
NestGetIdentityFromUser( const char *user )
{
    char *identity = NULL;
    int  state = Nestset_grid_mapfile( GridMapFile );
    
    if ( state < 0) {
        nest_warn( __FILE__, __FUNCTION__,  
                    "Couldn't read grid-mapfile at %s: Can't lookup %s!\n",
                    GridMapFile,
                    identity );
        return user;
    }

    grid_mapfile_entry  *entry;
    while( (entry = Nestget_grid_mapfile_entry(state) ) != NULL) {
        if( !strcmp(user, entry->name) ) {
            identity = entry->name;
            break;
        }
    }

    if ( !identity ) {
        nest_warn( __FILE__, __FUNCTION__,
				   "Unable to find a gridmap entry for user '%s'.\n",
				   user );
    }
    nest_debug( NEST_DEBUG_CLIENT | NEST_DEBUG_AUTH,
				"%s: returning %s from user '%s'.\n",
				__FUNCTION__, identity, user );
    Nestend_grid_mapfile( state );
    return identity;
}

// Match this identity to the a mapfile entry & return the matching user
bool
NestGetMapfileEntry( const char *name, grid_mapfile_entry *match )
{
    if(!strcmp(name,NEST_EVERYBODY) ||
       !strcmp(name,NEST_NOBODY) ||
       !strcmp(name,NEST_NO_GROUP))
        return (char*)name;
            
    int  state = Nestset_grid_mapfile( GridMapFile );
    
    if ( state < 0) {
        nest_warn( __FILE__, __FUNCTION__,  
                    "Couldn't read grid-mapfile at %s: Can't lookup %s!\n",
                    GridMapFile,
                    name );
        return false;
    }

    grid_mapfile_entry  *entry;
	bool		matched = false;
    while( (entry = Nestget_grid_mapfile_entry(state)) != NULL) {
        if( !strcmp(name, entry->identity) ) {
			memcpy( match, entry, sizeof( grid_mapfile_entry ) );
			matched = true;
            break;
        } else if( !strcmp(name, entry->name) ) {
			memcpy( match, entry, sizeof( grid_mapfile_entry ) );
			matched = true;
            break;
		}
    }
    Nestend_grid_mapfile( state );

    if ( !matched ) {
        nest_warn( __FILE__, __FUNCTION__,
				   "Unable to find a gridmap entry for name '%s'.\n",
				   name );
		return false;
    } else {
		nest_debug( NEST_DEBUG_CLIENT | NEST_DEBUG_AUTH,
					"%s: match found for name '%s'.\n",
					__FUNCTION__, name );
		return true;
	}
}

NestReplyStatus
NestCreateFullPath( const char *path )
{
	char *temp;
	char *delim;
	char oldchar;
	int result;

	temp = (char *) alloca(strlen(path)+1);
	strcpy(temp,path);

	delim = temp;

	while( (delim = strchr(delim,'/')) ) {

		if(delim==temp) {
			delim++;
			continue;
		}

		oldchar = *delim;
		*delim = 0;

		result = NestMkdir(temp,0700);
		if(result!=0) {
			if(errno==EEXIST) {
				/* no problem, keep going */
			} else {
				return ErrnoToNest(errno);
			}
		} else {
			/* ok, made it successfully */
		}

		*delim = oldchar;
		delim++;
	}

	/* Now, last chance */

	result = NestMkdir(temp,0700);
	if(result!=0) {
		if(errno==EEXIST) {
			return NEST_SUCCESS;
		} else {
			return ErrnoToNest(errno);
		}
	} else {
		return NEST_SUCCESS;
	}
}

void NestCollapsePath( const char *l, char *s )
{
	char *start = s;

	while(*l) {
		if(!strncmp(l,"//",2)) {
			l++;
		} else if(!strncmp(l,"/./",3)) {
			l+=2;
		} else if(!strncmp(l,"/..",3)) {
			l+=3;
			while( ((*s)!='/') && (s>start) ) {
				s--;
			}
		} else {
			*s++ = *l++;
		}
	}

	*s = 0;

	if(s==start) {
		strcpy(s,"/");
	}
}
