/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

extern "C" {
#include "globus_gridftp_server.h"
}

#include "nest_dsi_recv.h"
#include "nest_dsi_finish.h"
#include "nest_dsi_logger.h"

nestDsiRecv::nestDsiRecv( nestDsiSession				*parent_session,
						  globus_gfs_transfer_info_t	*transfer_info ) :
		nestDsiTransfer( parent_session, transfer_info )
{
	// For now, just log the event
	_logger->logEvent( NEST_DSI_LOG_XFER,
					   "RECV '%s'\n", info.pathname );
}

nestDsiRecv::~nestDsiRecv( void )
{

}

int
nestDsiRecv::setup( globus_gfs_operation_t op,
				 globus_gridftp_server_read_cb_t callback )
{
	globus_result_t	result;

	// Register the read
	result = globus_gridftp_server_register_read(
		op,
		(globus_byte_t *) buffer,
		buffer_size,
		callback,
		this );

	// How'd it go?
	if( result != GLOBUS_SUCCESS ) {
		_logger->logEvent( NEST_DSI_LOG_ALWAYS,
						   "setupRecv: reg read failed\n" );
		cancel( op );
		return -1;
	}

	// Ok, done
	return 0;
}

int
nestDsiRecv::sendToNest(
	globus_gfs_operation_t	op,
    globus_result_t			result,
    globus_byte_t			*buffer,
    globus_size_t			nbytes,
    globus_off_t			offset,
    globus_bool_t			eof )
{
	NestReplyStatus status;

	// Send the data to NeST
	if( nbytes ) {
		_logger->logEvent( NEST_DSI_LOG_XFER_FULL,
						   "sendToNest: Sending %d bytes @ %ld\n",
						   nbytes, offset );
		status = NestWriteBytes (
			session->getConnection(),
			info.pathname,
			offset,
			(char *) buffer,
			nbytes,
			true );
		if ( NEST_SUCCESS != status  ) {
			_logger->logEvent( NEST_DSI_LOG_ALWAYS,
							   "sendToNest: Failed: %d\n", status );
			transfer_status = status;
			return -1;
		}
	}

	return 0;
}
