/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

extern "C" {
#include "globus_gridftp_server.h"
}

#include "nest_dsi_send.h"
#include "nest_dsi_finish.h"
#include "nest_dsi_logger.h"

nestDsiSend::nestDsiSend( nestDsiSession				*parent_session,
						  globus_gfs_transfer_info_t	*transfer_info ) :
		nestDsiTransfer( parent_session, transfer_info )
{
	offset = 0;
	bytesInBuffer = 0;

	// Log the event
	_logger->logEvent( NEST_DSI_LOG_XFER,
					   "SEND '%s'\n", info.pathname );
}

nestDsiSend::~nestDsiSend( void )
{

}

int
nestDsiSend::setup( globus_gfs_operation_t op,
					globus_gridftp_server_write_cb_t callback )
{
	globus_result_t	result;

	// Register the read
	_logger->logEvent( NEST_DSI_LOG_XFER,
					   "setupSend: Registering write: count=%d, off=%ld\n",
					   bytesInBuffer, offset);
	result = globus_gridftp_server_register_write(
		op,
		(globus_byte_t *) buffer,
		bytesInBuffer,
		offset,
		0,
		callback,
		this );

	// How'd it go?
	if( result != GLOBUS_SUCCESS ) {
		_logger->logEvent( NEST_DSI_LOG_ALWAYS,
						   "setupSend: reg write failed\n" );
		cancel( op );
		return -1;
	}

	// Ok, done
	return 0;
}

int
nestDsiSend::writeCB( globus_gfs_operation_t	op,
					  globus_result_t			result,
					  globus_byte_t				*buffer,
					  globus_size_t				nbytes )
{
	if ( nbytes > (unsigned) bytesInBuffer )
	{
		_logger->logEvent( NEST_DSI_LOG_ALWAYS,
						   "writeCB: Too many bytes (%d > %d)\n",
						   nbytes, bytesInBuffer );
		return -1;
	}
	bytesInBuffer -= nbytes;
	offset += nbytes;
	_logger->logEvent( NEST_DSI_LOG_XFER_FULL,
					   "writeCB: bib=%d, offset=%ld\n",
					   bytesInBuffer, offset );
	return bytesInBuffer;
}

int
nestDsiSend::readFromNest( globus_size_t &nbytes )
{
	NestReplyStatus status;

	// Just wait if the buffer isn't empty...
	if ( bytesInBuffer ) {
		_logger->logEvent( NEST_DSI_LOG_XFER,
						   "readFromNest: Already %d bytes in buffer\n",
						   bytesInBuffer );
		return 1;
	}

	// Read the data to NeST
	_logger->logEvent( NEST_DSI_LOG_XFER_FULL,
					   "readFromNest: Reading %d bytes from offset %ld\n",
					   buffer_size, offset );
	int bytes_read = 0;
	status = NestReadBytes (
		session->getConnection(),
		info.pathname,
		offset,
		(char *) buffer,
		buffer_size,
		bytes_read );
	if ( NEST_SUCCESS != status  ) {
		_logger->logEvent( NEST_DSI_LOG_ALWAYS,
						   "Error reading from NeST: %d\n", status );
		transfer_status = status;
		return -1;
	}

	bytesInBuffer = nbytes = bytes_read;
	_logger->logEvent( NEST_DSI_LOG_XFER_FULL,
					   "readFromNest: Now %d bytes in buffer\n",
					   bytesInBuffer );

	return 0;
}
