/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "nest_user.h"
#include "nest_identity_manager.h"
#include "nest_protocol.h"


nestIdentityManager::nestIdentityManager (
	void )
{
    this->Identities = NULL;
    this->userlist = "";
    this->denylist = "";

    /*
        default case used by this->initialize is that
        nobody was present in either the userlist or denylist, previously!
    */
    this->prev_userlist = "";
    this->prev_denylist = "";

    return;
}

nestIdentityManager::~nestIdentityManager( void )
{
	/* Does nothing for now */
}

NestReplyStatus
nestIdentityManager::set_userlist(
	const char *userlist )
{
    if (!userlist) {
        this->userlist = NEST_EVERYBODY;
    } else {
        this->userlist = userlist;
    }
	return NEST_SUCCESS;
}

NestReplyStatus
nestIdentityManager::set_denylist(
	const char *denylist )
{
    if (!denylist) {
        this->denylist = "";
    } else {
        this->denylist = denylist;
    }
	return NEST_SUCCESS;
}

NestReplyStatus
nestIdentityManager::get_total_nest_users( int *total_nest_users )
{
    int temp_total_nest_users;
    string key = TOTAL_USERS_KEY;

    ClassAd *total_users_ad = this->Identities->GetClassAd( key );
    if(total_users_ad){
        bool result = total_users_ad->EvaluateAttrNumber( "TotalNestUsers",
                                                     temp_total_nest_users );
        if(!result){
            *total_nest_users = -1;
            return NEST_CLASSAD_ERROR;
        } else {
            *total_nest_users = temp_total_nest_users;
            return NEST_SUCCESS;
        }

    } else {
        *total_nest_users = -1;
        return NEST_CLASSAD_ERROR;
    }

}

NestReplyStatus
nestIdentityManager::get_total_nest_groups( int *total_nest_groups )
{
    int temp_total_nest_groups;
    string key = TOTAL_GROUPS_KEY;

    ClassAd *total_groups_ad = this->Identities->GetClassAd( key );
    if(total_groups_ad){
        bool result = total_groups_ad->EvaluateAttrNumber( "TotalNestGroups",
                                                     temp_total_nest_groups );
        if(!result){
            *total_nest_groups = -1;
            return NEST_CLASSAD_ERROR;
        } else {
            *total_nest_groups = temp_total_nest_groups;
            return NEST_SUCCESS;
        }

    } else {
        *total_nest_groups = -1;
        return NEST_CLASSAD_ERROR;
    }

}

bool
nestIdentityManager::is_present_in_list(const char *userpool,
                                        const nestUser *user)
{

    if (!userpool ||
        !user) {
        return false;
    }

	const char	*ident = user->getIdentity( );
	int 	    len = NestStrlen(ident);
	int			userpool_len = NestStrlen(userpool);

    char    tmp_user[NEST_NAMELEN];
	char	*tmp_ptr;

    if (len > userpool_len) {
        return false;
    }

    if (!strcmp(ident, userpool)) {
        return true;
    }

    NestSnprintf(tmp_user, NEST_NAMELEN, ",%s,", ident);
    if (strstr(userpool, tmp_user)) {
        return true;
    }

    NestSnprintf(tmp_user, NEST_NAMELEN, "%s,", ident);
    if ((tmp_ptr = strstr(userpool, tmp_user))) {
        if (tmp_ptr == userpool) {
            return true;
        }
    }

    NestSnprintf(tmp_user, NEST_NAMELEN, ",%s", ident);
    if ((tmp_ptr = strstr(userpool, tmp_user))) {
        if ((tmp_ptr + len + 1) == (userpool + userpool_len)) {
            return true;
        }
    }

    return false;
}

bool
nestIdentityManager::is_user_access_allowed (const nestUser *user)
{

    if (! user) {
        return false;
    }

	return user->isAdmin( );
}

bool
nestIdentityManager::is_group_access_allowed (const nestGroup *group,
                                              const nestUser *user,
                                              bool allow_members)
{

    if (! group || ! user) {
        return false;
    }

	if ( user->isAdmin( ) ) {
        return true;
    }

	if ( group->isNoGroup( ) ) {
        return false;
    }

    string key = GROUP_KEY_PREFIX;
    key = key + group->getName( );

    ClassAd *group_ad = this->Identities->GetClassAd( key );
    if (!group_ad) {
        return false;
    }

    string owner_user;
    if (!group_ad->EvaluateAttrString( "OwnerUser", owner_user )) {
        return false;
    }

    if(!allow_members) {
        return (strcmp(owner_user.c_str(), user->getIdentity() ) == 0);
    } else {
        return is_user_in_group(user, group );
	}
}

/*
    Refer to file "doc/Userlist.Denylist.Semantics" to look at
    how the following boolean function was derived:

    Result = ((~A (((C + E)(B + ~D)) + CE )) + (ABE ( ~C~D + CD )))

    where   A is "Present in Current-Denylist",
            B is "Present in Previous-Denylist",
            C is "Present in Current-Userlist",
            D is "Present in Previous-Userlist" and
            E is "Valid-Current-User".

    Looks awfully weird, I agree, but really helps to give a pre-defined
    and meaningful semantics with providing both user and deny lists, and
    the ability to add/delete users on the fly! Trust me!!

*/
bool
nestIdentityManager::is_valid_nest_user(const nestUser *user, bool is_init)
{
    /* Refer the /doc/Userlist.Denylist.Semantics file for the valid user
     * semantics. If the current/prev userlist is null then it is assumed to
     * contain all users from the grid-map-file. If the current/prev denylist
     * is null then no users are in that list. One exception is on the very
     * first synchronize call during initialization, if previous userlist is
     * null, then no users are in that list
     * This interpretation is necessary due to the fact that users can be added
     * on fly.
     */
	if ( ! user ) {
		return false;
	}

    bool deny = this->is_present_in_list(this->denylist.c_str(), user);
	bool ulist;
	bool valid = this->is_user_present( user );

    if(!strcmp(this->userlist.c_str(), ""))
        ulist = true;
    else
        ulist = this->is_present_in_list(this->userlist.c_str(), user);

	bool	ok = false;

	if ( deny ) {
		// Do nothing
	} else if ( ulist || valid ) {
		ok = true;
	} else {
		// Do nothing
	}

# if 0
	nest_debug( NEST_DEBUG_AUTH,
				"%s: init=%s deny=%s ulist=%s valid=%s ->%s\n",
				user, bs(is_init), bs(deny), bs(ulist), bs(valid), bs(ok) );
# endif

	return ok;
}

NestReplyStatus
nestIdentityManager::get_actual_nest_group(char *actual_nest_group) {

    if (!actual_nest_group) {
        return NEST_BAD_PARAMETERS;
    }

    group_ent ge;

    if (Nestgetgrpnam( ActualGroup, &ge, SystemGrpFile) < 0) {
        strcpy(actual_nest_group, "root");
        return NEST_SUCCESS;
    }

    strcpy(actual_nest_group, ge.grp_name);
    return NEST_SUCCESS;
}

bool
nestIdentityManager::is_user_present(const nestUser *user)
{

    if (!user) {
        return false;
    }

    string key = USER_KEY_PREFIX;
    key = key + user->getIdentity( );

    ClassAd *user_ad = this->Identities->GetClassAd( key );

    if (!user_ad) {
        return false;
    }

    bool    result,
            valid;

    result = user_ad->EvaluateAttrBool( "ValidEntry", valid );
    if (!result) {
        return false;
    }

    return valid;
}

bool
nestIdentityManager::is_group_present(const nestGroup *group)
{

    if (!group) {
        return false;
    }

	if ( group->isNoGroup() ) {
        return true;
    }

    string key = GROUP_KEY_PREFIX;
    key = key + group->getName( );

    return (this->Identities->GetClassAd(key) != NULL);

}


bool
nestIdentityManager::is_user_in_group(const nestUser *user,
                                      const nestGroup *group)
{

    string  key = GROUP_KEY_PREFIX;
    key = key + group->getName();

    ClassAd *group_class_ad = this->Identities->GetClassAd(key);
    if (!group_class_ad) {
        return false;
    }

    string  attr("Members");
    ExprTree *group_list = group_class_ad->Lookup( attr );

    if (!group_list) {
        return false;
    }

    ExprListIterator list_itr( (ExprList *) group_list);

    Value   val;

    if (!list_itr.CurrentValue( val )) {
        return false;
    }

    string  member;

    do {
        if (val.IsStringValue( member )) {
            if ( user->match( &member ) ) {
                return true;
            }
        }
    } while( list_itr.NextValue( val ) );

    return false;
}


bool
nestIdentityManager::is_group_in_user(const nestUser *user,
                                      const nestGroup *group)
{

    string  key = USER_KEY_PREFIX;
    key = key + user->getName();

    ClassAd *user_class_ad = this->Identities->GetClassAd(key);
    if (!user_class_ad) {
        return false;
    }

    string  attr("Groups");
    ExprTree *group_list = user_class_ad->Lookup( attr );

    if (!group_list) {
        return false;
    }

    ExprListIterator list_itr( (ExprList *) group_list);

    Value   val;

    if (!list_itr.CurrentValue( val )) {
        return false;
    }

    string  grp_entry;

    do {
        if (val.IsStringValue( grp_entry )) {
            if ( group->match( &grp_entry) ) {
                return true;
            }
        }
    } while( list_itr.NextValue( val ) );

    return false;
}


NestReplyStatus
nestIdentityManager::add_user_to_group_members(const nestUser *user,
                                               const nestGroup *group)
{

    if ( group->isNoGroup() ) {
        return NEST_SUCCESS;
    }

    string key = GROUP_KEY_PREFIX;
    key = key + group->getName();

    ClassAd *group_class_ad = this->Identities->GetClassAd(key);

    if (!group_class_ad) {
        return NEST_GROUP_NOT_FOUND;
    }

    string  attr("Members");
    ExprTree *member_list = group_class_ad->Lookup( attr );

    if (!member_list) {
        return NEST_CLASSAD_ERROR;
    }

    string	members;
    string	new_members;
	string	str_user(user->getName() );

    unparser.Unparse( members, member_list );
    //  delete member_list;

    NestReplyStatus reply = NestAdd_entry_to_list( new_members,
                                                   members,
                                                   str_user );
    NEST_REPLY_CHECK(reply);

    string modify_classad = "";
    modify_classad = modify_classad +
                        " [ " + attr + " = " + new_members + " ] ";
    modify_classad = " [ Updates = " + modify_classad + " ] ";

    ClassAd *mod_classad = parser.ParseClassAd( modify_classad, true );

    if (!this->Identities->ModifyClassAd( key, mod_classad )) {
        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}


NestReplyStatus
nestIdentityManager::add_group_to_user_groups(const nestUser *user,
                                              const nestGroup *group)
{

    string key = USER_KEY_PREFIX;
    key = key + user->getName();

    ClassAd *user_class_ad = this->Identities->GetClassAd(key);

    if (!user_class_ad) {
        return NEST_USER_NOT_FOUND;
    }

    string attr("Groups");
    ExprTree *group_list = user_class_ad->Lookup( attr );

    if (!group_list) {
        return NEST_CLASSAD_ERROR;
    }

    string groups;
    string new_groups;
    string str_group( group->getName() );
    unparser.Unparse( groups, group_list );
    //  delete group_list;

    NestReplyStatus reply = NestAdd_entry_to_list( new_groups,
												   groups,
												   str_group );
    NEST_REPLY_CHECK(reply);

    string modify_classad = "";
    modify_classad = modify_classad +
                        " [ " + attr + " = " + new_groups + " ] ";
    modify_classad = " [ Updates = " + modify_classad + " ] ";

    ClassAd *mod_classad = parser.ParseClassAd( modify_classad, true );

    if (!this->Identities->ModifyClassAd( key, mod_classad )) {
        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}

NestReplyStatus
nestIdentityManager::delete_user_from_group_members(const nestUser *user,
                                                    const nestGroup *group)
{
    if ( group->isNoGroup() ) {
        return NEST_SUCCESS;
    }

    string key = GROUP_KEY_PREFIX;
    key = key + group->getName();

    ClassAd *group_class_ad = this->Identities->GetClassAd(key);

    if (!group_class_ad) {
        return NEST_GROUP_NOT_FOUND;
    }

    string  attr("Members");
    ExprTree *group_list = group_class_ad->Lookup( attr );

    if (!group_list) {
        return NEST_CLASSAD_ERROR;
    }

    string members,
           new_members,
           str_user(user->getName() );
    unparser.Unparse( members, group_list );
    //  delete group_list;

    NestReplyStatus reply = NestDelete_entry_from_list( new_members,
                                                    members,
                                                    str_user );
    NEST_REPLY_CHECK(reply);

    string modify_classad = "";
    modify_classad = modify_classad +
                        " [ " + attr + " = " + new_members + " ] ";
    modify_classad = " [ Updates = " + modify_classad + " ] ";

    ClassAd *mod_classad = parser.ParseClassAd( modify_classad, true );

    if (!this->Identities->ModifyClassAd( key, mod_classad )) {
        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}


NestReplyStatus
nestIdentityManager::delete_group_from_user_groups(const nestUser *user,
												   const nestGroup *group)
{

    string  key = USER_KEY_PREFIX;
    key = key + user->getName();

    ClassAd *user_class_ad = this->Identities->GetClassAd(key);

    if (!user_class_ad) {
        return NEST_USER_NOT_FOUND;
    }

    string attr("Groups");
    ExprTree *group_list = user_class_ad->Lookup( attr );

    if (!group_list) {
        return NEST_CLASSAD_ERROR;
    }

    string groups;
    string new_groups;
    string str_group(group->getName() );
    unparser.Unparse( groups, group_list );
    //  delete group_list;

    NestReplyStatus reply = NestDelete_entry_from_list( new_groups,
                                                    groups,
                                                    str_group );
    NEST_REPLY_CHECK(reply);

    string modify_classad = "";
    modify_classad = modify_classad +
                        " [ " + attr + " = " + new_groups + " ] ";
    modify_classad = " [ Updates = " + modify_classad + " ] ";

    ClassAd *mod_classad = parser.ParseClassAd( modify_classad, true );

    if (!this->Identities->ModifyClassAd( key, mod_classad )) {
        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}


bool
nestIdentityManager::is_active_group(const nestGroup *group)
{

    if ( group->isNoGroup() ) {
        return true;
    }

    string key = GROUP_KEY_PREFIX;
    key = key + group->getName();

    ClassAd *group_class_ad = this->Identities->GetClassAd(key);

    if (!group_class_ad) {
        return false;
    }

    string  attr("Members");
    ExprTree *group_list = group_class_ad->Lookup( attr );

    if (!group_list) {
        return false;
    }

    ExprListIterator list_itr( (ExprList *) group_list);

    Value   val;
    if (!list_itr.CurrentValue( val )) {
        //  delete group_list;
        return false;
    }

    //  Verify!
    string  member;
    if (val.IsStringValue( member )) {
        //  delete group_list;
        return true;
    }

    //  delete group_list;
    return false;
}

NestReplyStatus
nestIdentityManager::get_best_group( nestGroup *group,
                                     const nestUser *user )
{

    if (!user) {
        return NEST_BAD_PARAMETERS;
    }

    string key = USER_KEY_PREFIX;
    key = key + user->getName();

    ClassAd *user_class_ad = this->Identities->GetClassAd( key );
    if (!user_class_ad) {
        return NEST_USER_NOT_FOUND;
    }

    ExprTree *group_list = user_class_ad->Lookup("Groups");
    if (!group_list) {
        return NEST_CLASSAD_ERROR;
    }

    ExprListIterator    group_itr((ExprList *) group_list);

    Value   group_val;

    if (!group_itr.CurrentValue( group_val )) {
        return NEST_GROUP_NOT_FOUND;
    }

	string	str_group;
    if (!group_val.IsStringValue( str_group )) {
        return NEST_GROUP_NOT_FOUND;
    }
	group->setName( &str_group );

    return NEST_SUCCESS;
}

NestReplyStatus
nestIdentityManager::get_group_owner( const nestGroup *group,
                                      nestUser *owner)
{
    if(!group) {
        return NEST_BAD_PARAMETERS;
    }

    string key = GROUP_KEY_PREFIX;
    key = key + group->getName();

    ClassAd *group_class_ad = this->Identities->GetClassAd( key );
    if(!group_class_ad) {
        return NEST_GROUP_NOT_FOUND;
    }

    bool result;
    string owner_name;
    result = group_class_ad->EvaluateAttrString( "OwnerUser", owner_name );
    
    if(!result) {
        return NEST_CLASSAD_ERROR;
    }

	owner->setIdentity( &owner_name );
           
    return NEST_SUCCESS;
}  

NestReplyStatus
nestIdentityManager::set_default_group(const nestUser *user,
                                       const nestGroup *group,
                                       bool test_presence)
{

    if (!user || !group) {
        return NEST_BAD_PARAMETERS;
    }

    if (test_presence) {
        if (!this->is_group_in_user(user, group)) {
            /*
                Group not found in user's Groups list!
            */
            return NEST_GROUP_NOT_FOUND;
        }
    }

    string key = USER_KEY_PREFIX;
    key = key + user->getName();

    string modify_classad = "";
    modify_classad = modify_classad +
                        " [ DefaultGroup = \"" + group->getName() + "\" ] ";
    modify_classad = " [ Updates = " + modify_classad + " ] ";

    ClassAd *mod_classad = parser.ParseClassAd( modify_classad, true );

    if (!this->Identities->ModifyClassAd( key, mod_classad )) {
        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}

NestReplyStatus
nestIdentityManager::reset_default_group(const nestUser *user,
                                         const nestGroup *group)
{

    if (!user || !group) {
        return NEST_BAD_PARAMETERS;
    }

    string key = USER_KEY_PREFIX;
    key = key + user->getName();

    ClassAd *user_ad = this->Identities->GetClassAd( key );

    if (!user_ad) {
        return NEST_USER_NOT_FOUND;
    }

    string  default_group;

    if (user_ad->EvaluateAttrString( "DefaultGroup", default_group )) {

        /*
            if user's default_group is not group
            don't bother
        */

        if ( ! group->match( &default_group ) ) {
            return NEST_SUCCESS;
        }
    }

	nestGroup	best_group;
    NestReplyStatus reply = this->get_best_group( &best_group,
                                                  user );

    if (reply == NEST_SUCCESS) {
        return this->set_default_group( user,
										&best_group,
                                        false );
    }

    /*
        oops - the user's last group has been removed from its
        "Groups" list

        add user to NEST_NO_GROUP
    */

	nestGroup no_group( NEST_NO_GROUP );
	nestUser  admin( NestAdminUser );
    reply = this->add_user_in_group( &admin, user, &no_group );

    NEST_REPLY_CHECK(reply);

    return this->set_default_group( user, &no_group, true );

}

NestReplyStatus
nestIdentityManager::add_user_in_group(const nestUser *cur_user,
									   const nestUser *user,
                                       const nestGroup *group )
{

    if (!this->is_group_access_allowed(group, cur_user)) {
        return NEST_NONPRIVELEDGED_USER;
    }

    if (!this->is_user_present(user)) {
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                    "Trying to add a non-existing user to a group %s\n",
                    user,
                    group );
        return NEST_USER_NOT_FOUND;
    }

    if (!this->is_group_present(group)) {
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                    "Trying to add user to a non-existing group %s\n",
                    group );
        return NEST_GROUP_NOT_FOUND;
    }

    /*
        if ((strcmp(group, NEST_NO_GROUP) != 0) &&
            this->is_user_in_group(user, group)) {
    */

    if (this->is_group_in_user(user, group)) {
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                    "Trying to add an already existing user: %s to group %s\n",
                    user,
                    group );
        return NEST_USER_EXISTS;
    }


    /*
        Start:  keep your fingers crossed!
                it shouldn't crash now
    */

    NestReplyStatus reply = this->add_user_to_group_members(user, group);
    NEST_REPLY_CHECK(reply);

    reply = this->add_group_to_user_groups(user, group);

    /*
        End:    relax - the prayer is over and you are safe
    */

    /*
        Oh yeah! these two operations needs to be atomic!
        NestTime to use a Collection Transaction!
    */

    /*
        Not sure we should have this

    this->set_default_group(user, group, false);
    */

    return reply;
}

NestReplyStatus
nestIdentityManager::delete_user_in_group(const nestUser *cur_user,
										  const nestUser *user,
                                          const nestGroup *group )
{

    if (!this->is_group_access_allowed(group, cur_user)) {
        return NEST_NONPRIVELEDGED_USER;
    }

    if (!this->is_user_present(user)) {
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                    "Trying to delete a non-existing user %s from group %s\n",
                    user,
                    group);
        return NEST_USER_NOT_FOUND;
    }

    if (!this->is_group_present(group)) {
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                    "Trying to delete user to a non-existing group %s\n",
                    group );
        return NEST_GROUP_NOT_FOUND;
    }

    /*
        if ((strcmp(group, NEST_NO_GROUP) != 0)&&
            !this->is_user_in_group(user, group)) {
    */

    if (!this->is_group_in_user(user, group)) {
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                    "Trying to delete a non-member %s from group %s\n",
                    user,
                    group );
        return NEST_USER_NOT_FOUND;
    }

    /*
        Start:  keep your fingers crossed!
                it shouldn't crash now
    */

    NestReplyStatus reply = this->delete_user_from_group_members(user, group);
    NEST_REPLY_CHECK(reply);

    reply = this->delete_group_from_user_groups(user, group);

    /*
        End:    relax - the prayer is over and you are safe
    */

    /*
        Oh yeah! these two operations needs to be atomic!
        NestTime to use a Collection Transaction!
    */

    this->reset_default_group(user, group);

    return reply;
}

NestReplyStatus
nestIdentityManager::add_user( const nestUser *new_user,
							   bool log_existing )
{

    if (!new_user ) {
        return NEST_BAD_PARAMETERS;
    }

	const nestGroup *group = new_user->getGroup( );
	const char *identity = new_user->getIdentity( );

	// identity & group must be valid
    if ( !identity || !group ) {
        return NEST_BAD_PARAMETERS;
    }

    if (!this->is_group_present(group) ) {
        if ( log_existing ) {
            nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                        "Trying to add user %s to a non-existing group %s\n",
                        identity, group );
        }
        return NEST_GROUP_NOT_FOUND;
    }

    string  key = USER_KEY_PREFIX;
    key = key + identity;

    ClassAd *user_ad = this->Identities->GetClassAd( key );

    if (user_ad) {
        bool    result;
        bool    valid;

        result = user_ad->EvaluateAttrBool( "ValidEntry", valid );
        if (!result) {
            return NEST_CLASSAD_ERROR;
        }

        if (valid) {
            if ( log_existing ) {
                nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                            "Trying to add an already existing user '%s'\n",
                            identity );
            }
            return NEST_USER_EXISTS;

        } else {

            string modify_classad1 = "";
            modify_classad1 = modify_classad1 +
                                    "[   Updates  = " +
                                        "[  ValidEntry = true; ] " +
                                    "]";


            ClassAd *mod_classad1 = parser.ParseClassAd( modify_classad1,
                                                         true );

            if (!this->Identities->ModifyClassAd( key, mod_classad1 )) {
                return NEST_CLASSAD_ERROR;
            }

            string modify_classad2 = "";

            modify_classad2 = modify_classad2 +
				"[   Updates  = " +
				"[ Groups = { \""+group->getName()+"\" } ]" +
				"]";


            ClassAd *mod_classad2 = parser.ParseClassAd( modify_classad2,
                                                         true );

            if (!this->Identities->ModifyClassAd( key, mod_classad2 )) {
                return NEST_CLASSAD_ERROR;
            }
        }
	} else {

        const char    *os_user = new_user->getOsName( );
		if ( NULL == os_user ) {
			os_user = "";
		}

        const char    *local = new_user->getLocalName( );
		if ( NULL == local ) {
			local = "";
		}

        string str_user_classad = "";
        str_user_classad = str_user_classad +
			"[  Type = \"" + IDENTITY_TYPE + "\"; " +
			"   PerceivedName = \"" + identity + "\"; " +
			"   ActualName = \"" + os_user + "\"; " +
			"   LocalName = \"" + local + "\"; " +
			"   ValidEntry = true; " +
			"   IsUser = true; " +
			"   Groups = {\"" + group->getName() + "\"}; " +
			"   DefaultGroup = \"" + group->getName() + "\"; " +
			"   Members = {}; " +
			"   StorageSpaceUsed = " + "\"0\"" + "; " +
			"   DefaultLotHandle = \"\" " +
			"] " ;
		user_ad = parser.ParseClassAd( str_user_classad, true );

        if (!user_ad) {
            nest_debug( NEST_DEBUG_ALWAYS,
                        "Parsing UserClassAd for user %s failed!\n",
                        new_user->getName() );
            return NEST_CLASSAD_ERROR;
        }

        if (!this->Identities->AddClassAd( key, user_ad )) {
            nest_debug( NEST_DEBUG_ALWAYS,
                        "Adding UserClassAd for user '%s' failed!\n",
                        identity );
            return NEST_CLASSAD_ERROR;
        }
    }

    NestReplyStatus reply = this->add_user_to_group_members(new_user, group);

    if (reply != NEST_SUCCESS) {
        this->delete_user(new_user, NestAdminUser);
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                    "Error while adding user to group members!\n" );
        return reply;
    }

    int total_nest_users;

    reply = this->get_total_nest_users(&total_nest_users);

    if (reply != NEST_SUCCESS) {
        this->delete_user(new_user, NestAdminUser);
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                    "Error while adding user to group members!\n" );
        return reply;
    }

    total_nest_users++;
    key = TOTAL_USERS_KEY;

    char str_total_nest_users[NEST_NAMELEN];
    NestSnprintf(str_total_nest_users, NEST_NAMELEN, "%d",
                                            total_nest_users );

    string modify_classad3 = "";
    modify_classad3 = modify_classad3 +
		"[ TotalNestUsers = " + str_total_nest_users + "]";

    modify_classad3 = " [ Updates = " + modify_classad3 + " ] ";

    ClassAd *mod_classad3 = parser.ParseClassAd( modify_classad3, true );

    if(!mod_classad3){
        return NEST_CLASSAD_ERROR;
    }
    if (!this->Identities->ModifyClassAd( key, mod_classad3 )) {
        this->delete_user(new_user, NestAdminUser);
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                    "Error while adding user to group members!\n" );

        return NEST_CLASSAD_ERROR;
    }

    NestSync();

    return NEST_SUCCESS;
}

NestReplyStatus
nestIdentityManager::update_usage(const nestUserOrGroup *name,
								  nestLotSize more_size)
{
    string key = name->isUser() ? USER_KEY_PREFIX : GROUP_KEY_PREFIX;
    key = key + name->getName();
    ClassAd *user_ad = this->Identities->GetClassAd(key);
    if (!user_ad) {
        return NEST_USER_NOT_FOUND;
    }

    bool result;
    char tmp[64];
    nestLotSize size;
    result = user_ad->EvaluateAttrString("StorageSpaceUsed", tmp, sizeof(tmp));
    if (!result) {
        return NEST_CLASSAD_ERROR;
    }
	size = atoll( tmp );

    NestSnprintf(tmp, sizeof(tmp), "\"%lld\"", size + more_size);

    string str_modify_classad = "";
    str_modify_classad = str_modify_classad +
		"[  StorageSpaceUsed = " + tmp + "]";

    str_modify_classad = " [ Updates = " + str_modify_classad + " ] ";

    ClassAd *modify_classad = parser.ParseClassAd( str_modify_classad, true );

    if (!this->Identities->ModifyClassAd( key, modify_classad )) {
        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}

NestReplyStatus
nestIdentityManager::delete_user(const nestUser *user,
                                 bool log_existing)
{

    if (!user) {
        return NEST_BAD_PARAMETERS;
    }

	if ( user->isEverybody( ) ) {
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                    "Trying to delete NEST_EVERYBODY %s\n",
                    NEST_EVERYBODY );
        return NEST_NONPRIVELEDGED_USER;
    }

    if (!this->is_user_present(user)) {
        if ( log_existing ) {
            nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                        "Trying to delete an already deleted user %s\n",
                        user );
        }
        return NEST_USER_NOT_FOUND;
    }

    string modify_classad1 = "";
    modify_classad1 = modify_classad1 +
                            "[   Updates  = " +
                                "[  ValidEntry = false ] " +
                            "]";

    ClassAd *mod_classad1 = parser.ParseClassAd( modify_classad1, true );

    string key = USER_KEY_PREFIX;
    key = key + user->getName();

    ClassAd *user_class_ad = this->Identities->GetClassAd(key);
    if (!user_class_ad) {
        return NEST_USER_NOT_FOUND;
    }

    if (!this->Identities->ModifyClassAd( key, mod_classad1 )) {
        return NEST_CLASSAD_ERROR;
    }

    ExprTree *group_list = user_class_ad->Lookup("Groups");
    if (!group_list) {
        return NEST_CLASSAD_ERROR;
    }
    ExprListIterator    group_itr((ExprList *) group_list);

    Value   group_val;
    string  group_str;

    if (!group_itr.CurrentValue( group_val )) {
        return NEST_SUCCESS;
    }

    do {
        if (group_val.IsStringValue( group_str )) {
			nestGroup	group( &group_str );
            this->delete_user_from_group_members(user, &group );
        }
    } while( group_itr.NextValue( group_val ) );

    int total_nest_users;

    NestReplyStatus reply = this->get_total_nest_users(&total_nest_users);

    if(reply!=NEST_SUCCESS){
        return NEST_CLASSAD_ERROR;
    }

    total_nest_users--;
    key = TOTAL_USERS_KEY;

    char str_total_nest_users[NEST_NAMELEN];
    NestSnprintf(str_total_nest_users, NEST_NAMELEN, "%d",
                                        total_nest_users );

    string modify_classad2 = "";
    modify_classad2 = modify_classad2 +
                        "[ TotalNestUsers = " + str_total_nest_users + "]";

    modify_classad2 = " [ Updates = " + modify_classad2 + " ] ";

    ClassAd *mod_classad2 = parser.ParseClassAd( modify_classad2, true );

    if(!mod_classad2){
        return NEST_CLASSAD_ERROR;
    }

    if (!this->Identities->ModifyClassAd( key, mod_classad2 )) {
        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}

NestReplyStatus
nestIdentityManager::add_group( const nestUser *cur_user,
								const nestGroup *group )
{

    if (this->is_group_present(group)) {
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
            "Trying to add an existing group %s\n",
            group );
        return NEST_GROUP_EXISTS;
    }

    string str_key,
           str_group_classad = "";

    str_key = GROUP_KEY_PREFIX;
    str_key = str_key + group->getName();
        
    str_group_classad =  str_group_classad +
		"[  Type = \"" + IDENTITY_TYPE + "\"; " +
		"   PerceivedName = \"" + group->getName() + "\"; " +
		"   ActualName = \"" + group->getName() + "\"; " +
		"   ValidEntry = true; " +
		"   IsUser = false; " +
		"   OwnerUser = \"" + cur_user->getName() + "\"; " +
		"   Groups = {}; " +
		"   DefaultGroup = \"\"; " +
		"   Members = {}; " +
		"   StorageSpaceUsed = " + "\"0\"" + "; " +
		"   DefaultLotHandle = \"\" " +
		"]";

    ClassAd *group_classad = parser.ParseClassAd( str_group_classad, true );

    if (!group_classad) {
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
            "Parsing GroupClassAd for group %s during add-group failed!\n",
            group );
        return NEST_CLASSAD_ERROR;
    }

    if (!this->Identities->AddClassAd( str_key, group_classad )) {
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
               "Adding GroupClassAd for group %s during add-group failed!\n",
               group );
        return NEST_CLASSAD_ERROR;
    }
    
    NestReplyStatus reply = add_user_in_group(cur_user,cur_user,group);
    if(reply != NEST_SUCCESS) {
        delete_group(cur_user, group);
    }
   
    int total_nest_groups;
    reply = get_total_nest_groups(&total_nest_groups);
    if(reply != NEST_SUCCESS) {
        delete_user_in_group(cur_user, cur_user, group);
        delete_group(cur_user, group);
    }
    
    total_nest_groups++;
    string key = TOTAL_GROUPS_KEY;

    char str_total_nest_groups[NEST_NAMELEN];
    NestSnprintf(str_total_nest_groups, NEST_NAMELEN, "%d",
                                            total_nest_groups );

    string modify_classad = "";
    modify_classad = modify_classad +
                      "[ TotalNestGroups = " + str_total_nest_groups + "]";

    modify_classad = " [ Updates = " + modify_classad + " ] ";

    ClassAd *mod_classad = parser.ParseClassAd( modify_classad, true );

    if(!mod_classad){
        delete_user_in_group(cur_user, cur_user, group);
        delete_group(cur_user, group);
        return NEST_CLASSAD_ERROR;
    }

    if (!this->Identities->ModifyClassAd( key, mod_classad )) {
        delete_user_in_group(cur_user, cur_user, group);
        delete_group(cur_user, group);
        return NEST_CLASSAD_ERROR;
    }

    /*
        Move the commented portion at the
        end of this file,
        do appropriate modifications,
        when group reservations are enabled!
    */

    return NEST_SUCCESS;
}

NestReplyStatus
nestIdentityManager::delete_group(const nestUser *cur_user,
								  const nestGroup *group )
{
    if (!this->is_group_access_allowed(group, cur_user)) {
        return NEST_NONPRIVELEDGED_USER;
    }

    if (!this->is_group_present(group)) {
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                    "Trying to  delete a non-existing group %s\n",
                    group );
        return NEST_GROUP_NOT_FOUND;
    }

    if (this->is_active_group(group)) {
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                    "Trying to delete an active group %s\n",
                    group );
        return NEST_GROUP_ACTIVE;
    }

    string key = GROUP_KEY_PREFIX;
    key = key + group->getName();

    if (!this->Identities->RemoveClassAd( key )) {
        nest_debug( NEST_DEBUG_ACCT_MNGMNT,
                    "DeleteClassAd failed during group-delete for group %s!\n",
                    group );
        return NEST_CLASSAD_ERROR;
    }

    int total_nest_groups;
    NestReplyStatus reply = get_total_nest_groups(&total_nest_groups);
    NEST_REPLY_CHECK(reply);
    
    total_nest_groups--;
    key = TOTAL_GROUPS_KEY;

    char str_total_nest_groups[NEST_NAMELEN];
    NestSnprintf(str_total_nest_groups, NEST_NAMELEN, "%d",
                                            total_nest_groups );

    string modify_classad = "";
    modify_classad = modify_classad +
                      "[ TotalNestGroups = " + str_total_nest_groups + "]";

    modify_classad = " [ Updates = " + modify_classad + " ] ";

    ClassAd *mod_classad = parser.ParseClassAd( modify_classad, true );

    if (!this->Identities->ModifyClassAd( key, mod_classad )) {
        return NEST_CLASSAD_ERROR;
    }
    
    /*
        Move the commented portion at the
        end of this file, when group reservations are enabled!
    */

    return NEST_SUCCESS;
}

NestReplyStatus
nestIdentityManager::get_gid(user_id_t *gid, const nestGroup *group)
{
	const char *group_name = group->getName();

    if (!gid) {
        return NEST_BAD_PARAMETERS;
    }

    if (!group_name) {
        group_name = NEST_NO_GROUP;
    }

    string key = GROUP_KEY_PREFIX;
    key = key + group_name;

    ClassAd *group_class_ad = this->Identities->GetClassAd(key);
    if (!group_class_ad) {
        return NEST_GROUP_NOT_FOUND;
    }

    string  attr;
    string  actual_group;

    attr = "ActualName";
    if (!group_class_ad->EvaluateAttrString(attr, actual_group)) {
        return NEST_CLASSAD_ERROR;
    }

    group_ent   grp_ent;
    if (Nestgetgrpnam(actual_group.c_str(), &grp_ent, SystemGrpFile) < 0) {
        return ErrnoToNest(errno);
    }
    *gid = (user_id_t) grp_ent.grp_gid;

    return NEST_SUCCESS;
}

NestReplyStatus
nestIdentityManager::get_space_used(nestLotSize *size,
									const nestUserOrGroup *ug )
{
    *size = 0;
    string key;
    if(ug->isUser() ) {
        key = USER_KEY_PREFIX;
    } else {
        key = GROUP_KEY_PREFIX;
    }
    key = key + ug->getName();

    ClassAd *user_ad = this->Identities->GetClassAd(key);
    if (!user_ad) {
        return NEST_LOT_NOT_FOUND;
    }

    bool result;
	char tmp[64];
    result = user_ad->EvaluateAttrString( "StorageSpaceUsed",
										  tmp, sizeof(tmp)  );
    if (!result) {
        return NEST_CLASSAD_ERROR;
    }
	*size = atoll( tmp );
    return NEST_SUCCESS;
}


NestReplyStatus
nestIdentityManager::initialize(ClassAdCollection *resources )
{

    this->Identities = resources;

    this->query.Bind(resources);
    this->identities_view = NEST_IDENTITIES_VIEW;

    ViewName    parent_view = "root";

    string      view_constraint(""),
                view_rank(""),
                view_partition_expr("");

    view_constraint = view_constraint +
		"other.Type == \"" + IDENTITY_TYPE + "\"";

    if (!this->Identities->CreateSubView( this->identities_view,
										  parent_view,
                                          view_constraint,
                                          view_rank,
                                          view_partition_expr ) ) {
        nest_debug( NEST_DEBUG_ALWAYS,
                    "nestIdentityManager Couldn't create view for itself!\n");
        return NEST_CLASSAD_ERROR;
    }

    string str_key = TOTAL_USERS_KEY;

    string str_total_users_classad = "";
    str_total_users_classad = str_total_users_classad +
                            "[ Type = \"" + TOTAL_USERS_KEY + "\";" +
                            "  TotalNestUsers = 0" +
                            "]";

    ClassAd *total_users_ad = parser.ParseClassAd( str_total_users_classad,
                                                   true );

    if (!total_users_ad) {
        nest_debug( NEST_DEBUG_ALWAYS,
                    "nestIdentityManager Error creating 'total_users'\n");
        return NEST_CLASSAD_ERROR;
    }

    if (!this->Identities->AddClassAd( str_key, total_users_ad )) {
        nest_debug( NEST_DEBUG_ALWAYS,
                    "nestIdentityManager Error adding 'total_users'\n");
        return NEST_CLASSAD_ERROR;
    }

    str_key = TOTAL_GROUPS_KEY;

    string str_total_groups_classad = "";
    str_total_groups_classad = str_total_groups_classad +
                            "[ Type = \"" + TOTAL_GROUPS_KEY + "\";" +
                            "  TotalNestGroups = 0" +
                            "]";

    ClassAd *total_groups_ad = parser.ParseClassAd( str_total_groups_classad,
                                                   true );

    if (!total_groups_ad) {
        nest_debug( NEST_DEBUG_ALWAYS,
                    "nestIdentityManager Error creating 'total_groups'\n");
        return NEST_CLASSAD_ERROR;
    }

    if (!this->Identities->AddClassAd( str_key, total_groups_ad )) {
        nest_debug( NEST_DEBUG_ALWAYS,
                    "nestIdentityManager Error adding 'total_groups'\n");
        return NEST_CLASSAD_ERROR;
    }

    char    actual_nest_group[NEST_NAMELEN];
    if (this->get_actual_nest_group(actual_nest_group) != NEST_SUCCESS) {
        nest_debug( NEST_DEBUG_ALWAYS,
                    "Cannot find a default nest_group!\n" );
        return NEST_GROUP_NOT_FOUND;
    }

    str_key = GROUP_KEY_PREFIX;
    str_key = str_key + NEST_NO_GROUP;

    string str_group_classad =  "";
    str_group_classad = str_group_classad +
		"[  Type = \"" + IDENTITY_TYPE + "\"; " +
		"   PerceivedName = \"" + NEST_NO_GROUP + "\"; " +
		"   ActualName = \"" + actual_nest_group + "\"; "+
		"   ValidEntry = true; " +
		"   IsUser = false; " +
		"   OwnerUser = \"" + actual_nest_group + "\"; " +
		"   Groups = {}; " +
		"   DefaultGroup = \"\"; " +
		"   Members = {}; " +
		"   DefaultLotHandle = \"\" " +
		"] " ;

    ClassAd *group_classad = parser.ParseClassAd( str_group_classad, true );

    if (!group_classad) {
        nest_debug( NEST_DEBUG_ALWAYS,
                    "Parsing GroupClassAd for group %s during init failed!\n",
                    actual_nest_group );
        return NEST_CLASSAD_ERROR;
    }

    if (!this->Identities->AddClassAd( str_key, group_classad )) {
        nest_debug( NEST_DEBUG_ALWAYS,
                    "nestIdentityManager Error adding group ClassAd\n");
        return NEST_CLASSAD_ERROR;
    }

	nest_debug( NEST_DEBUG_ACCT_MNGMNT, "Creating the 'everbody' user\n" );

	// Create the "everybody" user
	nestGroup	no_group( NEST_NO_GROUP );
	nestUser	everybody( NEST_EVERYBODY );
	everybody.setOsName( NestAdminUser );
	everybody.setGroup( &no_group );
    NestReplyStatus reply = this->add_user( &everybody );

    NEST_REPLY_CHECK(reply);

    str_key = USERLIST_KEY;

    string  str_userlist_classad = "";
    str_userlist_classad = str_userlist_classad +
		"[  Type = \"" + USERLIST_KEY + "\"; " +
		"   UserList = \"\"; " +
		"   DenyList = \"\"  " +
		"] ";

    ClassAd *userlist_ad = parser.ParseClassAd( str_userlist_classad, true );

    if (!userlist_ad) {
        nest_debug( NEST_DEBUG_ALWAYS,
                    "nestIdentityManager No userlist ad\n");
        return NEST_CLASSAD_ERROR;
    }

    if (!this->Identities->AddClassAd( str_key, userlist_ad )) {
        nest_debug( NEST_DEBUG_ALWAYS,
                    "nestIdentityManager Error adding userlist ClassAd\n");
        return NEST_CLASSAD_ERROR;
    }

    return this->synchronize(true);
}

NestReplyStatus
nestIdentityManager::recover(ClassAdCollection *resources )
{

    this->Identities = resources;
    this->query.Bind(resources);
    this->identities_view = NEST_IDENTITIES_VIEW;

    return this->synchronize();
}


NestReplyStatus
nestIdentityManager::shutdown()
{
    return NEST_SUCCESS;
}


NestReplyStatus
nestIdentityManager::synchronize(bool is_init)
{

    string key;
    bool result;

    key = USERLIST_KEY;

    ClassAd *userlist_ad = this->Identities->GetClassAd( key );

    if (!userlist_ad) {
        return NEST_CLASSAD_ERROR;
    }

    result  = userlist_ad->EvaluateAttrString( "UserList",
                                               this->prev_userlist );
    result &= userlist_ad->EvaluateAttrString( "DenyList",
                                               this->prev_denylist );

    if (!result) {
        return NEST_CLASSAD_ERROR;
    }

    NestReplyStatus reply;

    int state = Nestset_grid_mapfile(GridMapFile);
    if ( state < 0) {
        nest_debug( NEST_DEBUG_ALWAYS,
                    "Couldn't read system grid-mapfile at %s!\n",
                    GridMapFile );
        return NEST_SYSTEM_FILE_MISSING;
    }

    string str_key;

    struct grid_mapfile_entry *user;
	nestGroup	no_group( NEST_NO_GROUP );
    while( (user = Nestget_grid_mapfile_entry(state)) != NULL) {

        //  Multiple identities could be mapped to the same user!!
        str_key = USER_KEY_PREFIX;
        str_key = str_key + user->name;
		string str_user_classad = "";

		// Create the user object
		nestUser	nest_user( user->identity );
		nest_user.setOsName( user->name );
		nest_user.setGroup( &no_group );

		// And, validate it...
        if (this->is_valid_nest_user( &nest_user, is_init )) {
            reply = this->add_user( &nest_user, false );
            if (reply == NEST_USER_EXISTS) {
                reply = NEST_SUCCESS;
            }
        }
        else {
            reply = this->delete_user( &nest_user, false );
            if (reply == NEST_USER_NOT_FOUND) {
                reply = NEST_SUCCESS;
            }
        }

        NEST_REPLY_CHECK(reply);
    }   //  End-of-while
    Nestend_grid_mapfile(state);

    string modify_classad = "";
    modify_classad = modify_classad +
		" [ UserList = \"" + this->userlist + "\";" +
		"   DenyList = \"" + this->denylist + "\"] ";
    modify_classad =
		" [ Updates = " + modify_classad + " ] ";

    ClassAd *mod_classad = parser.ParseClassAd( modify_classad, true );

    if (!this->Identities->ModifyClassAd( key, mod_classad )) {
        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}
