/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#ifndef _NEST_LOT_MANAGER_H_

#define _NEST_LOT_MANAGER_H_


#include "nest_quota_manager.h"
#include "nest_device_manager.h"
#include "nest_device.h"

#include "nest_common.h"
#include "nest_user.h"

#define RMCurrentList   ".RMCurrentList"
#define RMPendingList   ".RMPendingList"

#define SCHED_NEXT      0

class nestLotManager
{

  public:

	nestLotManager( void );
	~nestLotManager();
	NestReplyStatus setup( const char *rootDir,
						   nestLotSize storage_space );

	bool kernel_quota_enabled();
	nestLotSize delete_dead_rsvs(nestLotSize req_blks,
							int is_user,
							bool aggressive = true);

	NestReplyStatus update_lot_usage(const char *logical_filename,
									 nestLotSize size);

	NestReplyStatus update_file_lot_usage( const char *logical_file,
										   nestFileSize size );

	NestReplyStatus update_dir_lot_usage( const char *logical_file,
										  nestLotSize size );

	NestReplyStatus update_user_lot_usage( const char *logical_file,
										   nestLotSize size );
        
	NestReplyStatus update_group_lot_usage( const char *logical_file,
											nestLotSize size );

	NestReplyStatus modify_lot_usage(const nestUser *user,
									 const char *logical_filename,
									 const NestLotHandle  *from_lot,
									 nestLotSize from_share,
									 const NestLotHandle  *to_lot);

	NestReplyStatus set_user_default_lot( const NestLotHandle *handle,
										  const nestUser *user);

	NestReplyStatus request_user_lot(NestLotHandle *handle,
									 const nestUser *user,
									 nestLotSize size,
									 int duration,
									 const nestUser *lot_user,
									 const nestUserOrGroup *parent_ug,
									 const NestLotHandle *parent_handle,
									 int lot_type = G_LOT,
									 int cost = 1);

	NestReplyStatus request_group_lot(NestLotHandle *handle,
									  const nestUser *user,
									  nestLotSize size,
									  int duration,
									  const nestGroup *lot_group,
									  const nestUserOrGroup *parent_ug,
									  const NestLotHandle *parent_handle,
									  int lot_type = G_LOT,
									  int cost = 1);
        
	NestReplyStatus get_parent_lot(const nestUser *user,
								   const nestUserOrGroup *lot_user,
								   const nestUserOrGroup *parent_user,
								   NestLotHandle *parent_handle,
								   nestLotSize size );

	NestReplyStatus get_lot_parent( const nestUserOrGroup *lot_user,
									const NestLotHandle *parent_handle,
									NestLotInternal *lot_parent );

#if 0   
	bool    is_access_allowed_to_request(const nestUser *lot_user,
										 nestUserOrGroup **parent_user,      
										 NestLotHandle *parent_handle );
#endif
	bool check_request_access( const NestLotHandle	*parent_handle,
							   const nestUserOrGroup *ug );

	bool check_request_access( const NestLotInternal	*parent_lot,
							   const nestUserOrGroup *ug );

	nestUserOrGroup *get_lot_user( const NestLotHandle *handle );
	nestUserOrGroup *get_lot_user( const NestLotInternal *lot );
   
	bool    is_access_allowed_to_lot(const NestLotHandle *handle,
									 const nestUserOrGroup *user,
									 bool allow_group_members = true,
									 bool allow_admin = true);

	bool    is_access_allowed_to_user(const nestUser *user,
									  const nestUserOrGroup *lot_user,
									  bool allow_group_members = true,
									  bool allow_admin = true);

#if 0
	NestReplyStatus list_user_lots(NestLotListInternal **result,
								   const nestUser *user,
								   const nestUser *lot_user);

	NestReplyStatus list_group_lots(NestLotListInternal **result,
									const nestUser *req_user,
									const nestGroup *group);
#endif

	NestReplyStatus update_lot(const NestLotHandle *handle,
							   const nestUserOrGroup *user,
							   nestLotSize more_size,
							   int more_duration,
							   int more_cost = 0,
							   bool update_total_size = true);

	NestReplyStatus change_lot_type(const NestLotHandle *handle,
									const nestUser *user,
									int lot_type,
									int more_duration);

	NestReplyStatus stat_lot( const NestLotHandle *handle,
							  NestLotInternal *result );

	NestReplyStatus stat_lot( const NestLotHandle *handle,
							  const nestUser *user,
							  NestLotInternal *result );

	NestReplyStatus terminate_lot(const NestLotHandle *handle,
								  const nestUserOrGroup *user,
								  int sched_next = SCHED_NEXT);
  
	NestReplyStatus         list_lots(NestLotListInternal **result,
									  const nestUser *user,
									  const nestUserOrGroup *lot_ug );

	nestLotSize free_space ( int is_user );
	nestLotSize total_space( int is_user );
	nestLotSize available_lot_space(NestLotHandle *handle);
	nestLotSize available_lot_space(const nestUser *user,
							   const char *logical_file);
	nestLotSize available_file_lot_space(const char *logical_file);
	nestLotSize available_dir_lot_space(const char *logical_file);
	nestLotSize available_user_lot_space(const nestUser *user);
	nestLotSize available_group_lot_space(const nestUser *user);
	NestReplyStatus available_group_lot_space(const nestUser *user,
											  const nestGroup *group,
											  nestLotSize *size);

	NestReplyStatus add_user_to_lot(const NestLotHandle *handle,
									const nestUser *user,
									const nestUser *lot_user);
	NestReplyStatus delete_user_from_lot(const NestLotHandle *handle,
										 const nestUser *user,
										 const nestUser *lot_user);

	NestReplyStatus get_file_usage(const char *logical_filename,
								   nestLotSize *usage);

	int init();
	NestReplyStatus initialize(ClassAdCollection *resources);
	NestReplyStatus recover(ClassAdCollection *resources);

	NestReplyStatus synchronize();
	NestReplyStatus shutdown();

  private:
	ClassAdCollection       *lot_catalog;
	ClassAdCollection       *identities;

	ViewName                lot_catalog_view;

	ClassAdParser           parser;
	ClassAdUnParser         unparser;
	LocalCollectionQuery    query;

	rsv_id_t                next_rsv_id;

	const char				*root_dir;
	nestLotSize             storage_space;
	int                     max_lot_duration;
	int                     default_lot_duration;
	bool                    has_kernel_quota;

	QuotaManager            *QuotaMngr;
	DeviceManager           *DeviceMngr;

	rsv_id_t                get_next_rsv_id();

	NestReplyStatus         sync_next_rsv_id();

	NestReplyStatus         request_lot(NestLotHandle *handle,
										const NestLotHandle *parent_handle,
										const nestUserOrGroup *parent_name,
										const nestUserOrGroup *user_or_group,
										nestLotSize size,
										int duration,
										int lot_type=G_LOT,
										int cost = 1);

	NestReplyStatus         list_sub_lots(NestLotListInternal **result,
										  const NestLotHandle *handle);

	NestReplyStatus         LotClassAd_to_NestLot(ClassAd *lot_ad,
												  NestLotInternal *rstat);

	NestReplyStatus         set_actual_name(const nestUserOrGroup *user );

	NestReplyStatus         get_lot_and_share(const char *arg,
											  char *lot,
											  nestLotSize *share);

	NestReplyStatus         change_lot_usage(const char *lot_key,
											 nestLotSize *size,
											 bool allocate_as_much = false);

	NestReplyStatus         add_lot_to_file(const char * lot_handle,
											nestLotSize size,
											const char * file_key);

	NestReplyStatus         extend_current_lot(const char * file_key,
											   nestLotSize size);

	NestReplyStatus         check_default_lot(const char * file_key,
											  nestLotSize size,
											  const char * user);

	NestReplyStatus         reclaim_lot_space( const char *file_key,
											   nestLotSize size );
};

extern nestLotManager	*lot_manager;

#endif
