/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "nest_common.h"
#include "nest_protocol.h"

nestProtocol::~nestProtocol() {}

bool
nestProtocol::sendInitialAck( )
{
    return sendAck();
}

bool
nestProtocol::sendFinalAck( int transferType )
{
    return sendAck();
}

bool
nestProtocol::persistent()
{
    return true;
}

bool
nestProtocol::sendDirectoryList( NestFileInfo *list )
{
    PROTOCOL_WARN;
    assert( 0 );
    return false;
}

bool
nestProtocol::sendNestLot( NestLotInternal *rstat )
{
    PROTOCOL_WARN;
    assert( 0 );
    return false;
}

bool
nestProtocol::sendLotHandle( NestLotHandle *handle )
{
    PROTOCOL_WARN;
    assert( 0 );
    return false;
}

bool
nestProtocol::sendAd( nestAd *ad )
{
    PROTOCOL_WARN;
    assert( 0 );
    return false;
}

bool
nestProtocol::sendPwd( const char *pwd )
{
    PROTOCOL_WARN;
    assert( 0 );
    return false;
}

bool
nestProtocol::sendVersion( const char *version )
{
    PROTOCOL_WARN;
    assert( 0 );
    return false;
}

bool
nestProtocol::sendProxyList( list <char *> *proxys )
{
    PROTOCOL_WARN;
    assert( 0 );
    return false;
}

bool
nestProtocol::sendDiskspace( nestLotSize avail, nestLotSize total  )
{
    PROTOCOL_WARN;
    assert( 0 );
    return false;
}

bool
nestProtocol::sendBandwidthStatistics( )
{
    PROTOCOL_WARN;
    assert( 0 );
    return false;
}

bool
nestProtocol::sendPhysicalName( const char *physical_name )
{
    PROTOCOL_WARN;
    assert( 0 );
    return false;
}

NestReplyStatus
nestProtocol::sendFileinfo( const NestFileInfo *info )
{
    PROTOCOL_WARN;
    assert( 0 );
    return false;
}

NestReplyStatus
nestProtocol::setFileattr( NestFileInfo *info,
						   const char *fname,
						   const void *fattr )
{
    PROTOCOL_WARN;
    assert( 0 );
    return NEST_UNKNOWN_ERROR;;
}

bool
nestProtocol::sendAcl( const char *user,
                       int is_user,
                       nest_access_rights_t access )
{
    (void) user;
    (void) is_user;
    (void) access;

    PROTOCOL_WARN;
    assert( 0 );
    return false;
}

bool
nestProtocol::sendAcl( const NestACL *Acl )
{
    (void) Acl;

    PROTOCOL_WARN;
    assert( 0 );
    return false;
}

bool
nestProtocol::sendTransferInfo( class nestRequest *req )
{
    PROTOCOL_WARN;
    assert( 0 );
    return false;
}

NestReplyStatus
nestProtocol::sendFilehandle( NestFileInfo *fi )
{
    PROTOCOL_WARN;
    assert( 0 );
    return NEST_NOT_YET_IMPLEMENTED;
}
    
NestReplyStatus 
nestProtocol::sendProxyIndex( DynamicLard *index )
{
    PROTOCOL_WARN;
    assert( 0 );
    return NEST_NOT_YET_IMPLEMENTED;
}
    

NestReplyStatus
nestProtocol::gssAccept( char **username, char **globus_username )
{
    PROTOCOL_WARN;
    assert( 0 );
    return false;
}

nestProtocol *
nestProtocol::clone()
{
    PROTOCOL_WARN;
    assert( 0 );
    return NULL;
}

NestReplyStatus
nestProtocol::receive_state( int pipe )
{
    // default is a no-op
    // if any particular protocol has state, here is where it can be sent
    // to a process worker
    // see nestProtocol::receive_state
    return NEST_SUCCESS;
}

NestReplyStatus 
nestProtocol::send_state( int pipe )
{
    // see nestProtocol::receive_state
    return NEST_SUCCESS;
}


bool
nestProtocol::sendTimestamp( )
{
    PROTOCOL_WARN;
    assert( 0 );
    return false;
}

bool
nestProtocol::endThirdPartyTransfer( int type )
{
    PROTOCOL_WARN;
    assert( 0 );
    return false;
}

// returns -1 on error
ssize_t
nestProtocol::Sendn( const char *buffer, ssize_t n )
{

    ssize_t nleft = n;
    const char *ptr = buffer;
    ssize_t nwritten;

    while ( nleft > 0 ) {
        if ( ( nwritten = Send( ptr, nleft ) ) < 0 ) {
            return -1;
        }
        nleft -= nwritten;
        ptr += nwritten;
    }

    return n;
}

// returns -1 on error
ssize_t
nestProtocol::Send( const string &buffer )
{

    ssize_t nleft = buffer.length( );
    const char *ptr = buffer.c_str( );
    ssize_t nwritten;

    while ( nleft > 0 ) {
        if ( ( nwritten = Send( ptr, nleft ) ) < 0 ) {
            return -1;
        }
        nleft -= nwritten;
        ptr += nwritten;
    }

    return buffer.length();
}

ssize_t
nestProtocol::Receiven( char *buffer, ssize_t n )
{

    ssize_t nleft = n;
    char *ptr = buffer;
    ssize_t nread;

    while( nleft > 0 ) {
        if ( ( nread = Receive( ptr, nleft ) ) < 0 ) {
            nest_perror( "read" );
            return -1;
        } else if ( nread == 0 ) {
            break;        // EOF
        }
    
        nleft -= nread;
        ptr   += nread;
    }

    return( n - nleft );
}
