/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#ifndef _NEST_PROTOCOL_FILESYSTEM_H
#define _NEST_PROTOCOL_FILESYSTEM_H

#include "nest_common.h"
#include "nest_protocol_storage.h"
#include "nest_speak.i.h"

#include "nest_pread.h"

class nestProtocolFilesystem : public nestProtocolStorage {
public:


    nestProtocolFilesystem( int filedescriptor );

    nestProtocolFilesystem( int filedescriptor, off_t offset );

    virtual ~nestProtocolFilesystem();

    virtual ssize_t Receive( char *buffer, ssize_t n );

    virtual ssize_t Send( const char *buffer, ssize_t n );

    virtual int getType( ) { return STORAGE_PROTOCOL; }

    virtual bool hasSocket() { return true; }

    int getSocket() { return filedescriptor; }

    virtual nestProtocol *clone();


    virtual int  set_filenames(char *logical_name, char *physical_name);

    virtual int  get_filenames(char *logical_name, char *physical_name);

    //virtual int  set_offset( off_t offset ); 

    virtual int  set_open_flags(int flags);

    virtual int  get_open_flags(int *flags);


    virtual int  set_open_filesize(nestFileSize size);

    virtual int  get_open_filesize(nestFileSize *size);


    virtual int  set_ids(user_id_t uid, user_id_t gid);

    virtual int  get_ids(user_id_t *uid, user_id_t *gid);


    virtual NestReplyStatus receive_state( int pipe );

    virtual NestReplyStatus send_state( int pipe );

    virtual int set_fdkey( const char *fdkey );

        // call this to tell the connection not to close the fd
    virtual void set_fd_cache_status( bool status );

    virtual bool get_fd_cache_status( );

	const char *getName( void ) { return "FileSystem"; };

private:
    void init( int filedescriptor );
    
    int     filedescriptor;
    off_t   offset;

    bool    use_pread;
    
    int     open_flags;
    off_t   open_filesize;
    char    logical_filename[NEST_MAXPATHLEN],
            physical_filename[NEST_MAXPATHLEN];
    char    *fdkey;

    bool    fd_is_cached;
    user_id_t   uid, gid;
};

#endif
