/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "nest_protocol_manager.h"
#include "nest_config.h"
#include "nest_globus.h"

#include "nest_protocol_nest.h"
#include "nest_protocol_http.h"
#include "nest_protocol_proxy.h"
#include "nest_protocol_filesystem.h"
#include "nest_protocol_advertise.h"

#define LAST_PROTOCOL_SOCKET -99;

nestProtocolManager::nestProtocolManager (
	void )
{
	maxsocket = -1;

	NestPort = 0;
	HttpPort = 0;
	ProxyPort = 0;

	protocol_index = -1;
}

nestProtocolManager::~nestProtocolManager ( void )
{
}

int
nestProtocolManager::init( ) {

    const char *port_string;
    if (GetConfigValue(Config_NestPort, &port_string) && *port_string != 0 ) {
        NestPort = atoi(port_string);
    }
    if (GetConfigValue(Config_ProxyPort, &port_string) && *port_string != 0 ) {
        ProxyPort = atoi(port_string);
    }
    if (GetConfigValue(Config_HTTPPort, &port_string) && *port_string != 0 ) {
        HttpPort = atoi(port_string);
    }

        // now attempt to set up our gsi credentials if needed 
    if ( NestPort != 0 ) {
        getCredential();
    }

        // set chirp socket
    if ( NestPort != 0 ) {
        sockets[NEST_PROTOCOL] = nestProtocolNest::listen( NestPort );
    } else {
        sockets[NEST_PROTOCOL] = 0;
    }
    reportOpenPort( "NeST",     NestPort, sockets[NEST_PROTOCOL] );

        // set chirp socket
    if ( ProxyPort != 0 ) {
        sockets[PROXY_PROTOCOL] = nestProtocolNest::listen( ProxyPort );
    } else {
        sockets[PROXY_PROTOCOL] = 0;
    }
    reportOpenPort( "Proxy",     ProxyPort, sockets[PROXY_PROTOCOL] );

        // set http socket
    if ( HttpPort != 0 ) {
        sockets[HTTP_PROTOCOL] = nestProtocolHttp::listen( HttpPort );
    } else {
        sockets[HTTP_PROTOCOL] = 0;
    }
    reportOpenPort( "HTTP",     HttpPort, sockets[HTTP_PROTOCOL] );

    setMaxsocket();

    return 0;
}

NestReplyStatus credentialStatus = NEST_NOT_YET_INITIALIZED;
gss_cred_id_t credential_handle = GSS_C_NO_CREDENTIAL;
OM_uint32 lifetime;
gss_name_t *name;

NestReplyStatus
nestProtocolManager::getCredentialInfo( char **subject, long *expires ) {
    if ( credentialStatus != NEST_SUCCESS ) {
        return credentialStatus;
    }
    NestReplyStatus status = inquireGlobusCredential( 
            name,
            &lifetime,
            &credential_handle );
    *subject = (char*)name;
    *expires = (long)lifetime;
    return status;
}

void
nestProtocolManager::getCredential( ) {
    const char *value = NULL;

        // first check and set any specified environmental values used by 
        // x509 and globus code
    if ( GetConfigValue( Config_GlobusLocation, &value ) && *value ) {
        if ( ! NestIsDir( value ) ) {
            nest_die( "Globus location %s is not a valid directory!\n", value);
        }
        NestSetenv( "GLOBUS_LOCATION", value, 1 );
    }
    if ( GetConfigValue( Config_X509CertDir, &value ) && *value ) {
        if ( ! NestIsDir( value ) ) {
            nest_die( "x509_cert_dir %s is not a valid directory!\n", value);
        }
        NestSetenv( "X509_CERT_DIR", value, 1 );
    }
    if ( GetConfigValue( Config_X509HostCert, &value ) && *value ) {
        if ( ! NestIsFile( value ) ) {
            nest_die( "x509_host_cert %s is not a valid file!\n", value);
        }
        NestSetenv( "X509_USER_CERT", value, 1 );
    }
    if ( GetConfigValue( Config_X509HostKey, &value ) && *value ) {
        if ( ! NestIsFile( value ) ) {
            nest_die( "x509_host_key %s is not a valid file!\n", value);
        }
        NestSetenv( "X509_USER_KEY", value, 1 );
    }
    if ( GetConfigValue( Config_GridMapFile, &value ) && *value ) {
        if ( ! NestIsFile( value ) ) {
            nest_die( "grid_mapfile %s is not a valid file!\n", value);
        }
        NestSetenv( "GRIDMAP", value, 1 );
    }

	credential_handle = GSS_C_NO_CREDENTIAL;
    credentialStatus = acquireGlobusCredential( 
            &credential_handle, GSS_C_ACCEPT );
    nest_debug( NEST_DEBUG_CHIRP | NEST_DEBUG_AUTH, 
            "Acquired initial globus credential: %s\n",
                NestErrorString( credentialStatus ) );
    if ( credentialStatus != NEST_SUCCESS ) {
        nest_warn( __FILE__, __FUNCTION__, 
                "unable to acquire initial gsi certificate: %s\n",
                NestErrorString( credentialStatus ) );
    }
}

void
nestProtocolManager::reportOpenPort( char *protocol, int port, int status )
{
    if ( status < 0 ) {
        nest_warn( __FILE__, __FUNCTION__, 
                "Unable to listen to %s requests on port %d.\n",
                protocol, port );
    } else if ( status == 0 ) {
        nest_debug( NEST_DEBUG_PROTOCOL, 
                "%s protocol disabled at user request\n", protocol );
    } else {
        nest_debug( NEST_DEBUG_PROTOCOL, 
                "Listening to %s requests on port %d (sd %d).\n", 
                protocol, port, status );
    }
}

NestReplyStatus 
nestProtocolManager::initiateThirdPartyTransfer( nestClient *client, 
        nestProtocol **connection, nestProtocolType protocol ) 
{
        // also, it might be nice to cache connections
        // also, how about authentication?
        // 

    NestReplyStatus stat;
    nest_debug( NEST_DEBUG_PROTOCOL,
				"Trying to initiateThirdParty with %s\n",
				nestProtocolManager::protocolName( protocol ) );

    if ( protocol == NEST_PROTOCOL ) {
        stat = nestProtocolNest::initiateThirdParty
            ( client, connection );
    } else if ( protocol == HTTP_PROTOCOL ) {
        stat = nestProtocolHttp::initiateThirdParty( client, connection );
    } else {
        nest_warn( __FILE__, __FUNCTION__, 
				   "Unable to negotiate 3 party transfer with %s.\n",
				   nestProtocolManager::protocolName( protocol ) );
        return NEST_NOT_YET_IMPLEMENTED;
    }

    
    nest_debug( NEST_DEBUG_PROTOCOL,
            "initiateThirdParty with %s: %s\n", protocol,
            NestErrorString( stat ) );
    return stat;
}

// this is used by the nest_worker who is passed a protocol and a csd
nestProtocol* 
nestProtocolManager::newConnection( int csd, int protocol )
{
    switch( protocol ) {
    case NEST_PROTOCOL:
        return (nestProtocol *) new nestProtocolNest( csd );
    case PROXY_PROTOCOL:
        return (nestProtocol *) new nestProtocolProxy( csd );
    case HTTP_PROTOCOL:
        return (nestProtocol *) new nestProtocolHttp( csd ); 
    case STORAGE_PROTOCOL:
        return (nestProtocol *) new nestProtocolFilesystem( csd, (off_t)0 );
    case ADVERTISE_PROTOCOL:
        return (nestProtocol *) new nestProtocolAdvertise( ); 
    default:
            // we should never get here
        assert( 0 );
        return NULL;
    }
}

// this function receives a protocol connection from a socket
nestProtocol *
nestProtocolManager::ReceiveConnection( int sd, int protocol_id ) {
    
    nestProtocol *connection;
    int socket;
    
        // receive the fd and create the protocol connection 
    nest_debug( NEST_DEBUG_CONCURRENCY, "Recving fd on pipe %d\n", sd );
    socket = NestReceiveFd( sd );
    nest_debug( NEST_DEBUG_CONCURRENCY, "Recved fd %d on pipe %d\n", 
            socket, sd );
    connection = newConnection( socket, protocol_id );

        // receive any other state if necessary
    nest_debug( NEST_DEBUG_CONCURRENCY, "Recving state for protocol type %d "
            "on pipe %d\n", connection->getType(), sd );
    if ( connection->receive_state( sd ) != NEST_SUCCESS ) {
        assert( 0 );
        return NULL;
    } else {
        nest_debug( NEST_DEBUG_CONCURRENCY, "Recved state on pipe %d\n", sd );
        return connection;
    }

}

// this function sends a protocol connection over a socket
NestReplyStatus
nestProtocolManager::SendConnection( int sd, nestProtocol *connection ) {

    if ( ! connection->hasSocket() ) return NEST_NOT_YET_IMPLEMENTED;
    NestReplyStatus status;

        // pass the fd
    nest_debug( NEST_DEBUG_CONCURRENCY, "Sending fd %d on pipe %d\n", 
            connection->getSocket(), sd );
	status = NestSendFd( sd, connection->getSocket() );
    if ( status != NEST_SUCCESS ) {
        return status;
    }

    nest_debug( NEST_DEBUG_CONCURRENCY, "Sending state on pipe %d\n", sd );
    status = connection->send_state( sd );
    nest_debug( NEST_DEBUG_CONCURRENCY, "Sent state on pipe %d\n", sd );
    return status; 
}

nestClient* 
nestProtocolManager::accept(
	int protocol_socket )
{

    nestProtocol *connection = NULL;
    nestClient   *client     = NULL;
    struct sockaddr_in addr;

        // all the protocols create a connection object
        // some protocols only accept if the fd is ready, others don't care
	switch( getProtocolFromSocket( protocol_socket ) ) {
	case HTTP_PROTOCOL:
		connection = nestProtocolHttp::accept( protocol_socket, &addr );
		break;

	case NEST_PROTOCOL:
		connection = nestProtocolNest::accept( protocol_socket, &addr );
		break;

	case PROXY_PROTOCOL:
		connection = nestProtocolProxy::accept( protocol_socket, &addr);
		break;

	default:
		assert( 0 );
	}

    if ( connection != NULL ) {
        assert( client == NULL );
        client = new nestClient( storage_manager->getInitialClientDir(),
								 connection, &addr );
    }
    return client;
}

int
nestProtocolManager::getProtocolPortFromSocket( int protocol_socket ) {
    switch( getProtocolFromSocket( protocol_socket ) ) {
    case PROXY_PROTOCOL: return ProxyPort;
    case NEST_PROTOCOL:  return NestPort;
    case HTTP_PROTOCOL:  return HttpPort;
    default:
        return -1;
    }
}


char *
nestProtocolManager::getProtocolNameFromSocket( int protocol_socket )
{
    return protocolName( getProtocolFromSocket( protocol_socket ) );
}

char *
nestProtocolManager::protocolName( int protocol )
{
    static char nest_string[] = "Chirp";
    static char http_string[] = "HTTP";
    static char stor_string[] = "Storage";
    static char prox_string[] = "Proxy";

    switch( protocol ) {
    case NEST_PROTOCOL:     return nest_string;
    case HTTP_PROTOCOL:     return http_string;
    case STORAGE_PROTOCOL:  return stor_string;
    case PROXY_PROTOCOL:    return prox_string;
    default:                return NULL;
    }
}

nestProtocolType
nestProtocolManager::protocolType( const char *name )
{
	static struct { const char	*label; nestProtocolType type; }
	table[] = {
		{ "nest", NEST_PROTOCOL },
		{ "chirp", NEST_PROTOCOL },
		{ "http", HTTP_PROTOCOL },
		{ "proxy", PROXY_PROTOCOL },
		{ "storage", STORAGE_PROTOCOL },
		{ "ADVERTISE", ADVERTISE_PROTOCOL },
		{ NULL, (nestProtocolType) -1 },
	};

	for( int i=0;  table[i].label;  i++ ) {
		if ( !strcasecmp( table[i].label, name ) ) {
			return table[i].type;
		}
	}
	return (nestProtocolType) -1;
}

int
nestProtocolManager::getProtocolFromSocket( int protocol_socket )
{
    if ( protocol_socket == sockets[NEST_PROTOCOL] ) 
        return NEST_PROTOCOL;
    if ( protocol_socket == sockets[HTTP_PROTOCOL] )
        return HTTP_PROTOCOL; 
    if ( protocol_socket == sockets[PROXY_PROTOCOL] )
        return PROXY_PROTOCOL;

        // we better not make it here
    assert( 0 );
    return -1;
}

int
nestProtocolManager::firstProtocolSocket( ) {
    protocol_index = -1;
    return nextProtocolSocket();
}

int
nestProtocolManager::lastProtocolSocket( ) {
    return LAST_PROTOCOL_SOCKET;
}

int
nestProtocolManager::nextProtocolSocket( ) {
    protocol_index++;
    if ( protocol_index == PROTOCOL_COUNT ) {
        return LAST_PROTOCOL_SOCKET;
    } else {    
        return sockets[protocol_index];
    }
}

void
nestProtocolManager::setMaxsocket( ) {
    for( int i = 0; i < PROTOCOL_COUNT; i++ ) {
        if ( sockets[i] > maxsocket ) maxsocket = sockets[i];
    }
}    

int
nestProtocolManager::maxSocket( ) {
    if ( maxsocket <= 0 ) {
        nest_die( "Unable to listen on any ports.\n"
				  "\tPerhaps a NeST is already running?\n" );
    }
    return maxsocket;
}    
