/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#ifndef _NEST_PROTOCOL_MANAGER_H
#define _NEST_PROTOCOL_MANAGER_H

#include "nest_common.h"
#include "nest_client.h"

#define   FTP_SERVER_PORT   21
#define  HTTP_SERVER_PORT   80
#define  GFTP_SERVER_PORT 2811
#define  NEST_SERVER_PORT 7878
#define PROXY_SERVER_PORT    0  // off by default.  Turn on for proxies.

class
nestProtocolManager
{

  public:

	nestProtocolManager( void );
	~nestProtocolManager( void );

	int init( );

    nestProtocol* getProtocolByType  ( int protocol_type   );
    nestProtocol* getProtocolBySocket( int protocol_socket );

    NestReplyStatus initiateThirdPartyTransfer(
		nestClient *client, 
		nestProtocol **connection,
		nestProtocolType protocol );

    nestClient* accept(  int protocol_socket );
    
    nestProtocol* newConnection( int csd, int protocol );

    NestReplyStatus SendConnection( int sd, nestProtocol *connection );

    nestProtocol *ReceiveConnection( int sd, int protocol );
    
    int getProtocolFromSocket( int protocol_socket ); 
    int getProtocolPortFromSocket( int protocol_socket );
    char * getProtocolNameFromSocket( int protocol_socket ); 

    static char * protocolName( int protocol );
    static nestProtocolType protocolType( const char *name );

    NestReplyStatus getCredentialInfo( char **subject, long *expires );

    bool blockingTransfer( nestProtocol *protocol,
						   int type,
						   int fd, 
						   long len );

    int firstProtocolSocket( );
    int  lastProtocolSocket( );
    int  nextProtocolSocket( );
    
    int  NestsetSocket( int domain,
						int type,
						int protocol, 
						int my_protocol);

    void setMaxsocket( ); 
    int  maxSocket( );

    void getCredential( );

  private:
    void reportOpenPort( char *protocol, int port, int status );

	// array for mapping sockets to their protocols
	int           sockets    [PROTOCOL_COUNT];

	int maxsocket;

	int  NestPort;
	int  HttpPort;
	int   FtpPort;
	int  GftpPort;
	int ProxyPort;

	int protocol_index;

};

extern nestProtocolManager	*protocol_manager;

#endif
