/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "nest_client.h"
#include "nest_client_pool.h"

#include "nest_common.h"
#include "nest_speak.h"
#include "nest_protocol_nest.h"
#include "globus_gss_assist.h"
#include "nest_globus.h"
#include "nest_advertiser.h"
#include "nest_main.h"


// static class variables
// why can't these be in the header file ???
static unsigned addr_sz;

int
nestProtocolNest::listen( int port )
{
    
        // now actually listen on the socket
    addr_sz = sizeof( struct sockaddr_in );
    return NestsetSocket( AF_INET, SOCK_STREAM, 0, port );
}

nestProtocol *
nestProtocolNest::accept( int socket, sockaddr_in *address )
{
    int csd =  NestAccept( socket, (sockaddr *)address, &addr_sz );
    return ( csd <= 0 ) ? NULL : new nestProtocolNest( csd );
}

NestReplyStatus
nestProtocolNest::initiateThirdParty( nestClient *client,
									  nestProtocol **connection )
{
    NestConnection socket;
    NestReplyStatus status;

        // open connection
    nest_debug(NEST_DEBUG_CHIRP,
			   "Attempting to connect to chirp server at %s\n",
            client->getRequest()->getRemoteHost() );
    status = NestOpenConnection( socket,
								 client->getRequest()->getRemoteHost());
    if ( status != NEST_SUCCESS ) return status;

        // attempt to move connection into transfer-ready state
    if ( client->getRequest()->getType() == THIRD_PARTY_PUT ) {
        status = RequestWriteFile(
                client->getRequest()->getRemoteFile(),
                client->getRequest()->getStart(),
                client->getRequest()->getLength(), 
                socket );
    } else {
        assert( client->getRequest()->getType() == THIRD_PARTY_GET );
        nestFileSize length = client->getRequest()->getLength();
        status = RequestReadFile( 
			client->getRequest()->getRemoteFile(),
			client->getRequest()->getStart(), 
			length,
			socket );
        client->getRequest()->setEnd(
			length - client->getRequest()->getStart() );
    }
    if ( status != NEST_SUCCESS ) {
        nest_debug( NEST_DEBUG_CHIRP, 
                "Connection to %s made but transfer request denied: %s\n",
                client->getRequest()->getRemoteHost(),
                NestErrorString( status ) );
        NestCloseConnection( socket );
        return status;
    }

    *connection = new nestProtocolNest( socket );
    return status;
}

nestProtocolNest::nestProtocolNest( int socket )
{
    this->socket    = socket;
    this->msgoffset = 0;
}

nestProtocol *
nestProtocolNest::clone( )
{
    return new nestProtocolNest( NestDup( this->socket ) );
}

nestProtocolNest::~nestProtocolNest()
{
    nest_debug( NEST_DEBUG_CHIRP, "Closing chirp socket: %d.\n", socket );
    NestClose( socket );
}

NestReplyStatus
nestProtocolNest::gssAccept( char **username , char **globus_username )
{
    return auth_globus_accept( socket, username, globus_username );
}

bool
nestProtocolNest::endThirdPartyTransfer( int type )
{
    if ( type == THIRD_PARTY_GET ) {
            // nothing to do
        return true;
    } else {
        assert( type == THIRD_PARTY_PUT );
        nest_debug( NEST_DEBUG_CHIRP, 
                    "nestProtocolNest ending 3pt - waiting for ack\n" );
        return ( ReceiveNestAck( socket ) == NEST_SUCCESS );
    }
}
    
ssize_t
nestProtocolNest::Send( const char *buffer, ssize_t n )
{
    return NestWrite( socket, buffer, n );
}

ssize_t
nestProtocolNest::Receive( char *buffer, ssize_t n )
{
    return NestRead( socket, buffer, n );
}

bool
nestProtocolNest::sendAd( nestAd *ad )
{

    char *adstring  = ad->toString();
    int len         = ad->length();

    NestSnprintf( msg, sizeof(msg), "%d %ld", NEST_SUCCESS, len    );
    nest_debug( NEST_DEBUG_CHIRP, "Sending ad len %ld (%.40s).\n", 
				len, msg );
    if ( ! sendMessage( (void *)msg ) ) return false;

    ssize_t sent = NestWriten( socket, (void *)adstring, len );
    return ( sent == len );

    /*
    while( sent < len ) {
        thissend = ( len - sent > NEST_MESSAGESIZE )
                    ? NEST_MESSAGESIZE
                    : len - sent;

        NestSnprintf( msg, sizeof(msg), "%s", &(adstring[sent]) );
        sent += thissend;
        if ( ! sendMessage( (void *)msg ) ) return false;
    }
    return ( sent == len );

    nest_debug( NEST_DEBUG_CHIRP, "sending ack following send nestad\n" );
    return sendAck( );
    */
}

bool
nestProtocolNest::sendDirectoryList( NestFileInfo *list )
{
    while( list != NULL ) {
        NestFileInfo_to_string( list, msg, sizeof(msg) );
        if ( ! sendMessage( (void *)msg ) ) return false;
        list = list->next;
    }
    return sendAck( );
}

#define PROXY_ITERATOR list<char*>::iterator

bool
nestProtocolNest::sendProxyList( list <char *> *proxys )
{
    char *glist;
    long length = 0;

    if ( proxys != NULL ) {
            // go through twice:  first to determine length, then to send
        for(PROXY_ITERATOR itr = proxys->begin(); itr != proxys->end(); itr++){
            char *name = (char *)*itr;
            length += NestStrlen( name ) + 1;
        }   
    }

        // send the length
    if ( length > 0 ) length += 2;  // add room for null term
    NestSnprintf( msg, sizeof(msg), "%d %ld\n", NEST_SUCCESS, length );
    nest_debug( NEST_DEBUG_PROXY, "Sending proxy list len %ld (%.40s).\n", 
            length, msg );
    if ( ! sendMessage( (void *)msg ) ) return false;

        // cut out early if NULL
    if ( proxys == NULL || proxys->size() <= 0 ) return true;

        // otherwise write the list
    glist = (char *)NestCalloc( length );
    long offset = 0;
    for( PROXY_ITERATOR itr = proxys->begin(); itr != proxys->end(); itr++ ) {
        char *name = (char *)*itr;
        NestSnprintf( &glist[offset], length + 2 - offset, "%s ", name );
        offset += NestStrlen( name ) + 1;
        nest_debug( NEST_DEBUG_PROXY, "Sending proxy %s.\n", name );
    }
    NestSnprintf( &glist[offset], length - offset, "\n\0" );

        // send the list
    ssize_t sent = NestWriten( socket, (void *)glist, length );
    NestFree( glist );
    if (sent != length ) {
        if ( errno != EPIPE) nest_perror( "NestWriten" );
        return false;
    }

    return true;
}

bool
nestProtocolNest::sendNestLot( NestLotInternal *r)
{

    if (!r) {
        sendErrorMessage( __FILE__, __FUNCTION__,  NEST_UNKNOWN_ERROR );
        return false;
    }
    NestSnprintf(
		msg,
		sizeof(msg),
		"%d %s %s \"%s\" %u %s %lld %lld %lld %d %d %d %d %d\%s",
		NEST_LOT_SCHEDULED,
		r->lot_handle,
		r->parent_lot_handle,
		r->owner->getName(),
		r->owner->isUser(),
		r->dev,
		r->total_size,   
		r->size,
		r->used,
		r->start_time,
		r->duration,
		r->time_left,
		r->cost,
		r->lot_type,
		r->lot_users );

    return sendMessage( (void *) msg);
}

bool
nestProtocolNest::sendLotHandle( NestLotHandle *handle )
{

    rsv_id_t    tmp_id;

    if (NestGet_id_from_handle(handle, &tmp_id) < 0) {
        sendErrorMessage( __FILE__, __FUNCTION__,  NEST_UNKNOWN_ERROR );
        return false;
    }
    nest_debug(NEST_DEBUG_CHIRP, "sendLotHandle: handle=%p, id=%d \n",
			   handle, tmp_id );

    NestSnprintf(msg, sizeof(msg), "%d %lu",
                                NEST_LOT_SCHEDULED,
                                tmp_id);

    return sendMessage( (void *) msg);
}

bool
nestProtocolNest::sendDiskspace(  nestLotSize avail, nestLotSize total )
{
    NestSnprintf( msg, sizeof(msg), "%d %lld %lld",
				  NEST_SUCCESS, avail, total );
    return sendMessage( (void *) msg );
}

bool
nestProtocolNest::sendErrorMessage( const char* filename,
									const char* func_name,
									int error )
{
    NestSnprintf( msg, sizeof( msg ), "%d", error );
    return sendMessage( msg );
}

bool
nestProtocolNest::sendTransferInfo(  nestRequest *req )
{
    NestSnprintf( msg,
                  NEST_MESSAGESIZE,
                  "%d %s %d %ld",
                  NEST_SUCCESS,
                  req->getNestFile(),
                  req->getType(),
                  req->getEnd() );
    return sendMessage( msg );
}

bool
nestProtocolNest::sendTimestamp(  ){
    char buff[NEST_MESSAGESIZE];
    time_t ticks = time( NULL );
    NestSnprintf( buff, sizeof(buff), "%d %ld", NEST_SUCCESS, (long)ticks );
    return sendMessage( buff );
}

#if 0
bool
nestProtocolNest::sendQuotaInfo(  struct UserUsage *usage )
{
    NestSnprintf( msg, NEST_MESSAGESIZE, "%d %ld %ld\n",
            NEST_SUCCESS, usage->current, usage->quota );
    return sendMessage( msg );
}
#endif

bool
nestProtocolNest::sendFilesize( nestFileSize size)
{
    char msg[NEST_MESSAGESIZE];
    NestSnprintf( msg, NEST_MESSAGESIZE, "%d %lld", NEST_SUCCESS, size );
    return sendMessage( msg );
}

NestReplyStatus
nestProtocolNest::sendFileinfo( const NestFileInfo *info) {
    char msg[NEST_MESSAGESIZE];
    char tmp[NEST_MESSAGESIZE];
	nest_debug( NEST_DEBUG_ALWAYS, "Info: %d %d '%s' %ld '%s'\n",
				info->type, info->uniquifier, info->name, info->size,
				info->LotList );
	NestFileInfo_to_string( info, tmp, sizeof(tmp) );
	NestSnprintf( msg, NEST_MESSAGESIZE, "%d %s", NEST_SUCCESS, tmp );
	nest_debug( NEST_DEBUG_ALWAYS, "-> '%s'\n", msg );
	return sendMessage( (void *)msg );
}

bool
nestProtocolNest::sendPhysicalName( const char *name)
{
    char msg[NEST_MESSAGESIZE];
    NestSnprintf( msg, NEST_MESSAGESIZE, "%d %s", NEST_SUCCESS, name );
    return sendMessage( msg );
}

bool
nestProtocolNest::sendPwd( const char *pwd )
{
    char msg[NEST_MESSAGESIZE];
    NestSnprintf( msg, sizeof(msg), "%d %s", NEST_SUCCESS, pwd );
    return sendMessage( msg );
}

bool
nestProtocolNest::sendAcl( const char *user,
                           int is_user,
                           nest_access_rights_t access)
{
    char msg[NEST_MESSAGESIZE];

    // Format the return string..
    nest_debug( NEST_DEBUG_CHIRP, "Sending access=0x%02x for user '%s'\n",
                access, user );

    // Send the initial success...
    if ( ! sendAck( ) ) {
        nest_debug( NEST_DEBUG_ALWAYS, 
                    "sendACL: Error sending initial ACK\n" );
        return false;
    }

    // Now the data message
    NestSnprintf( msg, NEST_MESSAGESIZE, "%d %s %d %d",
                  NEST_PARTIAL_MESSAGE, user, is_user, access );

    // Ship it out
    if ( ! sendMessage( msg ) ) {
        return false;
    }
    return sendAck( );
}

bool
nestProtocolNest::sendAcl( const NestACL *Acl)
{
    char msg[NEST_MESSAGESIZE];

    // Send the initial success...
    if ( ! sendAck( ) ) {
        nest_debug( NEST_DEBUG_ALWAYS, 
                    "sendACL: Error sending initial ACK\n" );
        return false;
    }

    // Send the whole list we have...
    nest_debug( NEST_DEBUG_CHIRP, "Sending ACL@%p\n", Acl );
    while( Acl != NULL ) {
        nest_debug( NEST_DEBUG_CHIRP, "Sending access=0x%02x for user '%s'\n",
                    Acl->acl.rights, Acl->acl.name );
        NestSnprintf( msg, NEST_MESSAGESIZE, "%d %s %d %d",
                      NEST_PARTIAL_MESSAGE,
                      Acl->acl.name, Acl->acl.is_user, Acl->acl.rights );
        if ( ! sendMessage( (void *)msg ) ) {
            nest_debug( NEST_DEBUG_ALWAYS, "sendACL: Error sending ACL\n" );
            return false;
        }
        Acl = Acl->next;
    }
    return sendAck( );
}

bool
nestProtocolNest::sendVersion( const char *version )
{
    char msg[NEST_MESSAGESIZE];
    nest_debug( NEST_DEBUG_CHIRP, "Sending version %s\n", version );
    NestSnprintf( msg, sizeof(msg), "%d \"%s\"", NEST_SUCCESS, version );
    return sendMessage( msg );
}

bool
nestProtocolNest::sendMessage(  const void *vptr )
{
    return SendNestMessage( socket, (char *) vptr ); 
}

bool
nestProtocolNest::sendFinalAck( int transferType )
{
    nest_debug( NEST_DEBUG_CHIRP,
            "Entering nestProtocolNest::sendFinalAck for type %d: sd %d\n",
            transferType, socket );

    if (   transferType == CLIENT_RECV_FILE 
        || transferType == PROXY_GET 
        || transferType == PROXY_PUT 
        ) 
    {
        /* NO-OP */ 
        return true;
    } else {
        return sendAck( );
    }
}

bool
nestProtocolNest::sendAck(  )
{
    char ack[NEST_MESSAGESIZE];
    nest_debug( NEST_DEBUG_CHIRP,
            "Sending chirp ack on sd %d\n", socket );
    NestSnprintf( ack, sizeof(ack), "%d", NEST_SUCCESS );
    return sendMessage( ack );
}

nestRequest*
nestProtocolNest::receiveRequest( void )
{

    if ( ! NestFdIsSet( socket, FD_READ ) ) {
        //nest_debug( NEST_DEBUG_CHIRP, "NestSocket %d not ready\n", socket );
        return NULL;
    }

    NestReplyStatus status;
    int read = 0;

    status = ReceiveNestPartialMessage( socket, &(msg[msgoffset]),
										NEST_MESSAGESIZE - msgoffset, &read );

    if ( msgoffset != 0 ) {
        nest_debug( NEST_DEBUG_CHIRP, 
					"Attempted to resume partial req (sd %d, off %d). "
					"Got %d additional bytes (%s): %s",
					getSocket(), msgoffset, read, 
					NestErrorString( status ), 
					msg );
    }

    msgoffset = ( msgoffset + read ) % NEST_MESSAGESIZE;
    switch( status ) {
    case NEST_SUCCESS:
        assert( msgoffset == 0 );
        break;
    case NEST_PARTIAL_MESSAGE:
        assert( msgoffset > 0 && msgoffset < NEST_MESSAGESIZE );
        msg[msgoffset] = '\0';
        nest_debug( NEST_DEBUG_CHIRP, 
					"WARN: nestProtocol (sock %d) recvd partial msg"
					" (len %d): \"%s\"\n",
					getSocket(), msgoffset, msg );
        return NULL;
    default:
        if ( status != NEST_NO_CONNECTION ) {
            nest_debug( NEST_DEBUG_CHIRP,
                "WARN: nestProtocol failed to recv request: %s\n",
                NestErrorString( status ) );
        }
            // clients don't need to do anything explicit to close
            // this is gotten for free.  When they close, their
            // fd appears as ready to read but when we read from it
            // we get an error - nice
        return new nestRequest( CLOSE_CONNECTION_REQUEST );
    }

	// Remove the end-of-line from the message
	int		len = strlen( msg );
	if (  ( len > 0 ) && ( msg[len-1] == '\n' ) ) {
		msg[len-1] = '\0';
	}
    nest_debug( NEST_DEBUG_CHIRP, "Received chirp req: '%s'\n", msg );
    nestRequest *req = parseRequest( msg );
    if ( req == NULL ) {
        nest_debug( NEST_DEBUG_CHIRP,
					"Bad request received on sd %d: \"%s\"\n",
					getSocket(), msg );
		req = new nestRequest( BAD_REQUEST, NEST_BAD_PARAMETERS );
    }
    return req;
}

bool
valid_userstr( const char *str )
{
	if ( NULL == str ) return false;
	if ( *str == '\0' ) return false;
	if ( ! strcmp( str, "0" ) ) return false;
	return true;
}

nestRequest*
nestProtocolNest::parseRequest( char *message )
{

        // leave these static so that NestParseLine can free memory later
        // however this does require that this is always done by one thread
    char **argv = NULL;
    int argc;

    bool valid_request = true;

    NestParseLine(message, &argv, &argc);
    if( !argv ) return NULL;
    if( !argv[0] ) return NULL;

    int req_type = atoi(argv[0]);
	nestRequest		*req = new nestRequest( req_type );

    switch( req_type ) {

		// These take no parameters
    case DIR_LIST_REQUEST:
    case AVAILABLE_SPACE_QUERY:
    case TIME_REQUEST:
    case SERVER_TERMINATE_REQUEST:
    case FORMAT_DISK_REQUEST:
    case PWD_REQUEST:
    case NEST_SYNC:
    case REQUEST_NEST_AD:
    case VERSION_REQUEST:
    case PROXY_QUERY:
    case PROXY_INDEX_PUSH:
    case PROXY_FLUSH:
        break;

    case CLIENT_SEND_FILE:
    case CLIENT_RECV_FILE:
	{
		if(!argv[1] || !argv[2] || !argv[3]) {
			valid_request = false;
			break;
		}

		req->setFile( argv[1] );

		// TODO  this is really bad, but end is supposed to be the
		// actual end not the size, it should really to fixed to be the
		// size.  for the time being, just hack it here
		long start = atol(argv[2]);
		long end = atol(argv[3]);
		if ( end != ENTIRE_FILE ) {
			end += start;
		}
		req->setStart( start );
		req->setEnd( end );
		break;
	}

    case TEST_CLIENT_TO_SERVER:
    case TEST_SERVER_TO_CLIENT:
		req->setEnd( argv[1] );
        break;

    case FILE_DELETE_REQUEST:
    case FILESIZE_REQUEST:
	case FILEINFO_REQUEST:
		req->setFile( argv[1] );
		break;

    case RMDIR_REQUEST:
    case CWD_REQUEST:
    case MKDIR_REQUEST:
		req->setDir( argv[1] );
		break;

    case PROXY_REGISTER:
		req->setProxyLocation( argv[1] );
        break;

    case FILE_RENAME_REQUEST:
		req->setFile( argv[1] );
		req->setDestFile( argv[2] );
		break;

    case CHANGE_USER_REQUEST:
		req->setUser( argv[1] );
		req->setAuthType( argv[2] );
		break;

    case THIRD_PARTY_PUT:
    case THIRD_PARTY_GET:
	{
		req->setFile( argv[1] );
		req->setRemoteHost( argv[2] );
		req->setRemoteFile( argv[3] );
		req->setStart( argv[4] );
		req->setEnd( argv[5] );

		nestProtocolType	ptype = (nestProtocolType) -1;
		if ( argv[6] ) {
			ptype = nestProtocolManager::protocolType( argv[6] );
		}
		req->setRemoteProtocol( (int) ptype );
        break;
	}

    case ADD_NEST_USER:
	{
#	  if 0
		nestUser	user( argv[3], argv[1] );
#	  else
		nestUser	user( argv[3] );
		user.setOsName( argv[1] );
#	  endif
		req->setUser( &user );
		nestGroup	group( argv[2] );
		req->setGroup( &group );
        break;
	}


    case ADD_USER_TO_GROUP:
    case DELETE_USER_FROM_GROUP:
		req->setUser( argv[1] );
		req->setGroup( argv[2] );
        break;

    case ADD_NEST_GROUP:
    case DELETE_NEST_GROUP:
		req->setGroup( argv[1] );
        break;

    case DELETE_NEST_USER:
		req->setUser( argv[1] );
        break;

		// NRL
    case LIST_GROUP_LOTS:
		req->setGroup( argv[1] );
		break;

    case LIST_USER_LOTS:
		if ( valid_userstr( argv[1] ) ) {
			req->setUser( argv[1] );
		}
        break;

    case GET_LOT_STAT:
    case TERMINATE_LOT:
		req->setLotHandle( argv[1] );
		break;

    case REQUEST_USER_LOT:
    case REQUEST_GROUP_LOT:
        if(!argv[1] || !argv[2] || !argv[3]) {
            valid_request = false;
            break;
        }

		// Parameter 1: Lot size
		req->setLotSize( argv[1] );

		// Parameter 2: Lot duration
		req->setLotDuration( argv[2] );

		// Parameter 3: Lot type
		req->setLotType( argv[3] );

		// Parameter 4: Lot user / group
		if ( valid_userstr( argv[4] ) ) {
			if( REQUEST_USER_LOT == req_type ) {
				req->setUser( argv[4] );
			} else {
				req->setGroup( argv[4] );
			}
		}

		// Parameter 5: Lot parent user
		if ( valid_userstr( argv[5] ) ) {
			req->setLotParentUser( argv[5] );
		}

		// Parameter 6: Lot parent group
		if ( valid_userstr( argv[6] ) ) {
			req->setLotParentGroup( argv[6] );
		}

		// Parameter 7: Parent lot
		if ( valid_userstr( argv[7] ) ) {
			req->setLotParentLot( argv[7] );
		}
        break;

    case UPDATE_LOT:
		req->setLotHandle( argv[1] );
		req->setLotSize( argv[2] );
		req->setLotDuration( argv[3] );
		break;

    case CHANGE_LOT_T:
		req->setLotHandle( argv[1] );
		req->setLotType( argv[2] );
		req->setLotDuration( argv[3] );
        break;

    case LOT_MOVE_FILE:
		req->setFile( argv[1] );
		req->setLotSourceHandle( argv[2] );
		req->setLotDestHandle( argv[3] );
		req->setLotBytesToMove( argv[4] );
        break;

    case ADD_USER_TO_LOT:
    case DELETE_USER_FROM_LOT:
		req->setLotHandle( argv[1] );
		if ( valid_userstr( argv[2] ) ) {
			req->setUser( argv[2] );
		}
        break;

    case ATTACH_LOT_TO_FILE:
    case DETACH_LOT_FROM_FILE:
		req->setLotHandle( argv[1] );
		req->setFile( argv[2] );
        break;

        // Collection ops
    case UNNAME_COLLECTION:
    case NAME_COLLECTION:
		req->setCollectionName( argv[1] );
		req->setFile( argv[2] );
        break;

    case SET_COLLECTION_MTIME:
    case SET_COLLECTION_VERSION:
    case SET_COLLECTION_COMMENT:
	{
        if(!argv[1] || !argv[2]) {
            valid_request = false;
            break;
        }

		// Fill it in....
		req->setCollectionName( argv[1] );

		// Save 'arg2' to the appropriate field
		if ( SET_COLLECTION_MTIME == req_type ) {
			req->setCollectionMtime( argv[2] );
		} else {
			// NOTE: This modifies argv[2]
			int		len = strlen( argv[2] );
			char	*end = argv[2] + len - 1;
			char	*buf;
			if ( ( len > 0 ) && ( *argv[2] == '\"' ) && ( *end == '\"') ) {
				buf = argv[2]+1;
				buf[len - 2] = '\0';
			} else {
				buf = argv[2];
			}

			if ( SET_COLLECTION_VERSION == req_type ) {
				req->setCollectionVersion( buf );
			} else {
				req->setCollectionComment( buf );
			}
		}
        break;
	}

    case COPY_REQUEST:
		req->setFile( argv[1] );
		req->setDestFile( argv[2] );
		req->setStart( (nestFileSize) 0 );
		req->setEnd( (nestFileSize) ENTIRE_FILE );
        break;

    case LINK_REQUEST:
		req->setFile( argv[1] );
		req->setDestFile( argv[2] );
		req->setIsSymLink( argv[3] );
        break;

    case PHYSNAME_REQUEST:
		req->setFile( argv[1] );
        break;

        // ACL Related
    case GET_ACL:
    case GET_GROUP_ACL:
		req->setFile( argv[1] );
		if ( valid_userstr( argv[6] ) ) {
			if ( GET_GROUP_ACL == req_type ) {
				req->setGroup( argv[2] );
			} else {
				req->setUser( argv[2] );
			}
		}
        break;

    case SET_ACL:
    case SET_GROUP_ACL:
		req->setFile( argv[1] );
		if ( valid_userstr( argv[2] ) ) {
			if ( GET_GROUP_ACL == req_type ) {
				req->setGroup( argv[2] );
			} else {
				req->setUser( argv[2] );
			}
		}
		req->setRights( argv[3] );
        break;

    case DEBUG_REQUEST:
        req->setDebugString( argv[1] );
        break;

        // we longer have a bandwidth statistics request
    case BANDWIDTH_STATISTICS_REQUEST:
    default:
        nest_debug( NEST_DEBUG_CHIRP,
					"ERROR:  Unknown req type rec'vd: %d.  Full msg: %s\n",
					req_type, message );
		delete req;
        req = new nestRequest(BAD_REQUEST, NEST_UNKNOWN_REQUEST_TYPE );
		break;
    }

	// Validate it
	if ( ! req->validate( ) ) {
		nest_debug( NEST_DEBUG_CHIRP,
					"Request '%d' failed vaidation\n", req_type );
		delete req;
		req = NULL;
	}

	NestFreeArgv( argv, argc );
    return req;
}
