/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#ifndef    _NEST_QUOTA_MANAGER_H_

#define _NEST_QUOTA_MANAGER_H_

#include <sys/stat.h>
#include <sys/time.h>
#include <stdlib.h>
#include <string.h>

#include "nest_common.h"
#include "nest_protocol.h"
#include "nest_user.h"

#include "nest_quota_instance.h"
#include "nest_device.h"

#define  NUM_QUOTA_INSTANCES  1

#include "fs_ext2_quota_instance.h"
#define  FS_EXT2              0

typedef struct qimap
{
    char            dev[NEST_NAMELEN];
    int             qi_index;
    struct qimap    *next;
} QuotaInstanceMap;


class QuotaManager
{

  public:

	QuotaManager();
	~QuotaManager();


	NestReplyStatus incr_quota(const nestUserOrGroup *name,
							   const char *dev,
							   nestLotSize more_size);

	NestReplyStatus decr_quota(const nestUserOrGroup *name,
							   const char *dev,
							   nestLotSize less_size);

	NestReplyStatus initialize(ClassAdCollection *resources,
							   const NestDevice *disk,
							   nestLotSize *user_usage);

	NestReplyStatus recover(ClassAdCollection *resources,
							const NestDevice *disk);

	NestReplyStatus shutdown();
	NestReplyStatus synchronize();

	static NestReplyStatus has_kernel_support(const NestDevice *dev);


  private:

	ClassAdCollection       *quota_catalog;

	ViewName                quota_catalog_view;
	ClassAdParser           parser;
	ClassAdUnParser         unparser;
	LocalCollectionQuery    query;

	/*
	  QuotaInstanceMap - maps a device to the respective
	  QuotaInstance or Quota-Enforcement.

	  quota_inst_map is implemented a linked-list and
	  locate_quota_inst() is used to search thru it.
	  eg:{
	       {"/dev/hda5", FS_EXT2},
	       {"/dev/sdb6", FS_EXT2},
	       {"/mem",      MEM}, ...
	  }
	*/

	QuotaInstanceMap    *quota_inst_map;
	QuotaInstance       *quota_inst[NUM_QUOTA_INSTANCES];

	int locate_quota_inst(const char *dev);

	NestReplyStatus add_entity(const nestUserOrGroup *name,
							   const char *dev,
							   nestLotSize size,
							   int inode_count);

	NestReplyStatus delete_entity(const nestUserOrGroup *name,
								  const char *dev);

	NestReplyStatus update_entity(const nestUserOrGroup *name,
								  const char *dev,
								  nestLotSize new_size,
								  int new_inode_count);

	NestReplyStatus view_quota_status(const nestUserOrGroup *name,
									  const char *dev,
									  nestLotSize *size,
									  int *inode_count);

	NestReplyStatus view_quota_usage(const nestUserOrGroup *name,
									 const NestDevice *dev,
									 nestLotSize *size,
									 int *inode_count);
};

#endif
