/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#ifndef NEST_REQUEST_HANDLER_H
#define NEST_REQUEST_HANDLER_H

#include "nest_error.h"
#include "nest_client.h"
#include "nest_hash_table.h"

class nestRequestHandler
{
  public:
	nestRequestHandler( void );
	~nestRequestHandler( void );

	int  HandleRequests(
		nestClientPool *requests,
		fd_set *readfds
		);
	nestRequest *ReceiveRequest(
		nestClient *client
		);
	int TransferFile(
		nestClient *,
		nestRequest *,
		bool already_trans = false
		);
	void setAllowedAuth(
		int		types
		);
	void taskDone(
		nestClient *client,
		NestReplyStatus status
		);


  private:
	int  HandleRequest(
		nestClient *client
		);
	NestReplyStatus NewClient(
		nestClient *newclient
		);
	bool RemoveClient (
		nestClient *client,
		fd_set *reads
		);
	int  SynchronousRequestHandler(
		nestClient *client,
		nestRequest *
		);
	NestReplyStatus SendNestAd(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus openStorageConnections(
		nestProtocol **storage,
		nestProtocol **otherStorage,
		nestClient   *client
		);
	NestReplyStatus openStorageConnection(
		nestProtocol **storage,
		nestClient *client,
		const nestRequest *request
		);
	NestReplyStatus SendDirectoryList(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus SendPwd(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus MakeLink(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus CollectionOp(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus DirOp(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus SendVersion(
		nestClient *,
		nestRequest *
		);
	int SendBandwidthStatistics(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus CopyFile (
		nestClient *,
		nestRequest *
		);
	NestReplyStatus DeleteFile (
		nestClient *,
		nestRequest *
		);
	NestReplyStatus RenameFile (
		nestClient *,
		nestRequest *
		);
	NestReplyStatus SendFileinfo(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus GetFilehandle(
		nestClient *client,
		nestRequest *
		);
	NestReplyStatus SetFileAttr(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus SendFilesize(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus SendAvailableSpace(
		nestClient *,
		nestRequest *
		);
	int NestSync(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus QueryProxys (
		nestClient *,
		nestRequest *
		);
	NestReplyStatus RegisterProxyClient (
		nestClient *,
		nestRequest *
		);
	NestReplyStatus RemoveProxyClient (
		nestClient *
		);
	NestReplyStatus FormatDisk (
		nestClient *,
		nestRequest *
		);
	NestReplyStatus AuthenticateUser(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus AddUser(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus AddGroup(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus DeleteUser(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus DeleteGroup(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus AddUserToGroup(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus DeleteUserFromGroup(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus ReserveUserSpace(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus ReserveGroupSpace(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus ListUserLots(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus ListGroupLots(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus GetLotStat(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus UpdateLot(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus ChangeLotType(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus MoveFileLot(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus AddUserToLot(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus DeleteUserFromLot(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus AttachLotToFile(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus DetachLotFromFile(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus SendPhysName(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus SendUserAcl(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus SetUserAcl(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus TerminateLot(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus DebugRequest(
		nestClient *,
		nestRequest *
		);
	NestReplyStatus TerminateServer(
		nestClient *,
		nestRequest *
		);

	void getAuthString(
		int types,
		string &str
		);

	    // for a main nest to know which proxys there are
	    // this hash table is an stl list for each client registering proxies
	struct nest_hash_table *registered_proxys;

	    // acceptable authentication models
	int allowedAuthentication;

};

extern nestRequestHandler	*req_handler;

#endif
