/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#ifndef NEST_SHARED_H
#define NEST_SHARED_H

#include <sys/types.h>
#include "nest_error.h"

#ifdef USE_CLASSAD_NAMESPACE
# define WANT_NAMESPACES
# include "classad_distribution.h"
using namespace classad;
#else
# include "classad_distribution.h"
#endif

// Simple bits about how to contact the server
#define NEST_SERVER_PORT 7878

// TODO: we should try changing this at some point to 1448
// many packets get split at this boundary and may be causing
// extra TCP roundtrips...
#define NEST_MESSAGESIZE	2048
#define NEST_MAXPATHLEN     (NEST_MESSAGESIZE/4)
#define NEST_NAMELEN        (NEST_MESSAGESIZE/8)

// Basic size types
typedef long long nestLotSize;
typedef long long nestFileSize;

// Dirent type
#define        NEST_FILE            1
#define        NEST_DIR             2
#define        NEST_LINK            3

// File information
typedef         unsigned char   nest_obj_type_t;
typedef         int             NestFilehandle;
typedef struct NestFileInfo_s {
    nest_obj_type_t type;
    NestFilehandle uniquifier;
    char name[NEST_MAXPATHLEN];
    long size;
    char LotList[NEST_MAXPATHLEN];
	int mtime;
	int ctime;
	int atime;
    ClassAd *classad;
    struct NestFileInfo_s *next;
} NestFileInfo;

// Lot definitions
#define G_LOT 0
#define BE_LOT 1
typedef long unsigned int    rsv_id_t;

typedef struct NestLotHandle_s {
    rsv_id_t    rsv_id;
} NestLotHandle;

// *** ACL Stuff ***

// ACE == Access Control Entry
typedef short int nest_access_rights_t;
typedef struct NestACE_s {
    char    name[NEST_MAXPATHLEN];
    int     is_user;
    nest_access_rights_t rights;
} NestACE;

// ACL == Access Control List
typedef struct NestACL_s {
    NestACE     		acl;
    struct NestACL_s	*next;
} NestACL;
typedef struct NestACLList NestACLList;

// Transfer info
typedef struct NestTransferInfo_s {
	struct timeval starttime;
	struct timeval lastprogress;
	bool read;
	nestFileSize transferred;
} NestTransferInfo;

typedef int NestConnection;


// *************************************************************
// Prototypes for functions shared by the clients and server
// *************************************************************


//  time utilities
void NestGettimespec( struct timespec *ts );
void NestGettimeofday( struct timeval *time );
time_t NestTime(time_t *tsecs);
long NestTimediff( const struct timeval *start,
				   const struct timeval *end, 
				   const char *function );
long NestElapsedTime( const struct timeval *start );

	// File utiltities
nestFileSize NestFilelength( const char *path );
bool NestIsFile( const char *path );
bool NestIsDir ( const char *path );
bool NestIsLink ( const char *path );
bool NestReadLink( const char *path, char *linkto, int length );
char *NestExtractFilename( char *pathname );


	// Bandwidth measurement
double NestBandwidth( nestFileSize bytes,
					  long millis );
double NestBandwidth( nestFileSize bytes,
					  struct timeval *start,
					  struct timeval *end,
					  const char *function );

	// Memory definitions / functions
#define     NestCalloc(x)    NestCallocDebug(  x, __FILE__, __LINE__ )
#define     NestMalloc(x)    NestMallocDebug(  x, __FILE__, __LINE__ )
#define     NestRealloc(x,y) NestReallocDebug( x, y, __FILE__, __LINE__ )
#define     NestFree(x)      NestFreeDebug(    x, __FILE__, __LINE__ )
#define     NestStrdup(x)    NestStrdupDebug(  x, __FILE__, __LINE__ )
void* NestCallocDebug( size_t size, char *func, int line );
void* NestCallocNormal( size_t size );
void* NestMallocDebug( size_t size, char *func, int line );
void* NestMallocNormal( size_t size );
void* NestReallocNormal(void *ptr, size_t size);
void* NestReallocDebug(void *ptr, size_t size, char *func, int line);
void  NestFreeDebug( void *ptr, char *func, int line );
void  NestFreeNormal( void *ptr );
char * NestStrdupNormal( const char * );
char * NestStrdupDebug( const char *, char *func, int line );
char* NestMakeString(int initial_size, const char *fmt, ...);
void *NestMemcpy( void *dst, const void *src, size_t n );
void *NestMemset( void *dst, int c, size_t n );
size_t NestStrlen( const char * );
int NestSnprintf( char *buf, int len, const char *fmt, ... );

	// NeST specific protos
NestTransferInfo * NestmakeTransferInfo( );
void NestParseLine(char *line, char ***argv, int *argc);
void NestFreeArgv( char **argv, int argc );

    // NeST lot stuff protos
int     NestAdd_id_to_handle(NestLotHandle *handle, rsv_id_t rsv_id);
int     NestGet_id_from_handle(const NestLotHandle *handle, rsv_id_t *rsv_id);
int     NestHandle_to_string(const NestLotHandle *handle, char *str, int len);
int     NestString_to_handle(NestLotHandle *handle, const char *str);

// "Nest File Info" operators
int NestFileInfo_to_string(const NestFileInfo *file, char *msg, int len);
int NestString_to_NestFileInfo(NestFileInfo *file, const char *msg);

// Simple string parsers & formatters
NestReplyStatus
NestParseSizeString( const char *arg,
					 long long *size,
					 char default_mod = 'B' );
NestReplyStatus
NestParseTimeString( const char *arg,
					 int *seconds,
					 char default_mod = 'S' );
const char *
NestFormatSizeString( long long size,
					  char *buf,
					  int bufsize,
					  bool raw = false);

const char *
NestFormatTimeString( int seconds,
					  char *buf,
					  int bufsize,
					  bool raw );

#endif
