/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/
#include "nest_simple_storage.h"
#include "nest_config.h"

#define RET_ERRNO(X,Y)  return ( (X) < 0 ) ? ErrnoToNest((Y)) : NEST_SUCCESS; 

nestStorageManager::nestStorageManager( void )
{

	storagedir = NULL;
	collection_manager = NULL;
	cache = NULL;

	read_locks  = NULL; 
	write_locks = NULL;
}

nestStorageManager::~nestStorageManager( void )
{
}

    //  Initialize  Storage Manager
int  
nestStorageManager::init( )
{
        // takes no args
    long storagespace = -1;

    const char *value;
    if (GetConfigValue(Config_StorageDir, &value) && *value != 0) {
        storagedir = NestStrdup( value );
    } else nest_die( "Must specify storage dir\n" );
    
    if (GetConfigValue(Config_StorageSpace, &value) && *value != 0 ) {
        storagespace = atol(value) * 1024;
    }

    if ( NestMkdir( storagedir, DIR_MODE ) < 0 ) {
        if ( errno != EEXIST ) {
            nest_die( "mkdir %s: %s\n", storagedir, strerror(errno));
        }
    }
    NestChdir( storagedir );

        // set to NULL, get absolute path
    NestReplyStatus status;
    char buffer[NEST_MAXPATHLEN];
    if ( ( status = NestGetcwd( buffer, NEST_MAXPATHLEN ) ) != NEST_SUCCESS ) {
        nest_die( "Couldn't read cwd: %s\n", NestErrorString( status ) );
    } else {
        storagedir = NestStrdup( buffer );
    }

        // set up our locks hash table
    read_locks  = nest_hash_table_create( 1037, NULL );
    write_locks = nest_hash_table_create( 1037, NULL );
    assert( read_locks != NULL && write_locks != NULL );

        // set up our filecache
        // query from filesystem if not user specified
    if ( storagespace < 0 ) {
        long long int free;
        long long int total;
        nest_fsstat fs_buf;

        if (NestStatFS(".", &fs_buf) < 0) {
            nest_warn( __FILE__, __FUNCTION__, 
                    "statfs failed in bytesAvailable(): %s\n", 
                    strerror( errno ) );
            return NEST_UNKNOWN_ERROR;
        }

        total = free = (fs_buf.f_bfree * fs_buf.f_bsize);
        storagespace = free / 1024;
    }
    cache = new nestFilecacheModel( 
            (float)storagespace / 1024,
            NEST_CACHE_CLOCK, // clock allows giving priority weights
            removeCacheEntry,
            true );         // track hit info

    nest_debug( NEST_DEBUG_STORAGE | NEST_DEBUG_PROXY, 
            "Storage dir is %s.\n", storagedir );
    return 0;
}

NestReplyStatus  
nestStorageManager::initialize() {
    return NEST_NOT_YET_IMPLEMENTED;
}

void
nestStorageManager::writeAd( nestProtocol *dest ) {
    char buffer[NEST_MESSAGESIZE];

    long long int capacity, total;
    if ( bytesAvailable(IS_USER, &capacity, &total ) 
            == NEST_SUCCESS )
    {
        NestSnprintf( buffer, NEST_MESSAGESIZE, "Disk = %lld\nFree = %lld\n", 
                total, capacity);
        dest->Sendn( buffer, NestStrlen( buffer ) );
    }

        // now lots?
    if(is_lots_enabled())
        NestSnprintf( buffer, NEST_MESSAGESIZE, "LotsEnabled = true\n");
    else
        NestSnprintf( buffer, NEST_MESSAGESIZE, "LotsEnabled = false\n");
    dest->Sendn( buffer, NestStrlen( buffer ) );

    cache->writeAd( dest );
}

NestReplyStatus  
nestStorageManager::sync() {
    NestSync();
    return NEST_SUCCESS;
}
    
NestReplyStatus 
nestStorageManager::cleanupGftpTransfer( nestClient *client ) {
    return NEST_SUCCESS;
}


NestReplyStatus
nestStorageManager::renameFile( const char *user, 
                                user_id_t  uid,
                                const char *oldpath,
                                const char *newpath,
                                const char *cwd,
                                bool unused = false ) 
{
    NestReplyStatus ret;
    string fulloldpath;
    string fullnewpath;

    ret = expand_path( fulloldpath, user, cwd, oldpath );
    if(ret!=NEST_SUCCESS) return ret;

    ret = expand_path( fullnewpath, user, cwd, newpath );
    if(ret!=NEST_SUCCESS) return ret;

    if(NestRename( fulloldpath.c_str(), fullnewpath.c_str() )<0) {
        nest_warn( __FILE__, __FUNCTION__, "Unable to move %s to %s: %s\n",
                   oldpath, newpath, strerror( errno ) );
        return ErrnoToNest(errno);
    } else {
        nest_debug( NEST_DEBUG_PROXY, "Renamed %s to %s: %d\n",
                    oldpath, newpath, ret );
        cache->renameFile( fulloldpath.c_str(), fullnewpath.c_str());
        return NEST_SUCCESS;
    }
}

NestReplyStatus  
nestStorageManager::maintenance() {
        // nothing to do
    return NEST_SUCCESS;
}

NestReplyStatus  
nestStorageManager::shutdown() {
        // nothing to do
    return NEST_SUCCESS;
}

    //  Meta-file operations
nestProtocol*    
nestStorageManager::openMetaFile(char *file, int flags ) {
        // make sure we're in the base dir
    NestChdir( storagedir );
    
        // open the file
    int fd = NestOpen( file, flags );
    //NestChmod( file, FILE_MODE );
    if ( fd < 0 ) {
        return NULL;
    } else {
        nestProtocolFilesystem *connection = new nestProtocolFilesystem( fd );
        string path;
        NestReplyStatus status = expand_path( path, "proxy", storagedir, file );
        assert( status == NEST_SUCCESS );
        connection->set_filenames( (char *)path.c_str(), NULL );
        lockFile( read_locks, path.c_str() );
        return (nestProtocol*)connection;
    }
}

NestReplyStatus  
nestStorageManager::renameMetaFile(char *oldpath,
                                           char *newpath ) {
    return NEST_NOT_YET_IMPLEMENTED;
}

NestReplyStatus  
nestStorageManager::statMetaFile(char *file,
                                         nest_stat *st ) {
    return NEST_NOT_YET_IMPLEMENTED;
}

NestReplyStatus  
nestStorageManager::unlinkMetaFile( char *file ) {
    return NEST_NOT_YET_IMPLEMENTED;
}

char*            
nestStorageManager::metaFilepath(char *file ) {
        // TODO this is wrong
    return file; 
}

long             
nestStorageManager::metaFilelength( char *file ) {
    long len = NestFilelength( file );
    return len;
}

NestReplyStatus
nestStorageManager::expand_path( 
        string &result, const char *user, const char *cwd, const char *file ) 
{
	char longpath[NEST_MAXPATHLEN];
	char shortpath[NEST_MAXPATHLEN];

    if ( file[0] == '/' ) {
            // absolute already
        if ( strlen(file) == 1 ) {
            // parrot hack here.  replace root with storagedir
            strcpy( longpath, storagedir );
        } else {
            strcpy( longpath, file );
        }
    } else {
            // relative needs full path affixed
        sprintf( longpath, "%s/%s", cwd, file );
    }

	NestCollapsePath(longpath,shortpath);
	result = shortpath;

	return NEST_SUCCESS;
}

    //  File operations
NestReplyStatus  
nestStorageManager::open(nestProtocol **result,
                                 const char *user,
                                 user_id_t  uid,
                                 const char *group,
                                 const char *file,
                                 int flags,
                                 off_t offset,
                                 ssize_t filesize,
                                 const char *cwd,
                                 int link_count = 0,
                                 int weight = 1) 
{
    NestReplyStatus status;

        // first make sure it fits in our cache
        // one potential problem is for files that previously existed on a 
        // system.  if read, they will be inserted into cache and take up
        // space that was already allocated to them
        // but this would stray outside of expected behavior of a proxy
        // cache which is expected to just start running under a condor
        // starter for example on a new empty system
    string path;
    if ( NEST_SUCCESS != expand_path( path, "proxy", cwd, file ) ) {
        nest_die( "Could not expand %s and %s.\n", cwd, file );
    }
    status = addCacheEntry( path.c_str(), filesize, weight );
    if ( status != NEST_SUCCESS ) return status;

    if ( flags & O_CREAT ) {
        char *cpath = NestStrdup(path.c_str());
        char *lastslash = strrchr(cpath,'/');
        if(lastslash) {
            *lastslash = 0;
            status = NestCreateFullPath(cpath);
        } else {
            status = NEST_SUCCESS;
        }
        NestFree(cpath);
        if ( status != NEST_SUCCESS ) return status;
    }

    int fd = NestOpen( path.c_str(), flags );
    if ( fd < 0 ) {
        *result = NULL;
        status = ErrnoToNest( errno ); 
    } else {
            // successful open, create connection, try to set size, insert in $
		nestProtocolFilesystem *con = 
            new nestProtocolFilesystem( fd, ((off_t)offset) );


            // get filesize and path for cache insert
        if ( filesize < 0 ) {
            filesize = NestFilelength( path.c_str() );
            if ( filesize < 0 ) filesize = 0;
        }
        status = NEST_SUCCESS;

            // lock the file so it can't be deleted, remember path for unlock
        lockFile( read_locks, path.c_str() );
        con->set_filenames( (char *)path.c_str(), NULL );
		*result = (nestProtocol*)con;
    }
    return status;
}

NestReplyStatus
nestStorageManager::addCacheEntry( const char *file, ssize_t filesize, 
        int weight ) 
{
    NestReplyStatus status;
    bool hit = false;
    status = cache->insert( &hit, file, (long)filesize, weight );
    if ( ! hit ) {
        nest_debug( NEST_DEBUG_PROXY, "Adding %s (%ld) to cache: %s.\n",
            file, filesize, NestErrorString( status ) );
    }
    return status;
}

NestReplyStatus
nestStorageManager::removeCacheEntry( const char *file ) {
    NestReplyStatus status = deleteFile( "proxy", file, storagedir, false );
        // can't print file since it gets freed in deleteFile
    nest_debug( NEST_DEBUG_PROXY, "Attempted to remove file from cache: %s.\n",
            /*file,*/ NestErrorString( status ) );

    return status;
}

char *           
nestStorageManager::filehandleToPath( NestFilehandle * ) {
    return NULL;
}
    
NestReplyStatus  
nestStorageManager::pathToFilehandle(const char *user,
                                            const char *group,
                                            const char *file,
                                            const char *cwd,
                                            NestFilehandle   *inode_no,
                                            int   link_count = 0) 
{
    return NEST_NOT_YET_IMPLEMENTED;
}

int
nestStorageManager::checkLocks( nest_hash_table *ht, const char *path ) {
    int *locks = (int*)nest_hash_table_lookup( ht, path );
    return ( !locks ) ? 0 : *locks;
}

NestReplyStatus
nestStorageManager::lockFile( nest_hash_table *ht, const char *path ) {
    NestReplyStatus status = NEST_SUCCESS;
    int *locks = NULL;

    void *entry = nest_hash_table_lookup( ht, path );
    if ( entry == NULL ) {
            // make one
        nest_debug( NEST_DEBUG_FLOCK, "Making lock entry for %s.\n", path );
        locks = (int *)NestMalloc( sizeof( int ) ); // this leaks?
        *locks = 1;
        nest_hash_table_insert( ht, path, (void *)locks );
    } else {
        locks = (int *)entry;
        (*locks)++;
    }
    
    nest_debug( NEST_DEBUG_FLOCK, "Locking file %s: %d\n", path, *locks );
    return status;
}

NestReplyStatus
nestStorageManager::unlockFile( nest_hash_table *ht, const char *path ) {
    NestReplyStatus status = NEST_SUCCESS;
    int* remaining;

    void *entry = nest_hash_table_lookup( ht, path );
    assert( entry != NULL );

    remaining = (int *)entry;
    (*remaining)--;
    assert( *remaining >= 0 );

    nest_debug( NEST_DEBUG_FLOCK, "Unlocking file %s: %d\n", path, *remaining );
    return status;
}

NestReplyStatus
nestStorageManager::rememberLength( const char *user, const char *cwd,
        const char *file, long *length ) 
{
    string path;
    if ( NEST_SUCCESS != expand_path( path, "proxy", cwd, file ) ) {
        nest_die( "Could not expand %s and %s.\n", cwd, file );
    }

    bool remembered = cache->rememberLength( path.c_str(), length );
    return ( remembered ) ?  NEST_SUCCESS : NEST_FILE_NOT_FOUND;
}


NestReplyStatus  
nestStorageManager::close(nestProtocol *protocol,
                          nestClient *client = NULL ) {

        // just do the unlock here, when connection is deleted then close it
	if ( protocol->getType() != STORAGE_PROTOCOL ) {
		nest_die( "Weirdness in %s:%s\n", __FILE__, __FUNCTION__ );
	}
	nestProtocolFilesystem *connection = (nestProtocolFilesystem*)protocol;
    char path[NEST_MAXPATHLEN];
    connection->get_filenames( path, NULL );
    unlockFile( read_locks, path );

    return NEST_SUCCESS;
}

NestReplyStatus  
nestStorageManager::deleteLink(const char *filename) {
    return NEST_NOT_YET_IMPLEMENTED;
}

NestReplyStatus  
nestStorageManager::deleteDirContents(const char *user, const char *filename) {
    return NEST_NOT_YET_IMPLEMENTED;
}

NestReplyStatus  
nestStorageManager::deleteFile(        const char *user,
                                       const char *filename,
                                       const char *cwd,
                                       bool inform_cache = true) 
{
    string path;
    NestReplyStatus status = expand_path( path, user, cwd, filename );
    if(status!=NEST_SUCCESS) return status;

    nest_debug( NEST_DEBUG_PROXY, "Storage removing %s.\n", path.c_str() );

        // check to make sure this file isn't locked
    int locks = checkLocks( read_locks,  path.c_str() )
              + checkLocks( write_locks, path.c_str() );
    if ( locks > 0 ) {
        nest_debug( NEST_DEBUG_PROXY, "Cant remove %s (%d locks).\n", 
                path.c_str(), locks );
        return NEST_FLOCK;
    }
    
    int ret = NestUnlink( path.c_str() );
    
        // make sure our cache reclaims this space
        // don't do this however if this delete call actually originated with
        // the cache
    if ( inform_cache && ret==0 ) {
        cache->remove( path.c_str() );   
    }
        
    RET_ERRNO( ret, errno );
}

NestReplyStatus  
nestStorageManager::get_type( nest_obj_type_t   *result,
                                      const char *filename,
                                      const char *user,
                                      const char *cwd ) 
{
    // the NAME_COLLECTION request relies on this...
    return NEST_SUCCESS;
}

NestReplyStatus  
nestStorageManager::get_physical_filename(char **physical,
                                                  const char *filename,
                                                  const char *user,
                                                  const char *cwd,
                                                  int link_count = 0) 
{
    return NEST_NOT_YET_IMPLEMENTED;
}

NestReplyStatus  
nestStorageManager::Link(const char *user,
                                 const char *oldpath,
                                 const char *newpath,
                                 int val,
                                 const char *cwd) 
{
    return NEST_NOT_YET_IMPLEMENTED;
}

NestReplyStatus  
nestStorageManager::sync(const char *user) 
{
    NestSync();
    return NEST_SUCCESS;
}


    //  Directory operations

NestReplyStatus 
nestStorageManager::frame_full_path( string &result,
                                            const char *user,
                                            const char *file ) 
{
    return NEST_NOT_YET_IMPLEMENTED;
}

NestReplyStatus 
nestStorageManager::get_dest_dir( char *dest,
                                         const char *cwd,
                                         const char *dir,
										 const char *user ) 
{
    return NEST_NOT_YET_IMPLEMENTED;
}

NestReplyStatus 
nestStorageManager::chDir(char **new_cwd,
                                 const char *user,
                                 const char *cwd,
                                 const char *dir) 
{
    string fullpath;
    NestReplyStatus r;

    r = expand_path(fullpath,user,cwd,dir);
    if(r!=NEST_SUCCESS) return r;

    if(NestIsDir(fullpath.c_str())) {
        *new_cwd = NestStrdup(fullpath.c_str());
        return NEST_SUCCESS;
    } else {
        return ErrnoToNest(errno);
    }
}

NestReplyStatus 
nestStorageManager::mkDir(const char *user,
                                 const char *cwd,
                                 const char *dir ) 
{
    NestReplyStatus r;
    string fullpath;

    r = expand_path(fullpath,user,cwd,dir);
    if(r!=NEST_SUCCESS) return r;

    if(NestMkdir( fullpath.c_str(), DIR_MODE )==0) {
        return NEST_SUCCESS;
    } else {
        return ErrnoToNest(errno);
    }
}

char *
nestStorageManager::getInitialClientDir() {
    return storagedir;
    //return "/";
}

NestReplyStatus 
nestStorageManager::rmDir(const char *user,
                                 const char *cwd,
                                 const char *dir,
                                 bool recursive = false ) 
{
    NestReplyStatus r;
    string fullpath;

    r = expand_path(fullpath,user,cwd,dir);
    if(r!=NEST_SUCCESS) return r;

    if(NestRmdir( fullpath.c_str() )==0) {
        return NEST_SUCCESS;
    } else {
        return ErrnoToNest(errno);
    }
}

char*            
nestStorageManager::getRootDir( ) {
    return "."; 
}

NestFileInfo*    
nestStorageManager::directoryContents(const char *user, const char *dirname ) {

	NestReplyStatus r;
	NestFileInfo *result=0, *info;
	DIR *dir;
	struct dirent *d;

	dir = opendir(dirname);
	if(dir) {
		while( (d=readdir(dir)) ) {
			if(!strcmp(d->d_name,".")) continue;
			if(!strcmp(d->d_name,"..")) continue;
			info = (NestFileInfo*)NestMalloc( sizeof( *result ) );
			r = nestStorageManager::fileinfo(info,d->d_name,user,dirname);
			if(r==NEST_SUCCESS) {
				info->next = result;
				result = info;
			} else {
				delete info;
			}
		}
		closedir(dir);
	}

	return result;
}


    //  User/Group Management operations
NestReplyStatus  
nestStorageManager::add_user(const char *username,
                             const char *groupname,
                             const char *cur_user,
                             const char *identity) 
{
    return NEST_NOT_YET_IMPLEMENTED;
}

NestReplyStatus  
nestStorageManager::add_user(const char *username,
                                     const char *cur_user) 
{
    return NEST_NOT_YET_IMPLEMENTED;
}

NestReplyStatus
nestStorageManager::change_lot_type(NestLotHandle   *handle,
                                    const char *user,
                                    int lot_type,
                                    int duration) 
{
    return NEST_NOT_YET_IMPLEMENTED;
}
    
NestReplyStatus 
nestStorageManager::add_user_to_lot(NestLotHandle *handle,
                                           const char *cur_user,
                                           const char *user)
{
    return NEST_NOT_YET_IMPLEMENTED;
}

NestReplyStatus 
nestStorageManager::delete_user_from_lot(NestLotHandle *handle,
                                           const char *cur_user,
                                           const char *user)
{
    return NEST_NOT_YET_IMPLEMENTED;
}

NestReplyStatus
nestStorageManager::attach_lot_to_file(NestLotHandle *handle,
                                        const char *file,
                                        const char *user,
                                        const char *cwd)
{
    return NEST_NOT_YET_IMPLEMENTED;
}

NestReplyStatus
nestStorageManager::detach_lot_from_file(NestLotHandle *handle,
                                        const char *file,
                                        const char *user,
                                        const char *cwd)
{
    return NEST_NOT_YET_IMPLEMENTED;
}
    
    
NestReplyStatus  
nestStorageManager::delete_user(const char *username,
                                        const char *cur_user) 
{
    return NEST_NOT_YET_IMPLEMENTED;
}

NestReplyStatus  
nestStorageManager::add_group(const char *groupname, const char *cur_user) {
    return NEST_NOT_YET_IMPLEMENTED;
}
NestReplyStatus  
nestStorageManager::delete_group(const char *groupname, const char *cur_user) {
    return NEST_NOT_YET_IMPLEMENTED;
}

NestReplyStatus  
nestStorageManager::add_user_to_group(const char *username,
                                              const char *groupname,
                                              const char *cur_user) 
{
    return NEST_NOT_YET_IMPLEMENTED;
}
NestReplyStatus  
nestStorageManager::delete_user_from_group(const char *username,
                                                   const char *groupname,
                                                   const char *cur_user) 
{
    return NEST_NOT_YET_IMPLEMENTED;
}

    //  Access control operations
NestReplyStatus  
nestStorageManager::get_acl(nest_access_rights_t *result,
                                    const char *for_name,
                                    int is_user,
                                    const char *file,
                                    const char *user,
                                    const char *cwd) 
{
    return NEST_NOT_YET_IMPLEMENTED;
}

NestReplyStatus  
nestStorageManager::get_all_acl(NestACL **result,
                                        const char *file,
                                        const char *user,
                                        const char *cwd) 
{
    return NEST_NOT_YET_IMPLEMENTED;
}

NestReplyStatus  
nestStorageManager::set_acl(nest_access_rights_t arg,
                                    const char *to_name,
                                    int is_user,
                                    const char *file,
                                    const char *user,
                                    const char *cwd) 
{
    return NEST_NOT_YET_IMPLEMENTED;
}

NestReplyStatus  
nestStorageManager::get_user_acl(nest_access_rights_t *result,
                                         const char *for_user,
                                         const char *file,
                                         const char *user,
                                         const char *cwd) 
{
    return NEST_NOT_YET_IMPLEMENTED;
}

NestReplyStatus  
nestStorageManager::set_user_acl(nest_access_rights_t arg,
                                         const char *to_user,
                                         const char *file,
                                         const char *user,
                                         const char *cwd) 
{
    return NEST_NOT_YET_IMPLEMENTED;
}

NestReplyStatus  
nestStorageManager::get_group_acl(nest_access_rights_t *result,
                                          const char *for_group,
                                          const char *file,
                                          const char *user,
                                          const char *cwd) 
{
    return NEST_NOT_YET_IMPLEMENTED;
}

NestReplyStatus  
nestStorageManager::set_group_acl(nest_access_rights_t arg,
                                          const char *to_group,
                                          const char *file,
                                          const char *user,
                                          const char *cwd) 
{
    return NEST_NOT_YET_IMPLEMENTED;
}

    //  Lot operations
NestReplyStatus 
nestStorageManager::request_user_lot(NestLotHandle *handle,
                                            const char *user,
                                            nestLotSize size,
                                            int duration,
                                            const char *lot_user,
                                            const char *parent_user,
                                            const char *parent_group,
                                            NestLotHandle *parent_handle,
                                            int lot_type=G_LOT,
                                            int cost=1)
{
    return NEST_NOT_YET_IMPLEMENTED;
}

NestReplyStatus 
nestStorageManager::request_group_lot(NestLotHandle *handle,
                                             const char *user,
                                             nestLotSize size,
                                             int duration,
                                             const char *lot_user,
                                             const char *parent_user,
                                             const char *parent_group,
                                             NestLotHandle *parent_handle,
                                             int lot_type=G_LOT,
                                             int cost=1) 
{
    return NEST_NOT_YET_IMPLEMENTED;
}
    
NestReplyStatus 
nestStorageManager::increaseAllocation(nestClient *client) {
    return NEST_SUCCESS;
}

NestReplyStatus 
nestStorageManager::list_user_lots(NestLotListInternal **result,
								   const char *user,
								   const char *lot_user ) 
{
    return NEST_NOT_YET_IMPLEMENTED;
}

NestReplyStatus 
nestStorageManager::list_group_lots(NestLotListInternal **result,
									const char *req_user,
									const char *group ) 
{
    return NEST_NOT_YET_IMPLEMENTED;
}

NestReplyStatus 
nestStorageManager::update_lot(NestLotHandle *handle,
                                      const char *user,
                                      nestLotSize size,
                                      int duration,
                                      int cost = 0) 
{
    return NEST_NOT_YET_IMPLEMENTED;
}

NestReplyStatus 
nestStorageManager::modify_lot_usage(const char *user,
                                            const char *filename,
                                            const char *cwd,
                                            NestLotHandle  *from_lot,
                                            nestLotSize from_share,
                                            NestLotHandle  *to_lot) 
{
    return NEST_NOT_YET_IMPLEMENTED;
}

NestReplyStatus 
nestStorageManager::get_lot_status(NestLotInternal* result,
                                   NestLotHandle *handle,
                                   const char *user)
{
    return NEST_NOT_YET_IMPLEMENTED;
}

NestReplyStatus 
nestStorageManager::set_user_default_lot( NestLotHandle *handle,
                                                 const char *user ) 
{
    return NEST_NOT_YET_IMPLEMENTED;
}

NestReplyStatus 
nestStorageManager::terminate_lot(NestLotHandle *handle,
                                         const char *user) 
{
    return NEST_NOT_YET_IMPLEMENTED;
}

    //  Other operations
NestReplyStatus
nestStorageManager::bytesAvailable( int is_user,
                                    long long int *free,
                                    long long int *total ) 
{
    *free =  (long long int)cache->getavail() * 1024;
    *total = (long long int)cache->getsize() * 1024; 
    return NEST_SUCCESS;
}

NestReplyStatus  
nestStorageManager::filelength(     long int *result,
                                    const char *filename,
                                    const char *user,
                                    const char *cwd )
{
    NestReplyStatus r;
    string path;

    r = expand_path(path,user,cwd,filename);
    if(r!=NEST_SUCCESS) return r;

    if(NestIsDir(path.c_str())) {
        *result = -1;
        return NEST_SUCCESS;
    } else {
        *result = NestFilelength(path.c_str());
        if(*result<0) {
            return NEST_FILE_NOT_FOUND;
        } else {
            return NEST_SUCCESS;
        }
    }
}

NestReplyStatus
nestStorageManager::fileinfo( NestFileInfo *info,
                              const char *filename,
                              const char *user,
                              const char *cwd,
                              int link_count = 0 )
{
	NestReplyStatus r;
	nest_stat buf;
	string fullpath;

	r = expand_path(fullpath,user,cwd,filename);
	if(r!=NEST_SUCCESS) return r;

	if(NestStat(fullpath.c_str(),&buf)==0) {
		r = NEST_SUCCESS;
		if(S_ISREG(buf.st_mode)) {
			info->type = NEST_FILE;
		} else if(S_ISDIR(buf.st_mode)) {
			info->type = NEST_DIR;
		} else if(S_ISLNK(buf.st_mode)) {
			info->type = NEST_LINK;
		} else {
			r = NEST_NOT_A_FILE;
		}
		strcpy(info->name,fullpath.c_str());
		info->uniquifier = buf.st_ino;
		info->size = buf.st_size;
		info->LotList[0] = 0;
		info->classad = 0;
		info->next = 0;
	} else {
		r = ErrnoToNest(errno);
	}

	return r;
}

char *
nestStorageManager::type( ) {
    return "Proxy";
}

bool
nestStorageManager::is_lots_enabled( ) {
    return false;
}

bool             
nestStorageManager::is_valid_user(const char *user, user_id_t *uid ) {
    return true;
}

NestReplyStatus  
nestStorageManager::formatDisk(const char *user,
                                       const char *group ) 
{
    return cache->format(); 
}

