/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/
#ifndef _NEST_STORAGE_MANAGER_H
#define _NEST_STORAGE_MANAGER_H

#include "nest_common.h"
#include "nest_hash_table.h"
#include "nest_protocol_storage.h"
#include "nest_protocol_filesystem.h"

class
nestStorageManager
{

  public:

	// Constructors & destructors
	nestStorageManager( void );
	virtual ~nestStorageManager( void );

    int  init( );

    NestReplyStatus  initialize();

    NestReplyStatus  sync();

    void writeAd( nestProtocol *);
    
    NestReplyStatus  maintenance();
    NestReplyStatus  shutdown();

    NestReplyStatus  renameFile( const char *user, 
                                        user_id_t  uid,
                                        const char *old, 
                                        const char *newpath,
                                        const char *cwd,
                                        bool unused = false );

    NestReplyStatus cleanupGftpTransfer( nestClient *client );

    //  Meta-file operations
    nestProtocol*    openMetaFile(char *file,
                                         int flags );

    NestReplyStatus  renameMetaFile(char *oldpath,
                                           char *newpath );

    NestReplyStatus  statMetaFile(char *file,
                                         nest_stat *st );

    NestReplyStatus  unlinkMetaFile( char *file );

    char*            metaFilepath(char *file );

    long             metaFilelength( char *file );

    //  File operations
    NestReplyStatus  open(nestProtocol **result,
                                 const char *user,
                                 user_id_t  uid,
                                 const char *group,
                                 const char *file,
                                 int flags,
                                 off_t offset,
                                 ssize_t filesize,
                                 const char *cwd,
								 int link_count = 0,
                                 int weight = 1);

    char *           filehandleToPath( NestFilehandle * );
    
    NestReplyStatus  pathToFilehandle(const char *user,
                                            const char *group,
                                            const char *file,
                                            const char *cwd,
                                            NestFilehandle   *inode_no,
                                            int   link_count = 0);



    NestReplyStatus  close(nestProtocol *connection,
                                  nestClient *client = NULL);

    NestReplyStatus  deleteLink(const char *filename);

    NestReplyStatus  deleteDirContents(const char *user,
                                              const char *filename);

    NestReplyStatus  deleteFile(const char *user,
                                       const char *filename,
                                       const char *cwd,
                                       bool update_lot = true);
    /*
    NestReplyStatus  renameFile( const char *user,
                                        const char *oldpath,
                                        const char *newpath,
                                        const char *cwd,
                                        bool update_lot = true );
    */
    NestReplyStatus  get_type( nest_obj_type_t   *result,
                                      const char *filename,
                                      const char *user,
                                      const char *cwd );

    NestReplyStatus  get_physical_filename(char **physical,
                                                  const char *filename,
                                                  const char *user,
                                                  const char *cwd,
                                                  int link_count = 0);

    NestReplyStatus  Link(const char *user,
                                 const char *oldpath,
                                 const char *newpath,
                                 int val,
                                 const char *cwd);

    NestReplyStatus  sync(const char *user);


    //  Directory operations

    NestReplyStatus frame_full_path( string &result,
                                            const char *user,
                                            const char *file );

    NestReplyStatus expand_path( string &result,
                                        const char *user,
                                        const char *cwd,
                                        const char *file );


    NestReplyStatus get_dest_dir( char *dest,
                                         const char *cwd,
                                         const char *dir,
										 const char *user );

    NestReplyStatus chDir(char **new_cwd,
                                 const char *user,
                                 const char *cwd,
                                 const char *dir);

    NestReplyStatus mkDir(const char *user,
                                 const char *cwd,
                                 const char *dir );

    NestReplyStatus rmDir(const char *user,
                                 const char *cwd,
                                 const char *dir,
                                 bool recursive = false );

    char*            getRootDir( );

    NestFileInfo*    directoryContents(const char *user,
                                              const char *dir );


    //  User/Group Management operations
    NestReplyStatus  add_user(const char *username,
                                     const char *groupname,
                                     const char *cur_user,
                                     const char *identity);

    NestReplyStatus  add_user(const char *username,
                                     const char *cur_user);

    NestReplyStatus  delete_user(const char *username,
                                        const char *cur_user);

    NestReplyStatus  add_group(const char *groupname,
                                      const char *cur_user);
    NestReplyStatus  delete_group(const char *groupname,
                                         const char *cur_user);

    NestReplyStatus  add_user_to_group(const char *username,
                                              const char *groupname,
                                              const char *cur_user);
    NestReplyStatus  delete_user_from_group(const char *username,
                                                   const char *groupname,
                                                   const char *cur_user);

    //  Access control operations
    NestReplyStatus  get_acl(nest_access_rights_t *result,
                                    const char *for_name,
                                    int is_user,
                                    const char *file,
                                    const char *user,
                                    const char *cwd);

    NestReplyStatus  get_all_acl(NestACL **result,
								 const char *file,
								 const char *user,
								 const char *cwd);

    NestReplyStatus  set_acl(nest_access_rights_t arg,
                                    const char *to_name,
                                    int is_user,
                                    const char *file,
                                    const char *user,
                                    const char *cwd);

    NestReplyStatus  get_user_acl(nest_access_rights_t *result,
                                         const char *for_user,
                                         const char *file,
                                         const char *user,
                                         const char *cwd);

    NestReplyStatus  set_user_acl(nest_access_rights_t arg,
                                         const char *to_user,
                                         const char *file,
                                         const char *user,
                                         const char *cwd);

    NestReplyStatus  get_group_acl(nest_access_rights_t *result,
                                          const char *for_group,
                                          const char *file,
                                          const char *user,
                                          const char *cwd);

    NestReplyStatus  set_group_acl(nest_access_rights_t arg,
                                          const char *to_group,
                                          const char *file,
                                          const char *user,
                                          const char *cwd);

    //  Lot operations
    NestReplyStatus request_user_lot(NestLotHandle *handle,
                                            const char *user,
                                            nestLotSize size,
                                            int duration,
                                            const char *lot_user,
                                            const char *parent_user,
                                            const char *parent_group,
                                            NestLotHandle *parent_handle,
                                            int lot_type=G_LOT,
                                            int cost=1);

    NestReplyStatus request_group_lot(NestLotHandle *handle,
                                             const char *user,
                                             nestLotSize size,
                                             int duration,
                                             const char *lot_group,
                                             const char *parent_user,
                                             const char *parent_group,
                                             NestLotHandle *parent_handle,
                                             int lot_type=G_LOT,
                                             int cost=1);

    NestReplyStatus list_user_lots(NestLotList **result,
                                          const char *user,
                                          const char *lot_user );

    NestReplyStatus list_group_lots(NestLotList **result,
                                           const char *req_user,
                                           const char *group );

    NestReplyStatus update_lot(NestLotHandle *handle,
                                      const char *user,
                                      nestLotSize size,
                                      int duration,
                                      int cost = 0);

    NestReplyStatus change_lot_type(NestLotHandle *handle,
                                           const char *user,
                                           int lot_type,
                                           int duration);

    NestReplyStatus modify_lot_usage(const char *user,
                                            const char *filename,
                                            const char *cwd,
                                            NestLotHandle  *from_lot,
                                            nestLotSize from_share,
                                            NestLotHandle  *to_lot);

    NestReplyStatus get_lot_status(NestLotInternal* result,
								   NestLotHandle *handle,
								   const char *user);

    NestReplyStatus set_user_default_lot( NestLotHandle *handle,
                                                 const char *user );

    NestReplyStatus terminate_lot(NestLotHandle *handle,
                                         const char *user);

    NestReplyStatus add_user_to_lot(NestLotHandle *handle,
                                           const char *cur_user,
                                           const char *user);

    NestReplyStatus delete_user_from_lot(NestLotHandle *handle,
                                                const char *cur_user,
                                                const char *user);

    NestReplyStatus attach_lot_to_file(NestLotHandle *handle,
                                              const char *file,
                                              const char *user,
                                              const char *cwd);

    NestReplyStatus detach_lot_from_file(NestLotHandle *handle,
                                                const char *file,
                                                const char *user,
                                                const char *cwd);

    nestLotSize available_lot_space(const char *user,
									const char *logical_file);

    nestLotSize available_file_lot_space( const char *logical_file );

    nestLotSize available_dir_lot_space( const char *logical_file );

    nestLotSize available_user_lot_space( const char *user );

    nestLotSize available_group_lot_space( const char *user );

    NestReplyStatus increaseAllocation(nestClient *client);
    NestReplyStatus resetAllocation(const char *file,
                                           const char *user,
                                           const char *cwd,
                                           long offset);

    //  Proxy operations
    NestReplyStatus addCacheEntry( 
			const char *file, ssize_t filesize, int weight );

    NestReplyStatus removeCacheEntry( const char *file );
    

    //  Other operations
    NestReplyStatus  bytesAvailable( int is_user,
                                            long long int *free,
                                            long long int *total );

    NestReplyStatus  filelength(  long *filesize,  
                                         const char *filename,
                                         const char *user,
                                         const char *cwd );

    NestReplyStatus  fileinfo(  NestFileInfo *info,
                                       const char *filename,
                                       const char *user,
                                       const char *cwd,
                                       int link_count = 0 );
    char *           getInitialClientDir( );

    bool             is_valid_user(const char *user, user_id_t *uid );

    NestReplyStatus  formatDisk(const char *user,
                                       const char *group );

    NestReplyStatus  rememberLength( 
            const char *user, const char *cwd,
            const char *file, long *filesize );

    char *           type();
    
    bool             is_lots_enabled(); 

  private:

    NestReplyStatus  lockFile( nest_hash_table *ht, const char *file );

    NestReplyStatus  unlockFile( nest_hash_table *ht, const char *file );

    NestReplyStatus  ensure_path( const char *cwd, const char *file );
    
    int checkLocks( nest_hash_table *ht, const char *file );


	char *storagedir;
	nestFilecacheModel *cache;

	struct nest_hash_table *read_locks; 
	struct nest_hash_table *write_locks;


};

#endif
