/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#ifndef NEST_SPEAK_H
#define NEST_SPEAK_H

#include <sys/types.h>

#include "nest_globus.h"
#include "nest_shared.h"
#include "nest_error.h"

// Declarations *only* for the external API
typedef struct {
  public:
    char                    lot_handle[NEST_NAMELEN];
    char                    parent_lot_handle[NEST_NAMELEN];
    char                    owner[NEST_NAMELEN];
    unsigned int            is_user;
    char                    lot_users[NEST_MESSAGESIZE];
    char                    dev[NEST_NAMELEN];
    nestLotSize             total_size;
    nestLotSize             size;
    nestLotSize             used;
    int                     start_time;
    int                     duration;
    int                     time_left;
    int                     cost;
    int                     lot_type;
} NestLot;

typedef struct
NestLotList_s {
    NestLot                stat;
    struct NestLotList_s   *next;
} NestLotList;

// Function types
typedef void NestProgressFunction( nestFileSize sofar,
                                   nestFileSize total,
                                   void *args );

typedef void NestListDirFunction( const NestFileInfo *finfo,
                                  void *args);

typedef void NestProxyListFunction( char *proxy, void *args);

typedef void NestAclFunction( const NestACE *ace, void *CbData );


    // opening and closing connections
NestReplyStatus NestOpenConnection( NestConnection &con,
                                    const char *nestlocation,
                                    int port = NEST_SERVER_PORT,
									const char *user = NULL,
									gss_cred_id_t cred = GSS_C_NO_CREDENTIAL );

NestReplyStatus NestOpenAnonymously( NestConnection &con,
                                    const char *nestlocation,
                                    int port = NEST_SERVER_PORT );  

NestReplyStatus NestCloseConnection( NestConnection con );


    // sending
NestReplyStatus NestSendFile(const char *localfile,
                             const char *remotefile,
                             NestConnection);

NestReplyStatus NestSendFile(const char *localfile,
                             const char *remotefile,
                             NestConnection,
                             NestProgressFunction *,
                             void *args );

NestReplyStatus NestWriteBytes ( NestConnection con,
                                 const char *remotefile,
                                 nestFileSize offset,
                                 const char *buffer,
                                 int size,
                                 bool ack_wanted );

NestReplyStatus NestThirdPartyPut( const char *nestfile,
                                   const char *remotehost,
                                   const char *remotefile,
                                   const char *remoteprotocol,
                                   NestConnection con );

    // receiving
NestReplyStatus NestReceiveFile( const char *remotefile,
                                 const char *localfile,
                                 NestConnection );

NestReplyStatus NestReceiveFile( const char *remotefile,
                                 const char *localfile,
                                 NestConnection,
                                 NestProgressFunction *,
                                 void *args );

NestReplyStatus NestReadBytes( NestConnection con,
                               const char *remotefile,
                               nestFileSize offset,
                               char *buf,
                               int size,
                               int &bytes_read );

NestReplyStatus NestThirdPartyGet( const char *nestfile,
                                   const char *remotehost,
                                   const char *remotefile,
                                   const char *remoteprotocol,
                                   NestConnection con );

    // collections
NestReplyStatus NestNameCollection( const char *path,
									const char *name,
									NestConnection con );

NestReplyStatus NestUnnameCollection( const char *name,
									  NestConnection con );
NestReplyStatus NestDateCollection( const char *name,
									time_t date,
									NestConnection con );
NestReplyStatus NestVersionCollection( const char *name,
									   const char *version,
									   NestConnection con );
NestReplyStatus NestCommentCollection( const char *name,
									const char *comment,
									NestConnection con );


    // group and user management
NestReplyStatus NestAddUser( NestConnection con,
                             const char *user,
                             const char *group,
                             const char *identity = NULL );

NestReplyStatus NestAddGroup( NestConnection con,
                              const char *group);

NestReplyStatus NestDelUser( NestConnection con,
                             const char *user);

NestReplyStatus NestDelGroup( NestConnection con,
                              const char *group);

NestReplyStatus NestAddUserToGroup( NestConnection con,
									const char *user,
									const char *group);

NestReplyStatus NestDelUserFromGroup( NestConnection con,
									  const char *user,
									  const char *group);

    // lot management
NestReplyStatus NestRequestUserLot( NestLotHandle *handle,
                                    NestConnection con,
                                    nestLotSize bytes,
                                    int seconds,
                                    const char *user = NULL,
                                    const char *parent_user = NULL,
                                    const char *parent_group = NULL,
                                    NestLotHandle *parent_handle = NULL,
                                    int lot_type = G_LOT);

NestReplyStatus NestRequestGroupLot( NestLotHandle *handle,
                                     NestConnection con,
                                     nestLotSize bytes,
                                     int seconds,
                                     const char *group,
                                     const char *parent_user = NULL,
                                     const char *parent_group = NULL,
                                     NestLotHandle *parent_handle = NULL,
                                     int lot_type = G_LOT);

NestReplyStatus NestListUserLots( NestLotList **list,
                                  NestConnection con,
                                  const char *user = NULL );

NestReplyStatus NestListGroupLots( NestLotList **list,
                               NestConnection con,
                               const char *group );

NestReplyStatus NestGetLotStat( NestLot *result,
								NestConnection con,
								NestLotHandle *handle );

NestReplyStatus NestUpdateLot( NestConnection con,
							   NestLotHandle *handle,
							   nestLotSize more_bytes,
							   int more_seconds );

NestReplyStatus NestChangeLotType( NestConnection con,
                                   NestLotHandle *handle,
                                   int lot_type,
                                   int more_seconds);

NestReplyStatus NestLotMoveFile( NestConnection con,
                                 const char *file,
                                 NestLotHandle *source,
                                 NestLotHandle *dest,
                                 nestLotSize move_size_in_bytes );

NestReplyStatus NestTerminateLot( NestConnection con,
                              NestLotHandle *handle);

NestReplyStatus NestAddUserToLot( NestConnection con,
                                  NestLotHandle *handle,
                                  const char *user);
   
NestReplyStatus NestDeleteUserFromLot( NestConnection con,
                                       NestLotHandle *handle,
                                       const char *user);

NestReplyStatus NestAttachLotToFile( NestConnection con,
                                     NestLotHandle *handle,
                                     const char *file);

NestReplyStatus NestDetachLotFromFile( NestConnection con,
                                       NestLotHandle *handle,
                                       const char *file);

NestReplyStatus NestChangeUserID( const char *username,
                                  NestConnection con,
                                  const char *server,
								  gss_cred_id_t cred = GSS_C_NO_CREDENTIAL );

    // proxy stuff
NestReplyStatus NestProxyRegister( const char *nestlocation,
								   NestConnection con );

NestReplyStatus NestProxyQuery( NestConnection con,
                                NestProxyListFunction *foreachproxy,
                                void *args );

    // miscellaneous
NestReplyStatus NestFormat( float &bytesavailable,
                            float &totalbytes,
                            NestConnection con );

NestReplyStatus NestSync( NestConnection con );

NestReplyStatus NestAvailableSpace( nestLotSize &bytesavailable,
                                    nestLotSize &totalbytes,
                                    NestConnection con);

NestReplyStatus NestDebug( const char *debug_levels,
                           NestConnection con );

NestReplyStatus NestListFiles( NestConnection con,
                               NestListDirFunction *foreachfile,
                               void *args );

NestReplyStatus NestFilesize( nestFileSize &filesize,
                              const char *remotefile,
                              NestConnection con );

NestReplyStatus NestFileinfo( NestConnection con,
							  const char *name,
							  NestFileInfo *info );

NestReplyStatus NestRemoveFile( const char *remotefile,
                                NestConnection con );

NestReplyStatus NestRenameFile( const char *oldfile,
                                const char *newfile,
                                NestConnection con );

NestReplyStatus NestGetTimestamp( long &timestamp,
                                  NestConnection con );

NestReplyStatus NestGetPwd( char **buf,
                        NestConnection con );

NestReplyStatus NestGetPhysicalName( char **buf,
                                 const char *file,
                                 NestConnection con );

/*
   this can be easily turned into a "real" classad if the client wants to:
    ClassAdParser cap;
    ClassAd *ad = cap->ParseClassAd( adc );
    ...
    free( ad );
*/
NestReplyStatus NestRequestAd( NestConnection con, char **ad, unsigned *adlen );

NestReplyStatus NestGetAcl( NestConnection con,
                            const char *file,
                            const char *user,
                            NestAclFunction *Callback,
                            void *CbData,
                            bool is_group );
NestReplyStatus NestSetAcl( NestConnection con, 
                            const char *file,
                            const char *user,
                            nest_access_rights_t access,
                            bool is_group );

NestReplyStatus NestGetVersion( char *buf, int buflen,
								NestConnection con );

NestReplyStatus NestGetLibVersion( char *buf, int buflen );

NestReplyStatus NestChangeDir( const char *dir,
							   NestConnection con );

NestReplyStatus NestMkDir( const char *dir,
                       NestConnection con );

NestReplyStatus NestRmDir( const char *dir,
                       NestConnection con );

NestReplyStatus NestRmRecursive( const char *dir,
                       NestConnection con );

int NestReplyStatusToErrno( NestReplyStatus error );

NestReplyStatus NestLink( const char *oldPath,
                          const char *newPath,
                          int isSymbolic,
                          NestConnection con );
    
NestReplyStatus NestCopy( const char *oldPath,
                          const char *newPath,
                          NestConnection con);

NestReplyStatus NestRemoteShutdown( NestConnection con );

void NestFormatAccessString( nest_access_rights_t access,
                             char *access_string );

NestReplyStatus NestParseAccessString( nest_access_rights_t *access,
                                       const char *string );
#endif
